/* -----------------------------------------------------------------------------
 * Copyright (c) 2013-2015 ARM Ltd.
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from
 * the use of this software. Permission is granted to anyone to use this
 * software for any purpose, including commercial applications, and to alter
 * it and redistribute it freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software in
 *    a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 *
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 3. This notice may not be removed or altered from any source distribution.
 *
 *
 * $Date:        05. June 2015
 * $Revision:    V1.2
 *
 * Driver:       Driver_ETH_MAC0
 * Configured:   via RTE_Device.h configuration file
 * Project:      Ethernet Media Access (MAC) Driver for Freescale MK64F
 * --------------------------------------------------------------------------
 * Use the following configuration settings in the middleware component
 * to connect to this driver.
 *
 *   Configuration Setting                     Value
 *   ---------------------                     -----
 *   Connect to hardware via Driver_ETH_MAC# = 0
 * -------------------------------------------------------------------------- */

/* History:
 *  Version 1.2
 *    Updated for KSDK 1.2.0
 *  Version 1.1
 *    Updated initialization, uninitialization and power procedures
 *  Version 1.0
 *    Initial release
 */

#include "EMAC_MK64F.h"

#define ARM_ETH_MAC_DRV_VERSION ARM_DRIVER_VERSION_MAJOR_MINOR(1,2) /* driver version */


/* Receive/transmit Checksum offload enable */
#ifndef EMAC_CHECKSUM_OFFLOAD
#define EMAC_CHECKSUM_OFFLOAD   1
#endif

/* EMAC Memory Buffer configuration */
#define EMAC_RX_BUF_CNT         4U      /* 0x1800 for Rx (4*1536=6K)          */
#define EMAC_TX_BUF_CNT         2U      /* 0x0C00 for Tx (2*1536=3K)          */
#define EMAC_BUF_SIZE           1536U   /* ETH Receive/Transmit buffer size   */


/* Driver Version */
static const ARM_DRIVER_VERSION DriverVersion = {
  ARM_ETH_MAC_API_VERSION,
  ARM_ETH_MAC_DRV_VERSION
};

/* Driver Capabilities */
static const ARM_ETH_MAC_CAPABILITIES DriverCapabilities = {
  (EMAC_CHECKSUM_OFFLOAD != 0) ? 1U : 0U, /* checksum_offload_rx_ip4  */
  (EMAC_CHECKSUM_OFFLOAD != 0) ? 1U : 0U, /* checksum_offload_rx_ip6  */
  (EMAC_CHECKSUM_OFFLOAD != 0) ? 1U : 0U, /* checksum_offload_rx_udp  */
  (EMAC_CHECKSUM_OFFLOAD != 0) ? 1U : 0U, /* checksum_offload_rx_tcp  */
  (EMAC_CHECKSUM_OFFLOAD != 0) ? 1U : 0U, /* checksum_offload_rx_icmp */
  (EMAC_CHECKSUM_OFFLOAD != 0) ? 1U : 0U, /* checksum_offload_tx_ip4  */
  (EMAC_CHECKSUM_OFFLOAD != 0) ? 1U : 0U, /* checksum_offload_tx_ip6  */
  (EMAC_CHECKSUM_OFFLOAD != 0) ? 1U : 0U, /* checksum_offload_tx_udp  */
  (EMAC_CHECKSUM_OFFLOAD != 0) ? 1U : 0U, /* checksum_offload_tx_tcp  */
  (EMAC_CHECKSUM_OFFLOAD != 0) ? 1U : 0U, /* checksum_offload_tx_icmp */
  (RTE_ENET_MII_MODE == 0) ?
  ARM_ETH_INTERFACE_MII :
  ARM_ETH_INTERFACE_RMII,                 /* media_interface          */
  0U,                                     /* mac_address              */
  1U,                                     /* event_rx_frame           */
  1U,                                     /* event_tx_frame           */
  1U,                                     /* event_wakeup             */
  0U                                      /* precision_timer          */
};


/* Local variables */
static enet_special_maccfg_t spec_cfg;
static enet_config_rmii_t    mii_cfg;
static enet_config_rx_fifo_t rx_fifo_cfg;
static enet_config_tx_fifo_t tx_fifo_cfg;

static __align(16) enet_bd_struct_t Rx_Desc[EMAC_RX_BUF_CNT];
static __align(16) enet_bd_struct_t Tx_Desc[EMAC_TX_BUF_CNT];

static __align(16) uint8_t Rx_Buf[EMAC_RX_BUF_CNT][EMAC_BUF_SIZE];
static __align(16) uint8_t Tx_Buf[EMAC_TX_BUF_CNT][EMAC_BUF_SIZE];

static EMAC_INFO Emac = { 0 };


/**
  \fn          uint32_t crc32_8bit_rev (uint32_t crc32, uint8_t val)
  \brief       Calculate 32-bit CRC (Polynom: 0x04C11DB7, data bit-reversed).
  \param[in]   crc32  CRC initial value
  \param[in]   val    Input value
  \return      Calculated CRC value
*/
static uint32_t crc32_8bit_rev (uint32_t crc32, uint8_t val) {
  uint32_t n;

  crc32 ^= __RBIT (val);
  for (n = 8U; n; n--) {
    if (crc32 & 0x80000000U) {
      crc32 <<= 1U;
      crc32  ^= 0x04C11DB7U;
    } else {
      crc32 <<= 1U;
    }
  }
  return (crc32);
}


/**
  \fn          uint32_t crc32_data (const uint8_t *data, uint32_t len)
  \brief       Calculate standard 32-bit Ethernet CRC.
  \param[in]   data  Pointer to buffer containing the data
  \param[in]   len   Data length in bytes
  \return      Calculated CRC value
*/
static uint32_t crc32_data (const uint8_t *data, uint32_t len) {
  uint32_t cnt, crc;

  crc = 0xFFFFFFFFU;
  for (cnt = len; cnt; cnt--) {
    crc = crc32_8bit_rev (crc, *data);
    data++;
  }
  return (crc ^ 0xFFFFFFFFU);
}


/**
  \fn          ARM_DRIVER_VERSION ARM_ETH_MAC_GetVersion (void)
  \brief       Get driver version.
  \return      \ref ARM_DRIVER_VERSION
*/
static ARM_DRIVER_VERSION GetVersion (void) {
  return DriverVersion;
}


/**
  \fn          ARM_ETH_MAC_CAPABILITIES GetCapabilities (void)
  \brief       Get driver capabilities.
  \return      \ref ARM_ETH_MAC_CAPABILITIES
*/
static ARM_ETH_MAC_CAPABILITIES GetCapabilities (void) {
  return DriverCapabilities;
}


/**
  \fn          int32_t Initialize (ARM_ETH_MAC_SignalEvent_t cb_event)
  \brief       Initialize Ethernet MAC Device.
  \param[in]   cb_event  Pointer to \ref ARM_ETH_MAC_SignalEvent
  \return      \ref execution_status
*/
static int32_t Initialize (ARM_ETH_MAC_SignalEvent_t cb_event) {

  if (Emac.flags & EMAC_FLAG_INIT) { return ARM_DRIVER_OK; }

  SystemCoreClockUpdate();

  /* Select clock source */
  #if defined(ENET_OSCERCLK_ENABLE) && (ENET_OSCERCLK_ENABLE == 1)
  /* Enable external clock input */
  OSC_HAL_SetExternalRefClkCmd (OSC, true);
  #endif

  #if (RTE_ENET_MII_MODE == 1)
  /* Set clock source for RMII */
  CLOCK_HAL_SetRmiiSrc (SIM, 0, ENET_RMII_CLOCK_SOURCE);
  #endif

  /* Assign pointers */
  Emac.cfg.macAddr       = &Emac.addr[0];
  Emac.cfg.rxFifoPtr     = &rx_fifo_cfg;
  Emac.cfg.txFifoPtr     = &tx_fifo_cfg;
  Emac.cfg.macSpecialCfg = &spec_cfg;
  Emac.cfg.rmiiCfgPtr    = &mii_cfg;

  /* Setup Tx/Rx descriptors */
  Emac.desc_cfg.txBds           = &Tx_Desc[0];
  Emac.desc_cfg.txBuffer        = &Tx_Buf[0][0];
  Emac.desc_cfg.txBdNumber      =  EMAC_TX_BUF_CNT;
  Emac.desc_cfg.txBuffSizeAlign =  EMAC_BUF_SIZE;

  Emac.desc_cfg.rxBds           = &Rx_Desc[0];
  Emac.desc_cfg.rxBuffer        = &Rx_Buf[0][0];
  Emac.desc_cfg.rxBdNumber      =  EMAC_RX_BUF_CNT;
  Emac.desc_cfg.rxBuffSizeAlign =  EMAC_BUF_SIZE;

  /* Register driver callback function */
  Emac.cb_event = cb_event;
  Emac.flags    = EMAC_FLAG_INIT;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t Uninitialize (void)
  \brief       De-initialize Ethernet MAC Device.
  \return      \ref execution_status
*/
static int32_t Uninitialize (void) {

  /* Reset clock source selection */
  #if (RTE_ENET_MII_MODE == 1)
  CLOCK_HAL_SetRmiiSrc (SIM, 0, kClockRmiiSrcExtalClk);
  #endif

  Emac.flags = 0U;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t PowerControl (ARM_POWER_STATE state)
  \brief       Control Ethernet MAC Device Power.
  \param[in]   state  Power state
  \return      \ref execution_status
*/
static int32_t PowerControl (ARM_POWER_STATE state) {

  switch (state) {
    case ARM_POWER_OFF:
      /* Disable Ethernet MAC peripheral */
      ENET_HAL_Disable (ENET);

      /* Disable ENET interrupts in NVCI */
      NVIC_DisableIRQ (ENET_Transmit_IRQn);
      NVIC_DisableIRQ (ENET_Receive_IRQn);
      NVIC_DisableIRQ (ENET_Error_IRQn);

      /* Disable ENET peripheral clock */
      SIM_HAL_DisableClock (SIM, kSimClockGateEnet0);

      Emac.flags = EMAC_FLAG_INIT;
      break;

    case ARM_POWER_LOW:
      return ARM_DRIVER_ERROR_UNSUPPORTED;

    case ARM_POWER_FULL:
      if (Emac.flags & EMAC_FLAG_POWER) {
        /* Driver already powered */
        break;
      }
      /* Clear variables */
      Emac.tx_index = 0;
      Emac.rx_index = 0;

      /* Enable ENET peripheral clock */
      SIM_HAL_EnableClock (SIM, kSimClockGateEnet0);

      /* Setup MAC configuration */
      Emac.cfg.macMode = kEnetMacNormalMode;

      Emac.cfg.rmiiCfgPtr->mode             = kEnetCfgMii;
      Emac.cfg.rmiiCfgPtr->speed            = kEnetCfgSpeed100M;
      Emac.cfg.rmiiCfgPtr->duplex           = kEnetCfgHalfDuplex;
      Emac.cfg.rmiiCfgPtr->isRxOnTxDisabled = false;
      Emac.cfg.rmiiCfgPtr->isLoopEnabled    = false;

      Emac.cfg.macCtlConfigure = 0;

      #if (EMAC_CHECKSUM_OFFLOAD)
      Emac.cfg.txFifoPtr->isStoreForwardEnabled = true;
      #else
      Emac.cfg.txFifoPtr->isStoreForwardEnabled = false;
      #endif
      Emac.cfg.txFifoPtr->txFifoWrite   = 0;
      Emac.cfg.txFifoPtr->txEmpty       = 0;
      Emac.cfg.txFifoPtr->txAlmostEmpty = 4;
      Emac.cfg.txFifoPtr->txAlmostFull  = 8;

      Emac.cfg.rxFifoPtr->rxFull        = 0; /* Store and forward enabled */
      Emac.cfg.rxFifoPtr->rxAlmostFull  = 4;
      Emac.cfg.rxFifoPtr->rxEmpty       = 0;
      Emac.cfg.rxFifoPtr->rxAlmostEmpty = 4;

      Emac.cfg.rxAccelerCfg    = 0;
      Emac.cfg.txAccelerCfg    = 0;
      Emac.cfg.pauseDuration   = 0;

      Emac.cfg.macSpecialCfg->rxMaxFrameLen    = EMAC_BUF_SIZE; /* ENET->ECR.MAX_FL */
      Emac.cfg.macSpecialCfg->rxTruncLen       = 0x7FF;         /* ENET->FTRL */
      Emac.cfg.macSpecialCfg->txInterPacketGap = 0x0C;          /* ENET->TIPG */

      /* Apply configuration */
      ENET_HAL_Config (ENET, &Emac.cfg, SystemCoreClock, &Emac.desc_cfg);

      /* Enable ENET peripheral interrupts in NVIC */
      NVIC_EnableIRQ (ENET_Transmit_IRQn);
      NVIC_EnableIRQ (ENET_Receive_IRQn);
      NVIC_EnableIRQ (ENET_Error_IRQn);

      Emac.flags |= EMAC_FLAG_POWER;
      break;

    default:
      return ARM_DRIVER_ERROR_UNSUPPORTED;
  }

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t GetMacAddress (ARM_ETH_MAC_ADDR *ptr_addr)
  \brief       Get Ethernet MAC Address.
  \param[in]   ptr_addr  Pointer to address
  \return      \ref execution_status
*/
static int32_t GetMacAddress (ARM_ETH_MAC_ADDR *ptr_addr) {

  if (ptr_addr == NULL) {
    return ARM_DRIVER_ERROR_PARAMETER;
  }

  ptr_addr->b[5] = (uint8_t)(ENET->PAUR >> 16);
  ptr_addr->b[4] = (uint8_t)(ENET->PAUR >> 24);
  ptr_addr->b[3] = (uint8_t)(ENET->PALR);
  ptr_addr->b[2] = (uint8_t)(ENET->PALR >>  8);
  ptr_addr->b[1] = (uint8_t)(ENET->PALR >> 16);
  ptr_addr->b[0] = (uint8_t)(ENET->PALR >> 24);

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t SetMacAddress (const ARM_ETH_MAC_ADDR *ptr_addr)
  \brief       Set Ethernet MAC Address.
  \param[in]   ptr_addr  Pointer to address
  \return      \ref execution_status
*/
static int32_t SetMacAddress (const ARM_ETH_MAC_ADDR *ptr_addr) {

  if (ptr_addr == NULL) {
    return ARM_DRIVER_ERROR_PARAMETER;
  }

  if ((Emac.flags & EMAC_FLAG_POWER) == 0) {
    /* Driver not yet powered */
    return ARM_DRIVER_ERROR;
  }
  memcpy (Emac.addr, ptr_addr, 6);
  /* Set Ethernet MAC Address registers */
  ENET->PALR = Emac.addr[0] << 24 |
               Emac.addr[1] << 16 |
               Emac.addr[2] <<  8 |
               Emac.addr[3]       ;
  ENET->PAUR = Emac.addr[4] << 24 |
               Emac.addr[5] << 16 ;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t SetAddressFilter (const ARM_ETH_MAC_ADDR *ptr_addr,
                                               uint32_t          num_addr)
  \brief       Configure Address Filter.
  \param[in]   ptr_addr  Pointer to addresses
  \param[in]   num_addr  Number of addresses to configure
  \return      \ref execution_status
*/
static int32_t SetAddressFilter (const ARM_ETH_MAC_ADDR *ptr_addr, uint32_t num_addr) {
  uint32_t cnt, crc;

  if ((ptr_addr == NULL) && (num_addr != 0U)) {
    return ARM_DRIVER_ERROR_PARAMETER;
  }

  if ((Emac.flags & EMAC_FLAG_POWER) == 0U) {
    /* Driver not yet powered */
    return ARM_DRIVER_ERROR;
  }

  ENET->GALR = 0U;
  ENET->GAUR = 0U;

  if (num_addr != 0U) {
    for (cnt = 0U; cnt < num_addr; cnt++) {
      /* Calculate CRC-32 on the multicast group address */
      crc = crc32_data (&ptr_addr[cnt].b[0], 6U);
      /* Add multicast group to hash table */
      ENET_HAL_SetMulticastAddrHash (ENET, crc, kEnetSpecialAddressEnable);
    }
  }

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t SendFrame (const uint8_t *frame, uint32_t len, uint32_t flags)
  \brief       Send Ethernet frame.
  \param[in]   frame  Pointer to frame buffer with data to send
  \param[in]   len    Frame buffer length in bytes
  \param[in]   flags  Frame transmit flags (see ARM_ETH_MAC_TX_FRAME_...)
  \return      \ref execution_status
*/
static int32_t SendFrame (const uint8_t *frame, uint32_t len, uint32_t flags) {
  static uint8_t *frame_end = NULL;
  static uint32_t frame_len = 0U;
  enet_bd_attr_t bd;
  uint64_t mask;
  uint8_t *dst;

  if ((frame == NULL) || (len == 0U)) {
    return ARM_DRIVER_ERROR_PARAMETER;
  }

  if ((Emac.flags & EMAC_FLAG_POWER) == 0U) {
    /* Driver not yet powered */
    return ARM_DRIVER_ERROR;
  }

  dst = frame_end;

  mask = ENET_TX_BD_READY_FLAG_MASK | ENET_BD_LEN_MASK;
  ENET_HAL_GetBufDescripAttr (&Tx_Desc[Emac.tx_index], mask, &bd);

  if (dst == NULL) {
    if (bd.flags & ENET_TX_BD_READY_FLAG) {
      /* Data buffer not yet transmitted or currently transmits */
      return ARM_DRIVER_ERROR_BUSY;
    }
    dst = ENET_HAL_GetBuffDescripData(&Tx_Desc[Emac.tx_index]);
    frame_len = len;
  }
  else {
    /* Sending fragments, add fragment length to total length */
    frame_len += bd.bdLen;
  }

  /* Copy data fragment to Tx buffer */
  for ( ; len > 7U; dst += 8, frame += 8, len -= 8U) {
    ((__packed uint32_t *)dst)[0] = ((__packed uint32_t *)frame)[0];
    ((__packed uint32_t *)dst)[1] = ((__packed uint32_t *)frame)[1];
  }
  /* Copy remaining 7 bytes */
  for ( ; len > 1U; dst += 2, frame += 2, len -= 2U) {
    ((__packed uint16_t *)dst)[0] = ((__packed uint16_t *)frame)[0];
  }
  if (len > 0U) { dst++[0] = frame++[0]; }

  if (flags & ARM_ETH_MAC_TX_FRAME_FRAGMENT) {
    /* More data to come, remember current write position */
    frame_end = dst;
  }
  else {
    /* Frame ready to be sent, release it to DMA */

    /* Update frame descriptor */
    Tx_Desc[Emac.tx_index].length = BD_SHORTSWAP(frame_len);
    Tx_Desc[Emac.tx_index].control |= kEnetTxBdLast;
    Tx_Desc[Emac.tx_index].control |= kEnetTxBdTransmitCrc;

    /* Mask/Unmask TXF interrupt */
    if (flags & ARM_ETH_MAC_TX_FRAME_EVENT) {
      Tx_Desc[Emac.tx_index].controlExtend1 |=  kEnetTxBdTxInterrupt;
    } else {
      Tx_Desc[Emac.tx_index].controlExtend1 &= ~kEnetTxBdTxInterrupt;
    }

    #if (EMAC_CHECKSUM_OFFLOAD)
    /* Set PINS and IINS bits */
    Tx_Desc[Emac.tx_index].controlExtend1 |= (1 << 12) | (1 << 11);
    #endif
    Tx_Desc[Emac.tx_index].controlExtend2  = 0;

    Tx_Desc[Emac.tx_index].control |= kEnetTxBdReady;

    /* Start frame transmission */
    ENET_HAL_SetTxBdActive (ENET);

    frame_len = 0;

    Emac.tx_index++;
    if (Emac.tx_index == EMAC_TX_BUF_CNT) { Emac.tx_index = 0; }
  }

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t ReadFrame (uint8_t *frame, uint32_t len)
  \brief       Read data of received Ethernet frame.
  \param[in]   frame  Pointer to frame buffer for data to read into
  \param[in]   len    Frame buffer length in bytes
  \return      number of data bytes read or execution status
                 - value >= 0: number of data bytes read
                 - value < 0: error occurred, value is execution status as defined with \ref execution_status 
*/
static int32_t ReadFrame (uint8_t *frame, uint32_t len) {
  uint8_t const *src;

  if ((Emac.flags & EMAC_FLAG_POWER) == 0U) {
    /* Driver not yet powered */
    return ARM_DRIVER_ERROR;
  }

  src = ENET_HAL_GetBuffDescripData(&Rx_Desc[Emac.rx_index]);

  /* Fast-copy data to frame buffer */
  for ( ; len > 7U; frame += 8, src += 8, len -= 8U) {
    ((__packed uint32_t *)frame)[0] = ((uint32_t *)src)[0];
    ((__packed uint32_t *)frame)[1] = ((uint32_t *)src)[1];
  }
  /* Copy remaining 7 bytes */
  for ( ; len > 1U; frame += 2, src += 2, len -= 2U) {
    ((__packed uint16_t *)frame)[0] = ((uint16_t *)src)[0];
  }
  if (len > 0U) { frame[0] = src[0]; }

  /* Release descriptor to uDMA */
  ENET_HAL_ClrRxBdAfterHandled (&Rx_Desc[Emac.rx_index], NULL, false);

  Emac.rx_index++;
  if (Emac.rx_index == EMAC_RX_BUF_CNT) { Emac.rx_index = 0U; }

  /* Start MAC receive descriptor ring pooling */
  ENET_HAL_SetRxBdActive (ENET);

  return ((uint32_t)len);
}

/**
  \fn          uint32_t GetRxFrameSize (void)
  \brief       Get size of received Ethernet frame.
  \return      number of bytes in received frame
*/
static uint32_t GetRxFrameSize (void) {
  enet_bd_attr_t bd;
  uint64_t mask;

  if ((Emac.flags & EMAC_FLAG_POWER) == 0U) {
    /* Driver not yet powered */
    return (0U);
  }
  mask = ENET_RX_BD_EMPTY_FLAG_MASK | ENET_BD_LEN_MASK;

  ENET_HAL_GetBufDescripAttr (&Rx_Desc[Emac.rx_index], mask, &bd);

  if (bd.flags & ENET_RX_BD_EMPTY_FLAG) {
    /* Data buffer is empty or reception in progress */
    return (0U);
  }
  return (bd.bdLen);
}


/**
  \fn          int32_t GetRxFrameTime (ARM_ETH_MAC_TIME *time)
  \brief       Get time of received Ethernet frame.
  \param[in]   time  Pointer to time structure for data to read into
  \return      \ref execution_status
*/
static int32_t GetRxFrameTime (ARM_ETH_MAC_TIME *time) {
  return ARM_DRIVER_ERROR_UNSUPPORTED;
}

/**
  \fn          int32_t GetTxFrameTime (ARM_ETH_MAC_TIME *time)
  \brief       Get time of transmitted Ethernet frame.
  \param[in]   time  Pointer to time structure for data to read into
  \return      \ref execution_status
*/
static int32_t GetTxFrameTime (ARM_ETH_MAC_TIME *time) {
  return ARM_DRIVER_ERROR_UNSUPPORTED;
}

/**
  \fn          int32_t Control (uint32_t control, uint32_t arg)
  \brief       Control Ethernet Interface.
  \param[in]   control  operation
  \param[in]   arg      argument of operation (optional)
  \return      \ref execution_status
*/
static int32_t Control (uint32_t control, uint32_t arg) {
  uint32_t i;

  if ((Emac.flags & EMAC_FLAG_POWER) == 0U) {
    /* Driver not powered */
    return ARM_DRIVER_ERROR;
  }

  switch (control) {
    case ARM_ETH_MAC_CONFIGURE:
      Emac.cfg.rmiiCfgPtr->isRxOnTxDisabled = false;

      /* Configure link speed */
      switch (arg & ARM_ETH_MAC_SPEED_Msk) {
        case ARM_ETH_MAC_SPEED_10M:
          Emac.cfg.rmiiCfgPtr->speed = kEnetCfgSpeed10M;
          break;

        case ARM_ETH_SPEED_100M:
          Emac.cfg.rmiiCfgPtr->speed = kEnetCfgSpeed100M;
          break;

        case ARM_ETH_SPEED_1G:
          return ARM_DRIVER_ERROR_UNSUPPORTED;

        default: return ARM_DRIVER_ERROR;
      }

      /* Configure duplex mode */
      switch (arg & ARM_ETH_MAC_DUPLEX_Msk) {
        case ARM_ETH_MAC_DUPLEX_HALF:
          Emac.cfg.rmiiCfgPtr->duplex = kEnetCfgHalfDuplex;
          break;

        case ARM_ETH_MAC_DUPLEX_FULL:
          Emac.cfg.rmiiCfgPtr->duplex = kEnetCfgFullDuplex;
          break;

        default: return ARM_DRIVER_ERROR;
      }

      /* Configure MII/RMII mode */
      #if (RTE_ENET_MII_MODE == 0)
      Emac.cfg.rmiiCfgPtr->mode = kEnetCfgMii;
      #else
      Emac.cfg.rmiiCfgPtr->mode = kEnetCfgRmii;
      #endif

      /* Configure loopback mode */
      if (arg & ARM_ETH_MAC_LOOPBACK) {
        Emac.cfg.rmiiCfgPtr->isLoopEnabled = true;
      } else {
        Emac.cfg.rmiiCfgPtr->isLoopEnabled = false;
      }

      #if (EMAC_CHECKSUM_OFFLOAD)
      /* Enable/Disable rx checksum verification */
      if (arg & ARM_ETH_MAC_CHECKSUM_OFFLOAD_RX) {
        Emac.cfg.macCtlConfigure |= kEnetRxAccelEnable;
        Emac.cfg.rxAccelerCfg = kEnetRxAccelIpCheckEnabled | kEnetRxAccelProtoCheckEnabled | kEnetRxAccelMacCheckEnabled;
      } else {
        Emac.cfg.macCtlConfigure &= ~kEnetRxAccelEnable;
        Emac.cfg.rxAccelerCfg = 0;
      }

      /* Enable/Disable tx checksum generation */
      if (arg & ARM_ETH_MAC_CHECKSUM_OFFLOAD_TX) {
        Emac.cfg.macCtlConfigure |= kEnetTxAccelEnable;
        Emac.cfg.txAccelerCfg = kEnetTxAccelIpCheckEnabled | kEnetTxAccelProtoCheckEnabled;
      } else {
        Emac.cfg.macCtlConfigure &= ~kEnetTxAccelEnable;
        Emac.cfg.txAccelerCfg = 0;
      }
      #else
      if (arg & (ARM_ETH_MAC_CHECKSUM_OFFLOAD_RX | ARM_ETH_MAC_CHECKSUM_OFFLOAD_TX)) {
        /* Checksum offload is disabled */
        return ARM_DRIVER_ERROR_UNSUPPORTED;
      }
      #endif

      /* Enable/Disable broadcast frame receive */
      if (arg & ARM_ETH_MAC_ADDRESS_BROADCAST) {
        Emac.cfg.macCtlConfigure &= ~kEnetRxBcRejectEnable;
      } else {
        Emac.cfg.macCtlConfigure |=  kEnetRxBcRejectEnable;
      }

      /* Enable/Disable multicast frame receive */
       if (arg & ARM_ETH_MAC_ADDRESS_MULTICAST) {
        ENET->GALR = 0xFFFFFFFFU;
        ENET->GAUR = 0xFFFFFFFFU;
      } else {
        ENET->GALR = 0U;
        ENET->GAUR = 0U;
      }

      /* Enable/Disable promiscuous mode (no filtering) */
      if (arg & ARM_ETH_MAC_ADDRESS_ALL) {
        Emac.cfg.macCtlConfigure |=  kEnetRxPromiscuousEnable;
      } else {
        Emac.cfg.macCtlConfigure &= ~kEnetRxPromiscuousEnable;
      }

      Emac.cfg.macCtlConfigure |= kEnetMacEnhancedEnable;

      /* Apply configuration */
      ENET_HAL_Config (ENET, &Emac.cfg, SystemCoreClock, &Emac.desc_cfg);

      /* Disable Rx and Tx interrupts */
      ENET_HAL_SetIntMode (ENET, kEnetTxFrameInterrupt, false);
      ENET_HAL_SetIntMode (ENET, kEnetRxFrameInterrupt, false);

      /* Ethernet MAC enable */
      ENET_HAL_Enable (ENET);
      break;

    case ARM_ETH_MAC_CONTROL_TX:
      /* Enable/disable MAC transmitter */
      ENET_HAL_SetIntMode (ENET, kEnetTxFrameInterrupt, arg != 0);
      break;

    case ARM_ETH_MAC_CONTROL_RX:
      /* Enable/disable MAC receiver */
      ENET_HAL_SetIntMode (ENET, kEnetRxFrameInterrupt, arg != 0);

      if (arg) {
        ENET_HAL_SetRxBdActive (ENET);
      }
      break;

    case ARM_ETH_MAC_FLUSH:
      /* Flush tx and rx buffers */
      if (arg == ARM_ETH_MAC_FLUSH_RX) {
        /* Disable RX interrupts */
        ENET_HAL_SetIntMode (ENET, kEnetRxFrameInterrupt, false);

        for (i = 0; i < EMAC_RX_BUF_CNT; i++) {
          ENET_HAL_ClrRxBdAfterHandled (&Rx_Desc[i], NULL, false);
        }

        /* Enable RX interrupts */
        ENET_HAL_SetIntMode (ENET, kEnetRxFrameInterrupt, true);
      }
      else {
        /* Disable TX interrupts */
        ENET_HAL_SetIntMode (ENET, kEnetTxFrameInterrupt, false);

        for (i = 0; i < EMAC_TX_BUF_CNT; i++) {
          ENET_HAL_ClrTxBdAfterSend (&Tx_Desc[i]);
        }

        /* Enable TX interrupts */
        ENET_HAL_SetIntMode (ENET, kEnetTxFrameInterrupt, true);
      }
      break;
    
    case ARM_ETH_MAC_SLEEP:
      if (arg) {
        ENET_HAL_SetIntMode (ENET, kEnetWakeupInterrupt, true);
        /* Enter sleep mode and wait for magic packet */
        ENET->ECR |= (1 << 3) | (1 << 2);
      }
      else {
        ENET_HAL_SetIntMode (ENET, kEnetWakeupInterrupt, false);
        /* Exit sleep mode */
        ENET->ECR &= ~(1 << 3);
      }
      break;

    case ARM_ETH_MAC_VLAN_FILTER:
      /* VLAN filtering is not supported */
    default:
      return ARM_DRIVER_ERROR_UNSUPPORTED;
  }
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t ControlTimer (uint32_t control, ARM_ETH_MAC_TIME *time)
  \brief       Control Precision Timer.
  \param[in]   control  operation
  \param[in]   time     Pointer to time structure
  \return      \ref execution_status
*/
static int32_t ControlTimer (uint32_t control, ARM_ETH_MAC_TIME *time) {
  return ARM_DRIVER_ERROR_UNSUPPORTED;
}

/**
  \fn          int32_t PHY_Read (uint8_t phy_addr, uint8_t reg_addr, uint16_t *data)
  \brief       Read Ethernet PHY Register through Management Interface.
  \param[in]   phy_addr  5-bit device address
  \param[in]   reg_addr  5-bit register address
  \param[out]  data      Pointer where the result is written to
  \return      \ref execution_status
*/
static int32_t PHY_Read (uint8_t phy_addr, uint8_t reg_addr, uint16_t *data) {
  uint32_t loop;

  if (data == NULL) {
    return ARM_DRIVER_ERROR_PARAMETER;
  }

  if ((Emac.flags & EMAC_FLAG_POWER) == 0U) {
    /* Driver not powered */
    return ARM_DRIVER_ERROR;
  }
  /* Clear MII Interrupt flag */
  ENET_HAL_ClearIntStatusFlag(ENET, kEnetMiiInterrupt);

  ENET_HAL_SetSMIRead (ENET, phy_addr, reg_addr, kEnetReadValidFrame);

  /* Wait until operation completed */
  loop = SystemCoreClock;
  while (ENET_HAL_GetIntStatusFlag(ENET, kEnetMiiInterrupt) == 0U) {
    loop--;
    if (loop == 0) {
      /* Loop counter timeout */
      return ARM_DRIVER_ERROR;
    }
  }

  *data = (uint16_t)ENET_HAL_GetSMIData (ENET);

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t PHY_Write (uint8_t phy_addr, uint8_t reg_addr, uint16_t data)
  \brief       Write Ethernet PHY Register through Management Interface.
  \param[in]   phy_addr  5-bit device address
  \param[in]   reg_addr  5-bit register address
  \param[in]   data      16-bit data to write
  \return      \ref execution_status
*/
static int32_t PHY_Write (uint8_t phy_addr, uint8_t reg_addr, uint16_t data) {
  uint32_t loop;

  if ((Emac.flags & EMAC_FLAG_POWER) == 0U) {
    /* Driver not powered */
    return ARM_DRIVER_ERROR;
  }
  /* Clear MII Interrupt flag */
  ENET_HAL_ClearIntStatusFlag(ENET, kEnetMiiInterrupt);

  ENET_HAL_SetSMIWrite (ENET, phy_addr, reg_addr, kEnetWriteValidFrame, data);

  /* Wait until operation completed */
  loop = SystemCoreClock;
  while (ENET_HAL_GetIntStatusFlag(ENET, kEnetMiiInterrupt) == 0U) {
    loop--;
    if (loop == 0) {
      /* Loop counter timeout */
      return ARM_DRIVER_ERROR;
    }
  }
  return ARM_DRIVER_OK;
}


/* Ethernet MAC Transmit Interrupt */
void ENET_Transmit_IRQHandler (void) {

  if (ENET_HAL_GetIntStatusFlag(ENET, kEnetTxFrameInterrupt)) {
    /* Frame has been transmitted and the corresponding descriptor updated */
    ENET_HAL_ClearIntStatusFlag (ENET, kEnetTxFrameInterrupt);

    if (Emac.cb_event) {
      Emac.cb_event (ARM_ETH_MAC_EVENT_TX_FRAME);
    }
  }
}

/* Ethernet MAC Receive Interrupt */
void ENET_Receive_IRQHandler (void) {

  if (ENET_HAL_GetIntStatusFlag(ENET, kEnetRxFrameInterrupt)) {
    /* Frame has been received and the corresponding descriptor updated */
    ENET_HAL_ClearIntStatusFlag (ENET, kEnetRxFrameInterrupt);

    if (Emac.cb_event) {
      Emac.cb_event (ARM_ETH_MAC_EVENT_RX_FRAME);
    }
  }
}

/* Ethernet MAC Error and miscelaneous Interrupt */
void ENET_Error_IRQHandler (void) {

  if (ENET_HAL_GetIntStatusFlag(ENET, kEnetWakeupInterrupt)) {
    /* Magic packet has been detected */
    ENET_HAL_ClearIntStatusFlag (ENET, kEnetWakeupInterrupt);

    if (Emac.cb_event) {
      Emac.cb_event (ARM_ETH_MAC_EVENT_WAKEUP);
    }
  }
}


/* MAC Driver Control Block */
ARM_DRIVER_ETH_MAC Driver_ETH_MAC0 = {
  GetVersion,
  GetCapabilities,
  Initialize,
  Uninitialize,
  PowerControl,
  GetMacAddress,
  SetMacAddress,
  SetAddressFilter,
  SendFrame,
  ReadFrame,
  GetRxFrameSize,
  GetRxFrameTime,
  GetTxFrameTime,
  ControlTimer,
  Control,
  PHY_Read,
  PHY_Write
};
