/* -----------------------------------------------------------------------------
 * Copyright (c) 2013-2015 ARM Ltd.
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from
 * the use of this software. Permission is granted to anyone to use this
 * software for any purpose, including commercial applications, and to alter
 * it and redistribute it freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software in
 *    a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 *
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 3. This notice may not be removed or altered from any source distribution.
 *
 *
 * $Date:        05. June 2015
 * $Revision:    V1.3
 *
 * Driver:       Driver_I2C0, Driver_I2C1, Driver_I2C2
 * Configured:   via RTE_Device.h configuration file
 * Project:      I2C Driver for Freescale MK64F
 * --------------------------------------------------------------------------
 * Use the following configuration settings in the middleware component
 * to connect to this driver.
 *
 *   Configuration Setting                 Value   I2C Interface
 *   ---------------------                 -----   -------------
 *   Connect to hardware via Driver_I2C# = 0       use I2C0
 *   Connect to hardware via Driver_I2C# = 1       use I2C1
 *   Connect to hardware via Driver_I2C# = 2       use I2C2
 * -------------------------------------------------------------------------- */

/* History:
 *  Version 1.3
 *    Updated for KSDK 1.2.0
 *  Version 1.2
 *    Updated initialization, uninitialization and power procedures
 *  Version 1.1
 *    Bus clear control implemented
 *  Version 1.0
 *    Initial release
 */

#include "I2C_MK64F.h"

#define ARM_I2C_DRV_VERSION ARM_DRIVER_VERSION_MAJOR_MINOR(1,3)    /* driver version */


/* Driver Version */
static const ARM_DRIVER_VERSION DriverVersion = {
  ARM_I2C_API_VERSION,
  ARM_I2C_DRV_VERSION
};

/* Driver Capabilities */
static const ARM_I2C_CAPABILITIES DriverCapabilities = { 
  1U  /* address_10_bit */
};


static void I2Cx_DMA_Callback (void *par, edma_chn_status_t ch);

#if defined(RTE_I2C0)
/* I2C0 DMA Configuration */
#if (RTE_I2C0_DMA_RX != 0)
edma_chn_state_t I2C0_DMA_RxState;
#endif

#if (RTE_I2C0_DMA_TX != 0)
edma_chn_state_t I2C0_DMA_TxState;
#endif

#if ((RTE_I2C0_DMA_TX != 0) || (RTE_I2C0_DMA_RX != 0))
edma_transfer_config_t I2C0_DMA_Cfg;

static I2C_DMA I2C0_DMA = {
  kDmaRequestMux0I2C0,
  #if (RTE_I2C0_DMA_RX != 0)
  &I2C0_DMA_RxState,
  #else
  NULL,
  #endif
  #if (RTE_I2C0_DMA_TX != 0)
  &I2C0_DMA_TxState,
  #else
  NULL,
  #endif
  &I2C0_DMA_Cfg
};
#endif

/* I2C0 Information (Run-Time) */
static I2C_CTRL I2C0_Ctrl;

/* I2C0 Resources */
static I2C_RESOURCES I2C0_Resources = {
  I2C0,
  &I2C0_Ctrl,
  I2C0_IRQn,
  kSimClockGateI2c0,
  #if ((RTE_I2C0_DMA_TX != 0) || (RTE_I2C0_DMA_RX != 0))
  &I2C0_DMA,
  #else
  NULL,
  #endif
  (PORT_Type *)
  RTE_I2C0_SDA_PORT,
  RTE_I2C0_SDA_PIN,
  RTE_I2C0_SDA_MUX,
  (PORT_Type *)
  RTE_I2C0_SCL_PORT,
  RTE_I2C0_SCL_PIN,
  RTE_I2C0_SCL_MUX
};
#endif /* RTE_I2C0 */

#if defined(RTE_I2C1)
/* I2C1 DMA Configuration */
#if (RTE_I2C1_DMA_RX != 0)
edma_chn_state_t I2C1_DMA_RxState;
#endif

#if (RTE_I2C1_DMA_TX != 0)
edma_chn_state_t I2C1_DMA_TxState;
#endif

#if ((RTE_I2C1_DMA_TX != 0) || (RTE_I2C1_DMA_RX != 0))
edma_transfer_config_t I2C1_DMA_Cfg;

static I2C_DMA I2C1_DMA = {
  kDmaRequestMux0I2C1I2C2,
  #if (RTE_I2C1_DMA_RX != 0)
  &I2C1_DMA_RxState,
  #else
  NULL,
  #endif
  #if (RTE_I2C1_DMA_TX != 0)
  &I2C1_DMA_TxState,
  #else
  NULL,
  #endif
  &I2C1_DMA_Cfg
};
#endif

/* I2C1 Information (Run-Time) */
static I2C_CTRL I2C1_Ctrl;

/* I2C1 Resources */
static I2C_RESOURCES I2C1_Resources = {
  I2C1,
  &I2C1_Ctrl,
  I2C1_IRQn,
  kSimClockGateI2c1,
  #if ((RTE_I2C1_DMA_TX != 0) || (RTE_I2C1_DMA_RX != 0))
  &I2C1_DMA,
  #else
  NULL,
  #endif
  (PORT_Type *)
  RTE_I2C1_SDA_PORT,
  RTE_I2C1_SDA_PIN,
  RTE_I2C1_SDA_MUX,
  (PORT_Type *)
  RTE_I2C1_SCL_PORT,
  RTE_I2C1_SCL_PIN,
  RTE_I2C1_SCL_MUX
};
#endif /* RTE_I2C1 */

#if defined(RTE_I2C2)
/* I2C2 DMA Configuration */
#if (RTE_I2C2_DMA_RX != 0)
edma_chn_state_t I2C2_DMA_RxState;
#endif

#if (RTE_I2C2_DMA_TX != 0)
edma_chn_state_t I2C2_DMA_TxState;
#endif

#if ((RTE_I2C2_DMA_TX != 0) || (RTE_I2C2_DMA_RX != 0))
edma_transfer_config_t I2C2_DMA_Cfg;

static I2C_DMA I2C2_DMA = {
  kDmaRequestMux0I2C1I2C2,
  #if (RTE_I2C2_DMA_RX != 0)
  &I2C2_DMA_RxState,
  #else
  NULL,
  #endif
  #if (RTE_I2C2_DMA_TX != 0)
  &I2C2_DMA_TxState,
  #else
  NULL,
  #endif
  &I2C2_DMA_Cfg
};
#endif

/* I2C2 Information (Run-Time) */
static I2C_CTRL I2C2_Ctrl;

/* I2C2 Resources */
static I2C_RESOURCES I2C2_Resources = {
  I2C2,
  &I2C2_Ctrl,
  I2C2_IRQn,
  kSimClockGateI2c2,
  #if ((RTE_I2C2_DMA_TX != 0) || (RTE_I2C2_DMA_RX != 0))
  &I2C2_DMA,
  #else
  NULL,
  #endif
  (PORT_Type *)
  RTE_I2C2_SDA_PORT,
  RTE_I2C2_SDA_PIN,
  RTE_I2C2_SDA_MUX,
  (PORT_Type *)
  RTE_I2C2_SCL_PORT,
  RTE_I2C2_SCL_PIN,
  RTE_I2C2_SCL_MUX
};
#endif /* RTE_I2C2 */


/**
  \fn          ARM_DRV_VERSION I2C_GetVersion (void)
  \brief       Get driver version.
  \return      \ref ARM_DRV_VERSION
*/
static ARM_DRIVER_VERSION I2CX_GetVersion (void) {
  return DriverVersion;
}


/**
  \fn          ARM_I2C_CAPABILITIES I2C_GetCapabilities (void)
  \brief       Get driver capabilities.
  \return      \ref ARM_I2C_CAPABILITIES
*/
static ARM_I2C_CAPABILITIES I2CX_GetCapabilities (void) {
  return DriverCapabilities;
}


/**
  \fn          static uint32_t PTx_Base (uint32_t port_base)
  \brief       Determine base address of GPIO register from base address of PORT
  \param[in]   port_base  Port base address
  \return      Gpio base address
*/
static uint32_t PTx_Base (uint32_t port_base) {
  uint32_t base;
  switch (port_base) {
    case PORTA_BASE: base = PTA_BASE; break;
    case PORTB_BASE: base = PTB_BASE; break;
    case PORTC_BASE: base = PTC_BASE; break;
    case PORTD_BASE: base = PTD_BASE; break;
    default:
    case PORTE_BASE: base = PTE_BASE; break;
  }
  return (base);
}


/**
  \fn          int32_t I2C_Initialize (ARM_I2C_SignalEvent_t cb_event, I2C_RESOURCES *i2c)
  \brief       Initialize I2C Interface.
  \param[in]   cb_event  Pointer to \ref ARM_I2C_SignalEvent
  \param[in]   i2c   Pointer to I2C resources
  \return      \ref ARM_I2C_STATUS
*/
static int32_t I2C_Initialize (ARM_I2C_SignalEvent_t cb_event, I2C_RESOURCES *i2c) {

  if (i2c->ctrl->flags & I2C_INIT) { return ARM_DRIVER_OK; }

  /* Configure pins */
  PORT_HAL_SetMuxMode      (i2c->sda_base, i2c->sda_pin, i2c->sda_mux);
  PORT_HAL_SetOpenDrainCmd (i2c->sda_base, i2c->sda_pin, true);
  PORT_HAL_SetMuxMode      (i2c->scl_base, i2c->scl_pin, i2c->scl_mux);
  PORT_HAL_SetOpenDrainCmd (i2c->scl_base, i2c->scl_pin, true);

  /* Reset Run-Time information structure */
  memset (i2c->ctrl, 0x00, sizeof (I2C_CTRL));

  /* Configure DMA */
  if (i2c->dma != NULL) {
    if (i2c->dma->rx != NULL) {
      if (kEDMAInvalidChannel != EDMA_DRV_RequestChannel (kEDMAAnyChannel,
                                                          i2c->dma->req_src,
                                                          i2c->dma->rx)) {
        /* Got a free RX channel */
        i2c->ctrl->flags |= I2C_DMA_RX;

        EDMA_DRV_InstallCallback (i2c->dma->rx, I2Cx_DMA_Callback, (void *)i2c);
      }
    }

    if (i2c->dma->tx != NULL) {
      if (kEDMAInvalidChannel != EDMA_DRV_RequestChannel (kEDMAAnyChannel,
                                                          i2c->dma->req_src,
                                                          i2c->dma->tx)) {
        /* Got a free TX channel */
        i2c->ctrl->flags |= I2C_DMA_TX;

        EDMA_DRV_InstallCallback (i2c->dma->tx, I2Cx_DMA_Callback, (void *)i2c);
      }
    }

    if ((i2c->ctrl->flags & (I2C_DMA_RX | I2C_DMA_TX)) != 0) {
      /* Partially prepare DMA channel config structure */
      i2c->dma->cfg->srcTransferSize    = kEDMATransferSize_1Bytes;
      i2c->dma->cfg->destTransferSize   = kEDMATransferSize_1Bytes;
      i2c->dma->cfg->srcLastAddrAdjust  = 0;
      i2c->dma->cfg->destLastAddrAdjust = 0;
      i2c->dma->cfg->srcModulo          = kEDMAModuloDisable;
      i2c->dma->cfg->destModulo         = kEDMAModuloDisable;
      i2c->dma->cfg->minorLoopCount     = 1;
    }
  }

  i2c->ctrl->cb_event = cb_event;
  i2c->ctrl->flags   |= I2C_INIT;

  return ARM_DRIVER_OK;
}


/**
  \fn          int32_t I2C_Uninitialize (I2C_RESOURCES *i2c)
  \brief       De-initialize I2C Interface.
  \param[in]   i2c  Pointer to I2C resources
  \return      \ref execution_status
*/
static int32_t I2C_Uninitialize (I2C_RESOURCES *i2c) {

  i2c->ctrl->flags = 0U;

  return ARM_DRIVER_OK;
}


/**
  \fn          int32_t ARM_I2C_PowerControl (ARM_POWER_STATE state, I2C_RESOURCES *i2c)
  \brief       Control I2C Interface Power.
  \param[in]   state  Power state
  \param[in]   i2c  Pointer to I2C resources
  \return      \ref execution_status
*/
static int32_t I2C_PowerControl (ARM_POWER_STATE state, I2C_RESOURCES *i2c) {

  switch (state) {
    case ARM_POWER_OFF:
      /* Disable I2C interrupt requests in NVIC */
      NVIC_DisableIRQ (i2c->irq_num);

      I2C_HAL_SetIntCmd (i2c->base, false);

      if (i2c->ctrl->status.busy) {
        /* Transfer in progress */
        if (I2C_HAL_IsMaster (i2c->base)) {
          /* Master generates STOP condition */
          I2C_HAL_SendStop (i2c->base);
        }
      }

      i2c->ctrl->status.busy             = 0U;
      i2c->ctrl->status.mode             = 0U;
      i2c->ctrl->status.direction        = 0U;
      i2c->ctrl->status.general_call     = 0U;
      i2c->ctrl->status.arbitration_lost = 0U;
      i2c->ctrl->status.bus_error        = 0U;

      i2c->ctrl->flags = I2C_INIT;

      /* Disable I2C peripheral */
      I2C_HAL_Disable (i2c->base);

      /* Disable I2C peripheral clock */
      SIM_HAL_DisableClock (SIM, i2c->clock_gate);
      break;

    case ARM_POWER_LOW:
      return ARM_DRIVER_ERROR_UNSUPPORTED;

    case ARM_POWER_FULL:
      if ((i2c->ctrl->flags & I2C_POWER) == 0U) {
        /* Enable I2C peripheral clock */
        SIM_HAL_EnableClock (SIM, i2c->clock_gate);

        /* Enable I2C interrupt requests in NVIC */
        NVIC_EnableIRQ (i2c->irq_num);

        /* Initialize peripheral to known state */
        I2C_HAL_Init (i2c->base);

        /* Enable START/STOP condition detection interrupt */
        I2C_HAL_SetStartStopIntCmd (i2c->base, true);

        /* Enable I2C interrupts */
        I2C_HAL_SetIntCmd (i2c->base, true);

        /* Enable I2C peripheral */
        I2C_HAL_Enable (i2c->base);

        /* Ready for operation */
        i2c->ctrl->flags |= I2C_POWER;
      }
      break;
  }

  return ARM_DRIVER_OK;
}


/**
  \fn          int32_t I2C_MasterTransmit (uint32_t       addr,
                                           const uint8_t *data,
                                           uint32_t       num,
                                           bool           xfer_pending,
                                           I2C_RESOURCES *i2c)
  \brief       Start transmitting data as I2C Master.
  \param[in]   addr          Slave address (7-bit or 10-bit)
  \param[in]   data          Pointer to buffer with data to send to I2C Slave
  \param[in]   num           Number of data bytes to send
  \param[in]   xfer_pending  Transfer operation is pending - Stop condition will not be generated
  \param[in]   i2c           Pointer to I2C resources
  \return      \ref execution_status
*/
static int32_t I2C_MasterTransmit (uint32_t       addr,
                                   const uint8_t *data,
                                   uint32_t       num,
                                   bool           xfer_pending,
                                   I2C_RESOURCES *i2c) {
  edma_software_tcd_t tcd;

  if ((data == NULL) || (num == 0U)) {
    return ARM_DRIVER_ERROR_PARAMETER;
  }

  if (i2c->ctrl->status.busy) {
    return (ARM_DRIVER_ERROR_BUSY);
  }

  i2c->ctrl->status.busy             = 1U;
  i2c->ctrl->status.mode             = 1U;
  i2c->ctrl->status.direction        = 0U;
  i2c->ctrl->status.bus_error        = 0U;
  i2c->ctrl->status.arbitration_lost = 0U;

  i2c->ctrl->xfer.addr  = addr;
  i2c->ctrl->xfer.data  = (uint8_t *)data;
  i2c->ctrl->xfer.num   = num;
  i2c->ctrl->xfer.cnt   = 0U;
  i2c->ctrl->xfer.flags = XFER_FLAG_START;
  i2c->ctrl->xfer.state = XFER_STATE_START;

  if (xfer_pending) {
    i2c->ctrl->xfer.flags |= XFER_FLAG_PENDING;
  }

  if (i2c->ctrl->flags & I2C_DMA_TX) {
    if (num > 1) {
      i2c->ctrl->xfer.flags |= XFER_FLAG_DMA;

      /* Prepare config structure */
      i2c->dma->cfg->srcAddr        = (uint32_t)&data[1];
      i2c->dma->cfg->destAddr       = (uint32_t)&I2C_D_REG(i2c->base);
      i2c->dma->cfg->srcOffset      = 1;
      i2c->dma->cfg->destOffset     = 0;
      i2c->dma->cfg->majorLoopCount = num - 1;

      EDMA_DRV_PrepareDescriptorTransfer (i2c->dma->tx, &tcd, i2c->dma->cfg, true, false);
      
      EDMA_HAL_STCDSetScatterGatherCmd   (&tcd, false);
      EDMA_HAL_STCDSetChannelMajorLink   (&tcd, i2c->dma->tx->channel, false);
      EDMA_HAL_STCDSetHalfCompleteIntCmd (&tcd, false);
      EDMA_HAL_STCDSetBandwidth          (&tcd, kEDMABandwidthStallNone);
      EDMA_HAL_STCDSetDisableDmaRequestAfterTCDDoneCmd (&tcd, true);
      
      EDMA_DRV_PushDescriptorToReg       (i2c->dma->tx, &tcd);
      EDMA_DRV_StartChannel              (i2c->dma->tx);
    }
  }

  /* Generate START condition */
  I2C_HAL_SendStart (i2c->base);

  return ARM_DRIVER_OK;
}


/**
  \fn          int32_t I2C_MasterReceive (uint32_t       addr,
                                          uint8_t       *data,
                                          uint32_t       num,
                                          bool           xfer_pending,
                                          I2C_RESOURCES *i2c)
  \brief       Start receiving data as I2C Master.
  \param[in]   addr          Slave address (7-bit or 10-bit)
  \param[out]  data          Pointer to buffer for data to receive from I2C Slave
  \param[in]   num           Number of data bytes to receive
  \param[in]   xfer_pending  Transfer operation is pending - Stop condition will not be generated
  \param[in]   i2c           Pointer to I2C resources
  \return      \ref execution_status
*/
static int32_t I2C_MasterReceive (uint32_t       addr,
                                  uint8_t       *data,
                                  uint32_t       num,
                                  bool           xfer_pending,
                                  I2C_RESOURCES *i2c) {
  edma_software_tcd_t tcd;

  if ((data == NULL) || (num == 0U)) {
    return ARM_DRIVER_ERROR_PARAMETER;
  }

  if (i2c->ctrl->status.busy) {
    return (ARM_DRIVER_ERROR_BUSY);
  }

  i2c->ctrl->status.busy             = 1U;
  i2c->ctrl->status.mode             = 1U;
  i2c->ctrl->status.direction        = 1U;
  i2c->ctrl->status.bus_error        = 0U;
  i2c->ctrl->status.arbitration_lost = 0U;

  i2c->ctrl->xfer.addr  = addr;
  i2c->ctrl->xfer.data  = data;
  i2c->ctrl->xfer.num   = num;
  i2c->ctrl->xfer.cnt   = 0U;
  i2c->ctrl->xfer.flags = XFER_FLAG_START;
  i2c->ctrl->xfer.state = XFER_STATE_START;

  if (xfer_pending) {
    i2c->ctrl->xfer.flags |= XFER_FLAG_PENDING;
  }

  if (i2c->ctrl->flags & I2C_DMA_RX) {
    if (num > 1) {
      i2c->ctrl->xfer.flags |= XFER_FLAG_DMA;

      /* Prepare config structure */
      i2c->dma->cfg->srcAddr        = (uint32_t)&I2C_D_REG(i2c->base);
      i2c->dma->cfg->destAddr       = (uint32_t)data;
      i2c->dma->cfg->srcOffset      = 0;
      i2c->dma->cfg->destOffset     = 1;
      i2c->dma->cfg->majorLoopCount = num - 1;

      EDMA_DRV_PrepareDescriptorTransfer (i2c->dma->rx, &tcd, i2c->dma->cfg, true, false);

      EDMA_HAL_STCDSetScatterGatherCmd   (&tcd, false);
      EDMA_HAL_STCDSetChannelMajorLink   (&tcd, i2c->dma->rx->channel, false);
      EDMA_HAL_STCDSetHalfCompleteIntCmd (&tcd, false);
      EDMA_HAL_STCDSetBandwidth          (&tcd, kEDMABandwidthStallNone);
      EDMA_HAL_STCDSetDisableDmaRequestAfterTCDDoneCmd (&tcd, true);

      EDMA_DRV_PushDescriptorToReg       (i2c->dma->rx, &tcd);
      EDMA_DRV_StartChannel              (i2c->dma->rx);
    }
  }

  /* Generate START condition */
  I2C_HAL_SendStart (i2c->base);

  return ARM_DRIVER_OK;
}


/**
  \fn          int32_t I2C_SlaveTransmit (const uint8_t *data, uint32_t num, I2C_RESOURCES *i2c)
  \brief       Start transmitting data as I2C Slave.
  \param[in]   data          Pointer to buffer with data to send to I2C Master
  \param[in]   num           Number of data bytes to send
  \param[in]   i2c           Pointer to I2C resources
  \return      \ref execution_status
*/
static int32_t I2C_SlaveTransmit (const uint8_t *data, uint32_t num, I2C_RESOURCES *i2c) {

  if ((data == NULL) || (num == 0U)) {
    return ARM_DRIVER_ERROR_PARAMETER;
  }

  if (i2c->ctrl->status.busy) {
    return (ARM_DRIVER_ERROR_BUSY);
  }

  i2c->ctrl->status.bus_error    = 0U;
  i2c->ctrl->status.general_call = 0U;

  i2c->ctrl->xfer.data  = (uint8_t *)data;
  i2c->ctrl->xfer.num   = num;
  i2c->ctrl->xfer.cnt   = 0U;
  i2c->ctrl->xfer.flags = 0U;
  i2c->ctrl->xfer.state = XFER_STATE_IDLE;

  return ARM_DRIVER_OK;
}


/**
  \fn          int32_t I2C_SlaveReceive (uint8_t *data, uint32_t num, I2C_RESOURCES *i2c)
  \brief       Start receiving data as I2C Slave.
  \param[out]  data          Pointer to buffer for data to receive from I2C Master
  \param[in]   num           Number of data bytes to receive
  \param[in]   i2c           Pointer to I2C resources
  \return      \ref execution_status
*/
static int32_t I2C_SlaveReceive (uint8_t *data, uint32_t num, I2C_RESOURCES *i2c) {

  if ((data == NULL) || (num == 0U)) {
    return ARM_DRIVER_ERROR_PARAMETER;
  }

  if (i2c->ctrl->status.busy) {
    return (ARM_DRIVER_ERROR_BUSY);
  }

  i2c->ctrl->status.bus_error    = 0U;
  i2c->ctrl->status.general_call = 0U;

  i2c->ctrl->xfer.data  = data;
  i2c->ctrl->xfer.num   = num;
  i2c->ctrl->xfer.cnt   = 0U;
  i2c->ctrl->xfer.flags = 0U;
  i2c->ctrl->xfer.state = XFER_STATE_IDLE;

  return ARM_DRIVER_OK;
}


/**
  \fn          int32_t I2C_GetDataCount (void)
  \brief       Get transferred data count.
  \return      number of data bytes transferred; -1 when Slave is not addressed by Master
*/
static int32_t I2C_GetDataCount (I2C_RESOURCES *i2c) {
  int32_t cnt;

  if (i2c->ctrl->xfer.flags != 0) {
    /* Transfer in progress or completed */
    cnt = i2c->ctrl->xfer.cnt;
  }
  else {
    /* Slave not yet addressed */
    cnt = -1;
  }

  return (cnt);
}


/**
  \fn          int32_t I2C_Control (uint32_t control, uint32_t arg, I2C_RESOURCES *i2c)
  \brief       Control I2C Interface.
  \param[in]   control  operation
  \param[in]   arg      argument of operation (optional)
  \param[in]   i2c      pointer to I2C resources
  \return      \ref execution_status
*/
static int32_t I2C_Control (uint32_t control, uint32_t arg, I2C_RESOURCES *i2c) {
  uint32_t   freq, i, j;
  GPIO_Type *gpio_sda;
  GPIO_Type *gpio_scl;

  if ((i2c->ctrl->flags & I2C_POWER) == 0U) {
    /* I2C not powered */
    return ARM_DRIVER_ERROR;
  }

  switch (control) {
    case ARM_I2C_OWN_ADDRESS:
      if (arg & ARM_I2C_ADDRESS_10BIT) {
        I2C_HAL_SetAddress10bit (i2c->base, (uint16_t)arg);
      }
      else {
        I2C_HAL_SetAddress7bit (i2c->base, (uint8_t)arg);
      }
      break;

    case ARM_I2C_BUS_SPEED:
        if      (i2c->base == I2C0) { i = I2C0_IDX; }
        else if (i2c->base == I2C1) { i = I2C1_IDX; }
        else if (i2c->base == I2C2) { i = I2C2_IDX; }
        else { return ARM_DRIVER_ERROR; }

        freq = CLOCK_SYS_GetI2cFreq (i);

        switch (arg) {
          case ARM_I2C_BUS_SPEED_STANDARD:
            /* Clock = 100kHz,  Rise Time = 1000ns */
            I2C_HAL_SetBaudRate (i2c->base, freq, 100, NULL);
            break;
          case ARM_I2C_BUS_SPEED_FAST:
            /* Clock = 400kHz,  Rise Time = 300ns */
            I2C_HAL_SetBaudRate (i2c->base, freq, 400, NULL);
            break;
        default:
          return ARM_DRIVER_ERROR_UNSUPPORTED;
      }
      break;

    case ARM_I2C_BUS_CLEAR:
      SystemCoreClockUpdate();

      /* Frequency is set to approx. standard speed (100kHz) */
      freq = SystemCoreClock / 100000 / 4;

      /* Configure SDA and SCL pins as GPIO pins */
      PORT_HAL_SetMuxMode (i2c->sda_base, i2c->sda_pin, kPortMuxAsGpio);
      PORT_HAL_SetMuxMode (i2c->scl_base, i2c->scl_pin, kPortMuxAsGpio);

      gpio_sda = (GPIO_Type *)(PTx_Base ((uint32_t)i2c->sda_base));
      gpio_scl = (GPIO_Type *)(PTx_Base ((uint32_t)i2c->scl_base));

      GPIO_HAL_SetPinDir (gpio_sda, i2c->sda_pin, kGpioDigitalInput);
      GPIO_HAL_SetPinDir (gpio_scl, i2c->scl_pin, kGpioDigitalOutput);

      for (i = 0U; i < 9U; i++) {
        if (GPIO_HAL_ReadPinInput (gpio_sda, i2c->sda_pin) == 1U) {
          /* Break if slave released SDA line */
          break;
        }
        /* Clock high */
        GPIO_HAL_SetPinOutput (gpio_scl, i2c->scl_pin);
        for (j = 0; j < freq; j++);

        /* Clock low */
        GPIO_HAL_ClearPinOutput (gpio_scl, i2c->scl_pin);
        for (j = 0; j < freq; j++);
      }

      /* Check SDA state */
      i = GPIO_HAL_ReadPinInput (gpio_sda, i2c->sda_pin);

      /* Generate STOP condition - SDA goes high while SCL is high */
      GPIO_HAL_SetPinDir (gpio_sda, i2c->sda_pin, kGpioDigitalOutput);
      GPIO_HAL_ClearPinOutput (gpio_sda, i2c->sda_pin);
      for (j = 0; j < freq; j++);
      GPIO_HAL_SetPinOutput   (gpio_scl, i2c->scl_pin);
      for (j = 0; j < freq; j++);
      GPIO_HAL_SetPinOutput   (gpio_sda, i2c->sda_pin);

      /* Configure SDA and SCL pins as I2C peripheral pins */
      PORT_HAL_SetMuxMode (i2c->sda_base, i2c->sda_pin, i2c->sda_mux);
      PORT_HAL_SetMuxMode (i2c->scl_base, i2c->scl_pin, i2c->scl_mux);

      /* Send event */
      if (i2c->ctrl->cb_event) {
        i2c->ctrl->cb_event (ARM_I2C_EVENT_BUS_CLEAR);
      }
      return (i == 1U) ? ARM_DRIVER_OK : ARM_DRIVER_ERROR;

    case ARM_I2C_ABORT_TRANSFER:
      if (i2c->ctrl->status.busy) {
        /* Disable I2C interrupts */
        I2C_HAL_SetIntCmd (i2c->base, false);

        if (I2C_HAL_IsMaster (i2c->base)) {
          /* Master generates STOP condition */
          I2C_HAL_SendStop (i2c->base);
        }
        else {
          /* Slave releases the bus after current byte transfer */
        }
      }
      break;
  }

  return ARM_DRIVER_OK;
}


/**
  \fn          ARM_I2C_STATUS I2C_GetStatus (I2C_RESOURCES *i2c)
  \brief       Get I2C status.
  \param[in]   i2c      pointer to I2C resources
  \return      I2C status \ref ARM_I2C_STATUS
*/
static ARM_I2C_STATUS I2C_GetStatus (I2C_RESOURCES *i2c) {
  return (i2c->ctrl->status);
}


/**
  \fn          void I2Cx_IRQHandler (I2C_RESOURCES *i2c)
  \brief       I2C Interrupt handler.
  \param[in]   i2c  Pointer to I2C resources
*/
static void I2Cx_IRQHandler (I2C_RESOURCES *i2c) {
  I2C_TRANSFER_INFO *tr = &i2c->ctrl->xfer;
  uint32_t event;
  uint8_t  data;
  bool     tx, nack;

  I2C_HAL_ClearInt (i2c->base);

  if (I2C_HAL_GetStartFlag (i2c->base)) {
    /* Start condition detected */
    I2C_HAL_ClearStartFlag (i2c->base);
    I2C_HAL_ClearInt       (i2c->base);

    switch (tr->state) {
      case XFER_STATE_IDLE:
        if (tr->flags & XFER_FLAG_DONE) {
          /* Slave transfer done, this is repeated START */
          event = ARM_I2C_EVENT_TRANSFER_DONE;

          if (tr->cnt < tr->num) {
            event |= ARM_I2C_EVENT_TRANSFER_INCOMPLETE;
          }

          if (i2c->ctrl->cb_event) {
            i2c->ctrl->cb_event (event);
          }
        }
        break;

      case XFER_STATE_START:
        tr->state = XFER_STATE_ADDR;

        if (tr->addr & ARM_I2C_ADDRESS_10BIT) {
          /* 10-bit addressing, send hi byte */
          data = (uint8_t)(0xF0U | ((tr->addr >> 7) & 0x06U));
        }
        else {
          /* 7-bit addressing */
          tr->flags |= XFER_FLAG_ADDR_SENT;

          data  = (uint8_t)(tr->addr << 1);
          data |= (uint8_t)i2c->ctrl->status.direction;
        }
        break;

      case XFER_STATE_RSTART:
        tr->state  = XFER_STATE_ADDR;
        tr->flags |= XFER_FLAG_ADDR_SENT;

        /* 10-bit addressing, send hi byte with R bit set */
        data = (uint8_t)(0xF0U | ((tr->addr >> 7) | 1));
        break;

      default:
        /* Ignore start condition */
        break;
    }
    if (tr->state == XFER_STATE_ADDR) {
      /* Set Tx direction */
      I2C_HAL_SetDirMode(i2c->base, kI2CSend);
      /* Send address byte */
      I2C_HAL_WriteByte (i2c->base, data);
    }
  }
  else if (I2C_HAL_GetStopFlag (i2c->base)) {
    /* Stop condition detected */
    I2C_HAL_ClearStopFlag (i2c->base);
    I2C_HAL_ClearInt      (i2c->base);

    event = ARM_I2C_EVENT_TRANSFER_DONE;

    if (tr->cnt < tr->num) {
      event |= ARM_I2C_EVENT_TRANSFER_INCOMPLETE;
    }
    tr->data = NULL;

    switch (tr->state) {
      case XFER_STATE_IDLE:
      case XFER_STATE_STOP:
        break;
      
      case XFER_STATE_ADDR_NACK:
        /* Slave address not acked */
        event |= ARM_I2C_EVENT_ADDRESS_NACK;
        break;

      default:
        /* Unexpected stop condition */
        i2c->ctrl->status.bus_error = 1U;

        event |= ARM_I2C_EVENT_BUS_ERROR;
    }

    /* Update status */
    i2c->ctrl->status.busy = 0U;
    i2c->ctrl->status.mode = 0U;

    /* Send event */
    if (i2c->ctrl->cb_event) {
      i2c->ctrl->cb_event (event);
    }
  }
  else if (I2C_HAL_GetStatusFlag (i2c->base, kI2CAddressAsSlave)) {
    /* Addressed as a slave */
    tx = I2C_HAL_GetStatusFlag(i2c->base, kI2CSlaveTransmit);

    if (tx) {
      /* Slave addressed as a transmitter */
      event = ARM_I2C_EVENT_SLAVE_TRANSMIT;

      I2C_HAL_SetDirMode (i2c->base, kI2CSend);
    }
    else {
      /* Slave addressed as a receiver */
      event = ARM_I2C_EVENT_SLAVE_RECEIVE;

      I2C_HAL_SetDirMode (i2c->base, kI2CReceive);
    }

    /* Check if this is a General Call address */
    if ((I2C_HAL_ReadByte(i2c->base) >> 1) == 0x00) {
      event |= ARM_I2C_EVENT_GENERAL_CALL;

      i2c->ctrl->status.general_call = 1U;
    }
    else {
      i2c->ctrl->status.general_call = 0U;
    }

    if ((tr->data == NULL) ||
        ((tx == true)  && (i2c->ctrl->status.direction == 0U)) ||
        ((tx == false) && (i2c->ctrl->status.direction == 1U))) {

      if (i2c->ctrl->cb_event) {
        i2c->ctrl->cb_event (event);
      }
    }
    tr->flags |= XFER_FLAG_ADDR_DONE;

    if (tx) {
      tr->state = XFER_STATE_SLAVE_TX;
      
      if (tr->data != NULL) {
        /* Send first byte */
        I2C_HAL_WriteByte (i2c->base, tr->data[0]);
        tr->cnt = 1;
      }
    }
    else {
      tr->state = XFER_STATE_SLAVE_RX;

      if (tr->data == NULL) {
        /* No data, disable ACK */
        I2C_HAL_SendNak (i2c->base);
      }
    }
  }
  else if (I2C_HAL_GetStatusFlag (i2c->base, kI2CArbitrationLost)) {
    /* Arbitration lost */
    I2C_HAL_ClearArbitrationLost (i2c->base);

    event = ARM_I2C_EVENT_TRANSFER_DONE | ARM_I2C_EVENT_ARBITRATION_LOST;

    /* Switch to slave mode */
    i2c->ctrl->status.busy             = 0U;
    i2c->ctrl->status.mode             = 0U;
    i2c->ctrl->status.arbitration_lost = 1U;

    i2c->ctrl->xfer.data  = NULL;
    i2c->ctrl->xfer.flags = 0U;
  }
  else {
    if (I2C_HAL_GetStatusFlag(i2c->base, kI2CTransferComplete)) {
      /* Byte Transfer Complete */
      nack = I2C_HAL_GetStatusFlag (i2c->base, kI2CReceivedNak);

      switch (tr->state) {
        case XFER_STATE_ADDR:
          /* Addressing phase */
          if (nack) {
            /* Slave address not acknowledged */
            tr->state = XFER_STATE_ADDR_NACK;
            I2C_HAL_SendStop (i2c->base);
          }
          else {
            if ((tr->flags & XFER_FLAG_ADDR_SENT) == 0U) {
              /* Addressing phase in progress */
              if (tr->flags & XFER_FLAG_RSTART) {
                tr->state = XFER_STATE_RSTART;
                /* Generate repeated start condition */
                I2C_HAL_SendStart (i2c->base);
              }
              else {
                /* 10-bit addressing mode, send low address byte */
                I2C_HAL_WriteByte (i2c->base, tr->addr);
              
                if (i2c->ctrl->status.direction) {
                  /* Read from slave */
                  tr->flags |= XFER_FLAG_RSTART;
                }
              }
            }
            else {
              /* Address sent and acknowledged, prepare for data transfer */
              if (tr->flags & XFER_FLAG_DMA) {
                /* Enable DMA transfer */
                I2C_HAL_SetDmaCmd (i2c->base, true);
              }

              if (i2c->ctrl->status.direction) {
                /* Master receiver */
                tr->state = XFER_STATE_MASTER_RX;

                I2C_HAL_SetDirMode (i2c->base, kI2CReceive);

                if (tr->num == 1U) {
                  I2C_HAL_SendNak (i2c->base);
                }
                else {
                  I2C_HAL_SendAck (i2c->base);
                }

                /* Dummy read to trigger TCF */
                I2C_HAL_ReadByte (i2c->base);
              }
              else {
                /* Master transmitter */
                tr->state = XFER_STATE_MASTER_TX;

                I2C_HAL_SetDirMode (i2c->base, kI2CSend);

                /* Send first byte */
                I2C_HAL_WriteByte (i2c->base, tr->data[0]);
                tr->cnt = 1;
              }
            }
          }
          break;
        
        case XFER_STATE_MASTER_RX:
          if ((tr->flags & XFER_FLAG_DMA) == 0U) {
            tr->cnt++;
          }

          switch (tr->num - tr->cnt) {
            case 0U: /* All data received, send stop */
              if (tr->flags & XFER_FLAG_PENDING) {
                /* Wait for repeated start */
                tr->state = XFER_STATE_START;

                i2c->ctrl->status.busy = 0U;
              }
              else {
                /* Send stop and end the transfer */
                tr->state = XFER_STATE_STOP;

                I2C_HAL_SendStop (i2c->base);
              }
              break;

            case 1U: /* NAK before last byte */
              I2C_HAL_SendNak (i2c->base);
              break;

            default: /* ACK current byte */
              I2C_HAL_SendAck (i2c->base);
              break;
          }
          /* Read received byte */
          tr->data[tr->cnt-1] = I2C_HAL_ReadByte(i2c->base);
          break;

        case XFER_STATE_MASTER_TX:
          if (!nack && (tr->cnt < tr->num)) {
            I2C_HAL_WriteByte (i2c->base, tr->data[tr->cnt]);

            tr->cnt++;
          }
          else {
            /* Data not acknowledged or all transmitted, send stop */
            if (tr->flags & XFER_FLAG_PENDING) {
              /* Wait for repeated start */
              tr->state = XFER_STATE_START;
              
              i2c->ctrl->status.busy = 0U;
            }
            else {
              /* Send stop and end the transfer */
              tr->state = XFER_STATE_STOP;

              I2C_HAL_SendStop (i2c->base);
            }
          }
          break;

        case XFER_STATE_SLAVE_RX:
          data = I2C_HAL_ReadByte (i2c->base);

          if (tr->data) {
            tr->data[tr->cnt] = data;
            tr->cnt++;

            switch (tr->num - tr->cnt) {
              case 0U:
                /* All data received */
                tr->data = NULL;

                /* NACK sent, expect START or STOP */
                tr->flags |= XFER_FLAG_DONE;
                tr->state  = XFER_STATE_IDLE;
                break;

              case 1U:
                /* Disable ACK, only one byte to receive */
                I2C_HAL_SendNak (i2c->base);
                break;

              default:
                break;
            }
          }
          break;

        case XFER_STATE_SLAVE_TX:
          if (nack) {
            /* NACK from master, expect START or STOP */
            tr->flags |= XFER_FLAG_DONE;
            tr->state  = XFER_STATE_IDLE;
          }
          else {
            if (tr->data) {
              I2C_HAL_WriteByte (i2c->base, tr->data[tr->cnt]);
              tr->cnt++;

              if (tr->cnt == tr->num) {
                /* All data transmitted */
                tr->data = NULL;
              }
            }
            else {
              /* Master requests more data */
              I2C_HAL_WriteByte (i2c->base, 0xFF);
            }
          }
          break;
      }
    }
  }
}


/**
  \fn          void I2Cx_DMA_Callback (void *par, edma_chn_status_t status)
  \brief       I2C DMA callback handler.
  \param[in]   par (void)Pointer to I2C resources
*/
static void I2Cx_DMA_Callback (void *par, edma_chn_status_t status) {
  I2C_RESOURCES *i2c = (I2C_RESOURCES *)par;

  I2C_HAL_SetDmaCmd (i2c->base, false);

  switch (status) {
    case kEDMAChnNormal:
      i2c->ctrl->xfer.cnt = i2c->ctrl->xfer.num;

      if (i2c->ctrl->status.direction) {
        /* Receiver */
        I2C_HAL_SendNak (i2c->base);

        /* Read received byte */
        i2c->ctrl->xfer.data[i2c->ctrl->xfer.cnt] = I2C_HAL_ReadByte(i2c->base);
      }
      break;
    case kEDMAChnIdle:
      break;

    case kEDMAChnError:
    default:
      DMA_CERR = 1 << i2c->dma->rx->channel;
      break;
  }
}


#if defined(RTE_I2C0)
/* I2C0 Driver wrapper functions */
static int32_t I2C0_Initialize (ARM_I2C_SignalEvent_t cb_event) {
  return I2C_Initialize(cb_event, &I2C0_Resources);
}
static int32_t I2C0_Uninitialize (void) {
  return I2C_Uninitialize(&I2C0_Resources);
}
static int32_t I2C0_PowerControl (ARM_POWER_STATE state) {
  return I2C_PowerControl(state, &I2C0_Resources);
}
static int32_t I2C0_MasterTransmit (uint32_t addr, const uint8_t *data, uint32_t num, bool xfer_pending) {
  return I2C_MasterTransmit(addr, data, num, xfer_pending, &I2C0_Resources);
}
static int32_t I2C0_MasterReceive (uint32_t addr, uint8_t *data, uint32_t num, bool xfer_pending) {
  return I2C_MasterReceive(addr, data, num, xfer_pending, &I2C0_Resources);
}
static int32_t I2C0_SlaveTransmit (const uint8_t *data, uint32_t num) {
  return I2C_SlaveTransmit(data, num, &I2C0_Resources);
}
static int32_t I2C0_SlaveReceive (uint8_t *data, uint32_t num) {
  return I2C_SlaveReceive(data, num, &I2C0_Resources);
}
static int32_t I2C0_GetDataCount (void) {
  return I2C_GetDataCount(&I2C0_Resources);
}
static int32_t I2C0_Control (uint32_t control, uint32_t arg) {
  return I2C_Control(control, arg, &I2C0_Resources);
}
static ARM_I2C_STATUS I2C0_GetStatus (void) {
  return I2C_GetStatus(&I2C0_Resources);
}
void I2C0_IRQHandler (void) {
  I2Cx_IRQHandler(&I2C0_Resources);
}


/* I2C0 Driver Control Block */
ARM_DRIVER_I2C Driver_I2C0 = {
  I2CX_GetVersion,
  I2CX_GetCapabilities,
  I2C0_Initialize,
  I2C0_Uninitialize,
  I2C0_PowerControl,
  I2C0_MasterTransmit,
  I2C0_MasterReceive,
  I2C0_SlaveTransmit,
  I2C0_SlaveReceive,
  I2C0_GetDataCount,
  I2C0_Control,
  I2C0_GetStatus
};
#endif


#if defined(RTE_I2C1)
/* I2C1 Driver wrapper functions */
static int32_t I2C1_Initialize (ARM_I2C_SignalEvent_t cb_event) {
  return I2C_Initialize(cb_event, &I2C1_Resources);
}
static int32_t I2C1_Uninitialize (void) {
  return I2C_Uninitialize(&I2C1_Resources);
}
static int32_t I2C1_PowerControl (ARM_POWER_STATE state) {
  return I2C_PowerControl(state, &I2C1_Resources);
}
static int32_t I2C1_MasterTransmit (uint32_t addr, const uint8_t *data, uint32_t num, bool xfer_pending) {
  return I2C_MasterTransmit(addr, data, num, xfer_pending, &I2C1_Resources);
}
static int32_t I2C1_MasterReceive (uint32_t addr, uint8_t *data, uint32_t num, bool xfer_pending) {
  return I2C_MasterReceive(addr, data, num, xfer_pending, &I2C1_Resources);
}
static int32_t I2C1_SlaveTransmit (const uint8_t *data, uint32_t num) {
  return I2C_SlaveTransmit(data, num, &I2C1_Resources);
}
static int32_t I2C1_SlaveReceive (uint8_t *data, uint32_t num) {
  return I2C_SlaveReceive(data, num, &I2C1_Resources);
}
static int32_t I2C1_GetDataCount (void) {
  return I2C_GetDataCount(&I2C1_Resources);
}
static int32_t I2C1_Control (uint32_t control, uint32_t arg) {
  return I2C_Control(control, arg, &I2C1_Resources);
}
static ARM_I2C_STATUS I2C1_GetStatus (void) {
  return I2C_GetStatus(&I2C1_Resources);
}
void I2C1_IRQHandler (void) {
  I2Cx_IRQHandler(&I2C1_Resources);
}


/* I2C1 Driver Control Block */
ARM_DRIVER_I2C Driver_I2C1 = {
  I2CX_GetVersion,
  I2CX_GetCapabilities,
  I2C1_Initialize,
  I2C1_Uninitialize,
  I2C1_PowerControl,
  I2C1_MasterTransmit,
  I2C1_MasterReceive,
  I2C1_SlaveTransmit,
  I2C1_SlaveReceive,
  I2C1_GetDataCount,
  I2C1_Control,
  I2C1_GetStatus
};
#endif


#if defined(RTE_I2C2)
/* I2C2 Driver wrapper functions */
static int32_t I2C2_Initialize (ARM_I2C_SignalEvent_t cb_event) {
  return I2C_Initialize(cb_event, &I2C2_Resources);
}
static int32_t I2C2_Uninitialize (void) {
  return I2C_Uninitialize(&I2C2_Resources);
}
static int32_t I2C2_PowerControl (ARM_POWER_STATE state) {
  return I2C_PowerControl(state, &I2C2_Resources);
}
static int32_t I2C2_MasterTransmit (uint32_t addr, const uint8_t *data, uint32_t num, bool xfer_pending) {
  return I2C_MasterTransmit(addr, data, num, xfer_pending, &I2C2_Resources);
}
static int32_t I2C2_MasterReceive (uint32_t addr, uint8_t *data, uint32_t num, bool xfer_pending) {
  return I2C_MasterReceive(addr, data, num, xfer_pending, &I2C2_Resources);
}
static int32_t I2C2_SlaveTransmit (const uint8_t *data, uint32_t num) {
  return I2C_SlaveTransmit(data, num, &I2C2_Resources);
}
static int32_t I2C2_SlaveReceive (uint8_t *data, uint32_t num) {
  return I2C_SlaveReceive(data, num, &I2C2_Resources);
}
static int32_t I2C2_GetDataCount (void) {
  return I2C_GetDataCount(&I2C2_Resources);
}
static int32_t I2C2_Control (uint32_t control, uint32_t arg) {
  return I2C_Control(control, arg, &I2C2_Resources);
}
static ARM_I2C_STATUS I2C2_GetStatus (void) {
  return I2C_GetStatus(&I2C2_Resources);
}
void I2C2_IRQHandler (void) {
  I2Cx_IRQHandler(&I2C2_Resources);
}


/* I2C2 Driver Control Block */
ARM_DRIVER_I2C Driver_I2C2 = {
  I2CX_GetVersion,
  I2CX_GetCapabilities,
  I2C2_Initialize,
  I2C2_Uninitialize,
  I2C2_PowerControl,
  I2C2_MasterTransmit,
  I2C2_MasterReceive,
  I2C2_SlaveTransmit,
  I2C2_SlaveReceive,
  I2C2_GetDataCount,
  I2C2_Control,
  I2C2_GetStatus
};
#endif
