/* -----------------------------------------------------------------------------
 * Copyright (c) 2013-2015 ARM Ltd.
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from
 * the use of this software. Permission is granted to anyone to use this
 * software for any purpose, including commercial applications, and to alter
 * it and redistribute it freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software in
 *    a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 *
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 3. This notice may not be removed or altered from any source distribution.
 *
 *
 * $Date:        05. June 2015
 * $Revision:    V1.2
 *
 * Driver:       Driver_MCI0
 * Configured:   via RTE_Device.h configuration file
 * Project:      MCI Driver for Freescale MK64F
 * --------------------------------------------------------------------------
 * Use the following configuration settings in the middleware component
 * to connect to this driver.
 *
 *   Configuration Setting                 Value
 *   ---------------------                 -----
 *   Connect to hardware via Driver_MCI# = 0
 * -------------------------------------------------------------------------- */

/* History:
 *  Version 1.2
 *    Updated for KSDK 1.2.0
 *    Manual handling of abort command CMD12 implemented
 *    Driver relies on kGpioSdhc0Cd and kGpioSdhc0Wp pins for handling CD and WP
 *  Version 1.1
 *    Updated initialization, uninitialization and power procedures
 *  Version 1.0
 *    Initial release
 */

#include "MCI_MK64F.h"

#define ARM_MCI_DRV_VERSION ARM_DRIVER_VERSION_MAJOR_MINOR(1,2)  /* driver version */

static MCI_CTRL MCI;

/* IRQ Handler prototype */
void SDHC_IRQHandler (void);


/* Driver Version */
static const ARM_DRIVER_VERSION DriverVersion = {
  ARM_MCI_API_VERSION,
  ARM_MCI_DRV_VERSION
};

/* Driver Capabilities */
static const ARM_MCI_CAPABILITIES DriverCapabilities = {
  MCI_CD_PIN,                                     /* cd_state          */
  0U,                                             /* cd_event          */
  MCI_WP_PIN,                                     /* wp_state          */
  0U,                                             /* vdd               */
  0U,                                             /* vdd_1v8           */
  0U,                                             /* vccq              */
  0U,                                             /* vccq_1v8          */
  0U,                                             /* vccq_1v2          */
  MCI_BUS_WIDTH_4,                                /* data_width_4      */
  MCI_BUS_WIDTH_8,                                /* data_width_8      */
  0U,                                             /* data_width_4_ddr  */
  0U,                                             /* data_width_8_ddr  */
  1U,                                             /* high_speed        */
  0U,                                             /* uhs_signaling     */
  0U,                                             /* uhs_tuning        */
  0U,                                             /* uhs_sdr50         */
  0U,                                             /* uhs_sdr104        */
  0U,                                             /* uhs_ddr50         */
  0U,                                             /* uhs_driver_type_a */
  0U,                                             /* uhs_driver_type_c */
  0U,                                             /* uhs_driver_type_d */
  1U,                                             /* sdio_interrupt    */
  0U,                                             /* read_wait         */
  0U,                                             /* suspend_resume    */
  0U,                                             /* mmc_interrupt     */
  0U,                                             /* mmc_boot          */
  0U,                                             /* rst_n             */
  0U,                                             /* ccs               */
  0U                                              /* ccs_timeout       */
};


/**
  \fn          ARM_DRV_VERSION GetVersion (void)
  \brief       Get driver version.
  \return      \ref ARM_DRV_VERSION
*/
static ARM_DRIVER_VERSION GetVersion (void) {
  return DriverVersion;
}


/**
  \fn          ARM_MCI_CAPABILITIES MCI_GetCapabilities (void)
  \brief       Get driver capabilities.
  \return      \ref ARM_MCI_CAPABILITIES
*/
static ARM_MCI_CAPABILITIES GetCapabilities (void) {
  return DriverCapabilities;
}


/**
  \fn            int32_t Initialize (ARM_MCI_SignalEvent_t cb_event)
  \brief         Initialize the Memory Card Interface
  \param[in]     cb_event  Pointer to \ref ARM_MCI_SignalEvent
  \return        \ref execution_status
*/
static int32_t Initialize (ARM_MCI_SignalEvent_t cb_event) {

  if (MCI.flags & MCI_INIT)  { return ARM_DRIVER_OK; }

  MCI.cb_event = cb_event;

  /* Clear status */
  MCI.status.command_active   = 0U;
  MCI.status.command_timeout  = 0U;
  MCI.status.command_error    = 0U;
  MCI.status.transfer_active  = 0U;
  MCI.status.transfer_timeout = 0U;
  MCI.status.transfer_error   = 0U;
  MCI.status.sdio_interrupt   = 0U;
  MCI.status.ccs              = 0U;

  MCI.flags = MCI_INIT;

  return ARM_DRIVER_OK;
}


/**
  \fn            int32_t Uninitialize (void)
  \brief         De-initialize Memory Card Interface.
  \return        \ref execution_status
*/
static int32_t Uninitialize (void) {

  MCI.flags = 0U;

  return ARM_DRIVER_OK;
}


/**
  \fn            int32_t PowerControl (ARM_POWER_STATE state)
  \brief         Control Memory Card Interface Power.
  \param[in]     state   Power state \ref ARM_POWER_STATE
  \return        \ref execution_status
*/
static int32_t PowerControl (ARM_POWER_STATE state) {
  sdhc_hal_config_t cfg;

  switch (state) {
    case ARM_POWER_OFF:
      /* Disable SDHC interrupt in NVIC */
      NVIC_DisableIRQ(SDHC_IRQn);

      /* Clear flags */
      MCI.flags = MCI_POWER;

      /* Clear status */
      MCI.status.command_active   = 0U;
      MCI.status.command_timeout  = 0U;
      MCI.status.command_error    = 0U;
      MCI.status.transfer_active  = 0U;
      MCI.status.transfer_timeout = 0U;
      MCI.status.transfer_error   = 0U;
      MCI.status.sdio_interrupt   = 0U;
      MCI.status.ccs              = 0U;

      /* Disable peripheral interrupts */
      SDHC_HAL_SetIntState(SDHC, false, SDHC_IRQ_EN_MASK);

      /* Reset/Dereset SDHC peripheral */
      SDHC_HAL_Reset (SDHC, SDHC_SYSCTL_RSTA_MASK, MCI_RESET_TIMEOUT);

      /* Disable SDHC peripheral clock */
      SIM_HAL_DisableClock (SIM, kSimClockGateSdhc0);
      break;

    case ARM_POWER_FULL:
      if ((MCI.flags & MCI_POWER) == 0) {
        /* Clear response variable */
        MCI.response = NULL;

        /* Enable SDHC peripheral clock */
        CLOCK_SYS_EnableSdhcClock (0);

        /* Reset/Dereset SDHC peripheral */
        SDHC_HAL_Reset (SDHC, SDHC_SYSCTL_RSTA_MASK, MCI_RESET_TIMEOUT);

        /* Configure peripheral */
        cfg.ledState              = kSdhcHalLedOff;
        cfg.endianMode            = kSdhcHalEndianLittle;
        cfg.dmaMode               = kSdhcHalDmaSimple;
        cfg.writeWatermarkLevel   = 128;
        cfg.readWatermarkLevel    = 128;
        cfg.enFlags               = SDHC_HAL_EN_D3CD_FLAG;
        cfg.bootParams.ackTimeout = 0;
        cfg.bootParams.mode       = kSdhcHalMmcbootNormal;
        cfg.bootParams.blockCount = 0;
        cfg.bootParams.enFlags    = 0;

        SDHC_HAL_Config (SDHC, &cfg);

        /* Set data timeout to maximum */
        SDHC_WR_SYSCTL_DTOCV (SDHC, 0xE);

        /* Clear all interrupt requests */
        SDHC_HAL_ClearIntFlags(SDHC, SDHC_IRQ_CLR_MASK);

        /* Enable SDHC peripheral interrupts */
        SDHC_HAL_SetIntState  (SDHC, true, SDHC_IRQ_EN_MASK);
        SDHC_HAL_SetIntSignal (SDHC, true, SDHC_IRQ_EN_MASK);

        /* Clear pending and enable SDHC interrupts in NVIC */
        NVIC_ClearPendingIRQ (SDHC_IRQn);
        NVIC_EnableIRQ       (SDHC_IRQn);

        MCI.flags |= MCI_POWER;
      }
      break;

    case ARM_POWER_LOW:
    default:
      return ARM_DRIVER_ERROR_UNSUPPORTED;
  }
  return ARM_DRIVER_OK;
}


/**
  \fn            int32_t CardPower (uint32_t voltage)
  \brief         Set Memory Card supply voltage.
  \param[in]     voltage  Memory Card supply voltage
  \return        \ref execution_status
*/
static int32_t CardPower (uint32_t voltage) {

  if ((MCI.flags & MCI_POWER) == 0U) { return ARM_DRIVER_ERROR; }
  return ARM_DRIVER_ERROR_UNSUPPORTED;
}


/**
  \fn            int32_t ReadCD (void)
  \brief         Read Card Detect (CD) state.
  \return        1:card detected, 0:card not detected, or error
*/
static int32_t ReadCD (void) {

#if (MCI_CD_PIN != 0)
  if (MCI.flags & MCI_POWER) {
    if (GPIO_DRV_ReadPinInput(kGpioSdhc0Cd) == MCI_CD_PIN_ACTIVE) {
      return (1);
    }
  }
#endif
  return (0);
}


/**
  \fn            int32_t ReadWP (void)
  \brief         Read Write Protect (WP) state.
  \return        1:write protected, 0:not write protected, or error
*/
static int32_t ReadWP (void) {

#if (MCI_WP_PIN != 0)
  if (MCI.flags & MCI_POWER) {
    if (GPIO_DRV_ReadPinInput(kGpioSdhc0Wp) == MCI_CD_PIN_ACTIVE) {
      return (1);
    }
  }
#endif
  return (0);
}


/**
  \fn            int32_t SendCommand (uint32_t  cmd,
                                      uint32_t  arg,
                                      uint32_t  flags,
                                      uint32_t *response)
  \brief         Send Command to card and get the response.
  \param[in]     cmd       Memory Card command
  \param[in]     arg       Command argument
  \param[in]     flags     Command flags
  \param[out]    response  Pointer to buffer for response
  \return        \ref execution_status
*/
static int32_t SendCommand (uint32_t cmd, uint32_t arg, uint32_t flags, uint32_t *response) {
  sdhc_hal_cmd_req_t cfg;

  if (((flags & MCI_RESPONSE_EXPECTED_Msk) != 0U) && (response == NULL)) {
    return ARM_DRIVER_ERROR_PARAMETER;
  }
  if ((MCI.flags & MCI_SETUP) == 0U) {
    return ARM_DRIVER_ERROR;
  }
  if (MCI.status.command_active) {
    return ARM_DRIVER_ERROR_BUSY;
  }

  if (flags & ARM_MCI_TRANSFER_DATA) {
    while (SDHC->PRSSTAT & SDHC_PRSSTAT_CDIHB_MASK);
  }
  while (SDHC->PRSSTAT & SDHC_PRSSTAT_CIHB_MASK);

  MCI.status.command_active   = 1U;
  MCI.status.command_timeout  = 0U;
  MCI.status.command_error    = 0U;
  MCI.status.transfer_timeout = 0U;
  MCI.status.transfer_error   = 0U;
  MCI.status.ccs              = 0U;

  if (flags & ARM_MCI_CARD_INITIALIZE) {
    SDHC_HAL_InitCard (SDHC, 100U);
  }

  cfg.arg   = arg;
  cfg.index = cmd & 0xFF;

  cfg.dataBlkSize  = MCI.bl_sz;
  cfg.dataBlkCount = MCI.bl_cnt;

  cfg.flags = SDHC_HAL_ENABLE_DMA;
  
  if (cfg.index == 0x0C) {
    cfg.flags |= SDHC_HAL_CMD_TYPE_ABORT;
  }

  MCI.response =  response;
  MCI.flags   &= ~MCI_RESP_LONG;

  switch (flags & ARM_MCI_RESPONSE_Msk) {
    case ARM_MCI_RESPONSE_NONE:
      /* No response expected */
      break;

    case ARM_MCI_RESPONSE_SHORT:
      /* Short response expected */
      cfg.flags |= SDHC_HAL_RESP_LEN_48;
      break;

    case ARM_MCI_RESPONSE_SHORT_BUSY:
      /* Short response with busy expected */
      cfg.flags |= SDHC_HAL_RESP_LEN_48_BC;
      break;

    case ARM_MCI_RESPONSE_LONG:
      MCI.flags |= MCI_RESP_LONG;
      /* Long response expected */
      cfg.flags |= SDHC_HAL_RESP_LEN_136;
      break;

    default:
      return ARM_DRIVER_ERROR;
  }
  if (flags & ARM_MCI_RESPONSE_INDEX) {
    /* Check for command index error */
    cfg.flags |= SDHC_HAL_ENABLE_INDEX_CHECK;
  }

  if (flags & ARM_MCI_RESPONSE_CRC) {
    /* Check for CRC error */
    cfg.flags |= SDHC_HAL_ENABLE_CRC_CHECK;
  }

  if (flags & ARM_MCI_TRANSFER_DATA) {
    /* Enable data transfer */
    cfg.flags |= SDHC_HAL_DATA_PRESENT;

    if (MCI.flags & MCI_DATA_READ) {
      /* Read transfer */
      cfg.flags |= SDHC_HAL_ENABLE_DATA_READ;
    }

    if (MCI.flags & MCI_DATA_MULB) {
      /* Multiple block transfer */
      cfg.flags |= SDHC_HAL_ENABLE_BLOCK_COUNT |
                    SDHC_HAL_MULTIPLE_BLOCK    ;
    }
    MCI.status.transfer_active = 1U;
  }

  /* Send the command */
  SDHC_HAL_SendCmd (SDHC, &cfg);

  return ARM_DRIVER_OK;
}


/**
  \fn            int32_t SetupTransfer (uint8_t *data,
                                        uint32_t block_count,
                                        uint32_t block_size,
                                        uint32_t mode)
  \brief         Setup read or write transfer operation.
  \param[in,out] data         Pointer to data block(s) to be written or read
  \param[in]     block_count  Number of blocks
  \param[in]     block_size   Size of a block in bytes
  \param[in]     mode         Transfer mode
  \return        \ref execution_status
*/
static int32_t SetupTransfer (uint8_t *data, uint32_t block_count, uint32_t block_size, uint32_t mode) {

  if ((data == NULL) || (block_count == 0U) || (block_size == 0U)) { return ARM_DRIVER_ERROR_PARAMETER; }

  if ((MCI.flags & MCI_SETUP) == 0U) {
    return ARM_DRIVER_ERROR;
  }
  if (MCI.status.transfer_active) {
    return ARM_DRIVER_ERROR_BUSY;
  }
  if (mode & ARM_MCI_TRANSFER_STREAM) {
    /* Stream or SDIO multibyte data transfer not supported by peripheral */
    return ARM_DRIVER_ERROR;
  }

  /* Set DMA source/destination address */
  SDHC->DSADDR = (uint32_t)data;

  /* Set transfer block count and size */
  MCI.bl_cnt = block_count;
  MCI.bl_sz  = block_size;

  if (block_count == 1) {
    /* Single block transfer */
    MCI.flags &= ~MCI_DATA_MULB;
  }
  else {
    /* Multiple block transfer */
    MCI.flags |=  MCI_DATA_MULB;
  }

  if (mode & ARM_MCI_TRANSFER_WRITE) {
    /* Direction: From controller to card */
    MCI.flags &= ~MCI_DATA_READ;
  }
  else {
    /* Direction: From card to controller */
    MCI.flags |=  MCI_DATA_READ;
  }

  return (ARM_DRIVER_OK);
}


/**
  \fn            int32_t AbortTransfer (void)
  \brief         Abort current read/write data transfer.
  \return        \ref execution_status
*/
static int32_t AbortTransfer (void) {
  int32_t  status;

  if ((MCI.flags & MCI_SETUP) == 0U) { return ARM_DRIVER_ERROR; }

  status = ARM_DRIVER_OK;

  /* Disable peripheral interrupts */
  SDHC_HAL_SetIntSignal(SDHC, false, SDHC_IRQ_EN_MASK);
  
  /* Software reset for DAT line */
  SDHC_HAL_Reset (SDHC, SDHC_SYSCTL_RSTC_MASK, MCI_RESET_TIMEOUT);

  MCI.status.command_active  = 0U;
  MCI.status.transfer_active = 0U;
  MCI.status.sdio_interrupt  = 0U;
  MCI.status.ccs             = 0U;

  /* Clear pending interrupts */
  SDHC_HAL_ClearIntFlags(SDHC, SDHC_HAL_GetIntFlags (SDHC));

  /* Enable peripheral interrupts */
  SDHC_HAL_SetIntSignal(SDHC, true, SDHC_IRQ_EN_MASK);

  return status;
}


/**
  \fn            int32_t Control (uint32_t control, uint32_t arg)
  \brief         Control MCI Interface.
  \param[in]     control  Operation
  \param[in]     arg      Argument of operation (optional)
  \return        \ref execution_status
*/
static int32_t Control (uint32_t control, uint32_t arg) {
  uint32_t pclk;
  uint32_t pr, div, set_pr, set_div, low;
  int32_t  diff;

  if ((MCI.flags & MCI_POWER) == 0U) { return ARM_DRIVER_ERROR; }

  switch (control) {
    case ARM_MCI_BUS_SPEED:
      pclk = CLOCK_SYS_GetSdhcFreq (0);

      /* Disable clock to the card */
      SDHC_CLR_SYSCTL (SDHC, SDHC_SYSCTL_SDCLKEN_MASK);

      /* Find prescaler and divider */
      low     = pclk;                       /* Closest setting has high init val  */
      set_pr  = 0x80;                       /* Maximum prescaler value            */
      set_div = 0x0F;                       /* Maximum divider value              */

      for (pr = 1; pr <= SDHC_HAL_MAX_CLKFS; pr <<= 1) {
        for (div = 1; div <= SDHC_HAL_MAX_DVS; div++) {
          diff = arg - (pclk / (pr * div));
          if (diff >= 0) {
            if (low > diff) {
              low     = diff;
              set_pr  = pr;
              set_div = div;
            }
          }
        }
      }
      /* Set new clock */
      SDHC_WR_SYSCTL_DVS     (SDHC, set_div - 1);
      SDHC_WR_SYSCTL_SDCLKFS (SDHC, set_pr >> 1);

      /* Wait until internal card clock is stable */
      while (SDHC_BRD_PRSSTAT_SDSTB(SDHC) == 0);

      /* Enable clock to the card */
      SDHC_SET_SYSCTL (SDHC, SDHC_SYSCTL_SDCLKEN_MASK);

      /* Bus speed configured */
      MCI.flags |= MCI_SETUP;
      return ((int32_t)(pclk / (set_div * set_pr)));

    case ARM_MCI_BUS_SPEED_MODE:
      switch (arg) {
        case ARM_MCI_BUS_DEFAULT_SPEED:
          /* Speed mode up to 25MHz */
        case ARM_MCI_BUS_HIGH_SPEED:
          /* Speed mode up to 50MHz */
          break;
        default: return ARM_DRIVER_ERROR_UNSUPPORTED;
      }
      break;

    case ARM_MCI_BUS_CMD_MODE:
      switch (arg) {
        case ARM_MCI_BUS_CMD_OPEN_DRAIN:
          /* Configure command line in open-drain mode */
          return ARM_DRIVER_ERROR;

        case ARM_MCI_BUS_CMD_PUSH_PULL:
          /* Configure command line in push-pull mode */
          break;

        default:
          return ARM_DRIVER_ERROR_UNSUPPORTED;
      }
      break;

    case ARM_MCI_BUS_DATA_WIDTH:
      switch (arg) {
        case ARM_MCI_BUS_DATA_WIDTH_1:
          SDHC_HAL_SetDataTransferWidth (SDHC, kSdhcHalDtw1Bit);
          break;
        case ARM_MCI_BUS_DATA_WIDTH_4:
          SDHC_HAL_SetDataTransferWidth (SDHC, kSdhcHalDtw4Bit);
          break;
        case ARM_MCI_BUS_DATA_WIDTH_8:
          SDHC_HAL_SetDataTransferWidth (SDHC, kSdhcHalDtw4Bit);
          break;
        default:
          return ARM_DRIVER_ERROR_UNSUPPORTED;
      }
      break;

    case ARM_MCI_CONTROL_CLOCK_IDLE:
      if (arg) {
        /* Clock generation enabled when idle */
        SDHC_SET_SYSCTL (SDHC, SDHC_SYSCTL_IPGEN_MASK |
                               SDHC_SYSCTL_HCKEN_MASK |
                               SDHC_SYSCTL_PEREN_MASK );
      }
      else {
        /* Clock generation disabled when idle */
        SDHC_CLR_SYSCTL (SDHC, SDHC_SYSCTL_IPGEN_MASK |
                               SDHC_SYSCTL_HCKEN_MASK |
                               SDHC_SYSCTL_PEREN_MASK );
      }
      break;

    case ARM_MCI_DATA_TIMEOUT:
      arg >>= 13;
      if (arg > 0x0E) {
        arg = 0x0E;
      }
      SDHC_WR_SYSCTL_DTOCV (SDHC, arg);
      break;
    
    case ARM_MCI_MONITOR_SDIO_INTERRUPT:
      MCI.status.sdio_interrupt = 0U;
      /* Clear SDIO Interrupt status */
      SDHC_HAL_ClearIntFlags(SDHC, SDHC_HAL_CARD_INT);
      /* Enable SDIO Interrupt request */
      SDHC_HAL_SetIntSignal (SDHC, true, SDHC_HAL_CARD_INT);
      break;

    default: return ARM_DRIVER_ERROR_UNSUPPORTED;
  }

  return ARM_DRIVER_OK;
}


/**
  \fn            ARM_MCI_STATUS GetStatus (void)
  \brief         Get MCI status.
  \return        MCI status \ref ARM_MCI_STATUS
*/
static ARM_MCI_STATUS GetStatus (void) {
  return MCI.status;
}


/* SDHC IRQ Handler */
void SDHC_IRQHandler (void) {
  uint32_t irq, msk, event;

  event = 0U;
  irq   = SDHC_HAL_GetIntFlags (SDHC);


  if (irq & SDHC_HAL_CMD_ALL_INT) {
    irq &= SDHC_HAL_CMD_ALL_INT;
    /* Command interrupt */
    MCI.status.command_active = 0U;

    if ((irq & (SDHC_HAL_CMD_TIMEOUT_ERR_INT|SDHC_HAL_CMD_CRC_ERR_INT)) == SDHC_HAL_CMD_TIMEOUT_ERR_INT) {
      /* Response timeout error */
      MCI.status.command_timeout = 1U;
      event = ARM_MCI_EVENT_COMMAND_TIMEOUT;
    }
    else if ((irq & (SDHC_HAL_CMD_TIMEOUT_ERR_INT|SDHC_HAL_CMD_CRC_ERR_INT)) == SDHC_HAL_CMD_CRC_ERR_INT) {
      /* Response CRC error */
      MCI.status.command_error = 1U;
      event = ARM_MCI_EVENT_COMMAND_ERROR;
    }
    else if (irq & SDHC_HAL_CMD_INDEX_ERR_INT) {
      /* Command index invalid */
      MCI.status.command_error = 1U;
      event = ARM_MCI_EVENT_COMMAND_ERROR;
    }
    else if (irq & SDHC_HAL_CMD_END_BIT_ERR_INT) {
      /* Command end bit error */
      event = ARM_MCI_EVENT_COMMAND_ERROR;
    }
    else {
      msk = SDHC_HAL_CMD_COMPLETE_INT | SDHC_HAL_CMD_TIMEOUT_ERR_INT | SDHC_HAL_CMD_CRC_ERR_INT;

      if ((irq & msk) == SDHC_HAL_CMD_COMPLETE_INT) {
        /* Command completed */
        if (MCI.response) {

          if (MCI.flags & MCI_RESP_LONG) {
            MCI.response[3] = (SDHC->CMDRSP[3] << 8) | (SDHC->CMDRSP[2] >> 24);
            MCI.response[2] = (SDHC->CMDRSP[2] << 8) | (SDHC->CMDRSP[1] >> 24);
            MCI.response[1] = (SDHC->CMDRSP[1] << 8) | (SDHC->CMDRSP[0] >> 24);
            MCI.response[0] =  SDHC->CMDRSP[0] << 8;
          }
          else {
            MCI.response[0] = SDHC->CMDRSP[0];
          }
        }
        event = ARM_MCI_EVENT_COMMAND_COMPLETE;
      }
    }
  }
  else if (irq & SDHC_HAL_DATA_ALL_INT) {
    irq &= SDHC_HAL_DATA_ALL_INT;
    /* Data transfer */
    MCI.status.transfer_active = 0U;

    msk = SDHC_HAL_DATA_CRC_ERR_INT | SDHC_HAL_DATA_END_BIT_ERR_INT | SDHC_HAL_DMA_ERR_INT;
    if (irq & msk) {
      /* CRC, end bit or DMA error */
      MCI.status.transfer_error = 1U;
      event = ARM_MCI_EVENT_TRANSFER_ERROR;
    }
    else if (irq & SDHC_HAL_DATA_TIMEOUT_ERR_INT) {
      /* Data transfer timeout */
      MCI.status.transfer_timeout = 1U;
      event = ARM_MCI_EVENT_TRANSFER_TIMEOUT;
    }
    else {
      if (irq & SDHC_HAL_DATA_COMPLETE_INT) {
        /* Data transfer complete */
        event = ARM_MCI_EVENT_TRANSFER_COMPLETE;
      }
    }
  }
  else {
    if (irq & SDHC_HAL_CARD_INT) {
      /* SDIO interrupt */
      MCI.status.sdio_interrupt = 1U;
      event = ARM_MCI_EVENT_SDIO_INTERRUPT;

      /* Disable SDIO Interrupt request */
      SDHC_HAL_SetIntSignal (SDHC, false, SDHC_HAL_CARD_INT);
    }
  }

  SDHC_HAL_ClearIntFlags(SDHC, irq);
  
  if (event && (MCI.cb_event != NULL)) {
    MCI.cb_event (event);
  }
}


/* MCI Driver Control Block */
ARM_DRIVER_MCI Driver_MCI0 = {
  GetVersion,
  GetCapabilities,
  Initialize,
  Uninitialize,
  PowerControl,
  CardPower,
  ReadCD,
  ReadWP,
  SendCommand,
  SetupTransfer,
  AbortTransfer,
  Control,
  GetStatus
};
