/* -----------------------------------------------------------------------------
 * Copyright (c) 2013-2015 ARM Ltd.
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from
 * the use of this software. Permission is granted to anyone to use this
 * software for any purpose, including commercial applications, and to alter
 * it and redistribute it freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software in
 *    a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 *
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 3. This notice may not be removed or altered from any source distribution.
 *
 *
 * $Date:        18. June 2015
 * $Revision:    V1.2
 *
 * Driver:       Driver_SAI0
 * Configured:   via RTE_Device.h configuration file
 * Project:      SAI Driver for Freescale MK64F
 * --------------------------------------------------------------------------
 * Use the following configuration settings in the middleware component
 * to connect to this driver.
 *
 *   Configuration Setting                 Value   SPI Interface
 *   ---------------------                 -----   -------------
 *   Connect to hardware via Driver_SAI# = 0       use SAI0 (SAI0)
 * -------------------------------------------------------------------------- */

/* History:
 *  Version 1.2
 *    - PowerControl for Power OFF and Uninitialize functions made unconditional.
 *    - Corrected status bit-field handling, to prevent race conditions.
 *  Version 1.1
 *    Updated for KSDK 1.2.0
 *  Version 1.0
 *    Initial release
 */

#include "SAI_MK64F.h"


#define ARM_SAI_DRV_VERSION ARM_DRIVER_VERSION_MAJOR_MINOR(1,2)   // driver version
// Driver Version
static const ARM_DRIVER_VERSION DriverVersion = {
  ARM_SAI_API_VERSION,
  ARM_SAI_DRV_VERSION
};

#if ((RTE_SAI0) && (RTE_SAI0_DMA_TX == 1))
static void SAI_TX_DMA_Callback (void *par, edma_chn_status_t status);
#endif

#if ((RTE_SAI0) && (RTE_SAI0_DMA_RX == 1))
static void SAI_RX_DMA_Callback (void *par, edma_chn_status_t status);
#endif


// FIFO watermark
#ifndef SAI0_TX_FIFO_WATERMARK
#define SAI0_TX_FIFO_WATERMARK (1U)
#endif

#if (SAI0_TX_FIFO_WATERMARK > 8)
#error "FIFO watermark max value is 8"
#endif

// FIFO watermark
#ifndef SAI0_RX_FIFO_WATERMARK
#define SAI0_RX_FIFO_WATERMARK (1U)
#endif

#if (SAI0_RX_FIFO_WATERMARK > 8)
#error "FIFO watermark max value is 8"
#endif

// SAI0 DMA Configuration
#if ((RTE_SAI0) && (RTE_SAI0_DMA_TX == 1))
edma_chn_state_t       SAI0_DMA_TxState;
edma_transfer_config_t SAI0_DMA_TxCfg;

static SAI_DMA SAI0_DMA_Tx = {
  kDmaRequestMux0I2S0Tx,
  &SAI0_DMA_TxState,
  &SAI0_DMA_TxCfg
};
#endif

#if ((RTE_SAI0) && (RTE_SAI0_DMA_RX == 1))
edma_chn_state_t       SAI0_DMA_RxState;
edma_transfer_config_t SAI0_DMA_RxCfg;

static SAI_DMA SAI0_DMA_Rx = {
  kDmaRequestMux0I2S0Rx,
  &SAI0_DMA_RxState,
  &SAI0_DMA_RxCfg
};
#endif

// SAI Run-time information
static SAI_INFO SAI0_info = {0};

static const SAI_RESOURCES SAI0_Resources = {
  I2S0,
  kSimClockGateSai0,
  I2S0_Tx_IRQn,
  I2S0_Rx_IRQn,
#if ((RTE_SAI0) && (RTE_SAI0_DMA_TX == 1))
 &SAI0_DMA_Tx,
#endif
#if ((RTE_SAI0) && (RTE_SAI0_DMA_RX == 1))
 &SAI0_DMA_Rx,
#endif
  SAI0_TX_FIFO_WATERMARK,
  SAI0_RX_FIFO_WATERMARK,
 &SAI0_info
};

static const uint32_t mute_data = 0;
const SAI_RESOURCES * const sai = &SAI0_Resources;


// SAI Capabilities
static ARM_SAI_CAPABILITIES SAI_Capabilities = {
  1,   ///< supports asynchronous Transmit/Receive
  1,   ///< supports synchronous Transmit/Receive
  1,   ///< supports user defined Protocol
  1,   ///< supports I2S Protocol
  1,   ///< supports MSB/LSB justified Protocol
  1,   ///< supports PCM short/long frame Protocol
  1,   ///< supports AC'97 Protocol
  0,   ///< supports Mono mode
  0,   ///< supports Companding
  1,   ///< supports MCLK (Master Clock) pin
  1,   ///< supports Frame error event: \ref ARM_SAI_EVENT_FRAME_ERROR
};

/**
  \fn          ARM_DRIVER_VERSION SAI_GetVersion (void)
  \brief       Get driver version.
  \return      \ref ARM_DRIVER_VERSION
*/
static ARM_DRIVER_VERSION SAI_GetVersion (void) {
  return (DriverVersion);
}

/**
  \fn          ARM_SAI_CAPABILITIES SAI0_GetCapabilities (void)
  \brief       Get driver capabilities.
  \return      \ref ARM_SAI_CAPABILITIES
*/
static ARM_SAI_CAPABILITIES SAI0_GetCapabilities (void) {
  return (SAI_Capabilities);
}

/**
  \fn          int32_t SAI0_Initialize (ARM_SAI_SignalEvent_t cb_event)
  \brief       Initialize SAI Interface.
  \param[in]   cb_event  Pointer to \ref ARM_SAI_SignalEvent
  \return      \ref execution_status
*/
static int32_t SAI0_Initialize (ARM_SAI_SignalEvent_t cb_event) {

  if (sai->info->flags & SAI_FLAG_INITIALIZED) {
    // Driver is already initialized
    return ARM_DRIVER_OK;
  }

  // Initialize SAI Run-Time resources
  sai->info->cb_event             = cb_event;
  sai->info->status.frame_error   = 0U;
  sai->info->status.rx_busy       = 0U;
  sai->info->status.rx_overflow   = 0U;
  sai->info->status.tx_busy       = 0U;
  sai->info->status.tx_underflow  = 0U;

  // DMA Initialize
#if ((RTE_SAI0) && (RTE_SAI0_DMA_TX == 1))
  if (kEDMAInvalidChannel != EDMA_DRV_RequestChannel (kEDMAAnyChannel,
                                                      sai->tx_dma->req_src,
                                                      sai->tx_dma->ch_info)) {

    EDMA_DRV_InstallCallback (sai->tx_dma->ch_info, SAI_TX_DMA_Callback, NULL);
  } else {
    return ARM_DRIVER_ERROR;
  }

  // Partly prepare DMA channel config structure
  sai->tx_dma->cfg->srcLastAddrAdjust  = 0;
  sai->tx_dma->cfg->destLastAddrAdjust = 0;
  sai->tx_dma->cfg->srcModulo          = kEDMAModuloDisable;
  sai->tx_dma->cfg->destModulo         = kEDMAModuloDisable;
  sai->tx_dma->cfg->minorLoopCount     = 1;
#endif

#if ((RTE_SAI0) && (RTE_SAI0_DMA_RX == 1))
  if (kEDMAInvalidChannel != EDMA_DRV_RequestChannel (kEDMAAnyChannel,
                                                      sai->rx_dma->req_src,
                                                      sai->rx_dma->ch_info)) {

    EDMA_DRV_InstallCallback (sai->rx_dma->ch_info, SAI_RX_DMA_Callback, NULL);
  } else {
    return ARM_DRIVER_ERROR;
  }

  // Partly prepare DMA channel config structure
  sai->rx_dma->cfg->srcLastAddrAdjust  = 0;
  sai->rx_dma->cfg->destLastAddrAdjust = 0;
  sai->rx_dma->cfg->srcModulo          = kEDMAModuloDisable;
  sai->rx_dma->cfg->destModulo         = kEDMAModuloDisable;
  sai->rx_dma->cfg->minorLoopCount     = 1;
#endif

  sai->info->flags = SAI_FLAG_INITIALIZED;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t SAI0_Uninitialize void)
  \brief       De-initialize SAI Interface.
  \return      \ref execution_status
*/
static int32_t SAI0_Uninitialize (void) {

  // Reset SAI status flags
  sai->info->flags = 0U;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t SAI0_PowerControl (ARM_POWER_STATE state)
  \brief       Control SAI Interface Power.
  \param[in]   state  Power state
  \return      \ref execution_status
*/
static int32_t SAI0_PowerControl (ARM_POWER_STATE state) {

  switch (state) {
    case ARM_POWER_OFF:

      // Disable SAI IRQ
      NVIC_DisableIRQ(sai->tx_irq_num);
      NVIC_DisableIRQ(sai->rx_irq_num);

      // Disable SAI register interface clock
      SIM_HAL_DisableClock (SIM, sai->clock_gate);
      
      sai->info->status.frame_error   = 0U;
      sai->info->status.rx_busy       = 0U;
      sai->info->status.rx_overflow   = 0U;
      sai->info->status.tx_busy       = 0U;
      sai->info->status.tx_underflow  = 0U;

      sai->info->flags = SAI_FLAG_INITIALIZED;
      break;

    case ARM_POWER_LOW:
      return ARM_DRIVER_ERROR_UNSUPPORTED;

    case ARM_POWER_FULL:
      if (sai->info->flags & SAI_FLAG_POWERED) {
        return ARM_DRIVER_OK;
      }
      
      sai->info->status.frame_error   = 0U;
      sai->info->status.rx_busy       = 0U;
      sai->info->status.rx_overflow   = 0U;
      sai->info->status.tx_busy       = 0U;
      sai->info->status.tx_underflow  = 0U;

      // Enable SAI register interface clock
      SIM_HAL_EnableClock (SIM, sai->clock_gate);

      // Disable transmitter and receiver, reset registers
      SAI_HAL_TxInit (sai->base);
      SAI_HAL_RxInit (sai->base);

      // Disable SAI interrupts
      SAI_HAL_TxSetIntCmd (sai->base, kSaiIntrequestWordStart,   false);
      SAI_HAL_TxSetIntCmd (sai->base, kSaiIntrequestSyncError,   false);
      SAI_HAL_TxSetIntCmd (sai->base, kSaiIntrequestFIFOWarning, false);
      SAI_HAL_TxSetIntCmd (sai->base, kSaiIntrequestFIFOError,   false);
      SAI_HAL_TxSetIntCmd (sai->base, kSaiIntrequestFIFORequest, false);
      SAI_HAL_RxSetIntCmd (sai->base, kSaiIntrequestWordStart,   false);
      SAI_HAL_RxSetIntCmd (sai->base, kSaiIntrequestSyncError,   false);
      SAI_HAL_RxSetIntCmd (sai->base, kSaiIntrequestFIFOWarning, false);
      SAI_HAL_RxSetIntCmd (sai->base, kSaiIntrequestFIFOError,   false);
      SAI_HAL_RxSetIntCmd (sai->base, kSaiIntrequestFIFORequest, false);

      // Set FIFO watermark
      SAI_HAL_TxSetWatermark (sai->base, SAI0_TX_FIFO_WATERMARK);
      SAI_HAL_RxSetWatermark (sai->base, SAI0_RX_FIFO_WATERMARK);

      sai->info->flags = SAI_FLAG_POWERED | SAI_FLAG_INITIALIZED;

      // Clear and Enable SAI IRQ
      NVIC_ClearPendingIRQ(sai->tx_irq_num);
      NVIC_ClearPendingIRQ(sai->rx_irq_num);
      NVIC_EnableIRQ(sai->tx_irq_num);
      NVIC_EnableIRQ(sai->rx_irq_num);

      break;

    default: return ARM_DRIVER_ERROR_UNSUPPORTED;
  }
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t SAI0_Send (const void *data, uint32_t num)
  \brief       Start sending data to SAI transmitter.
  \param[in]   data  Pointer to buffer with data to send to SAI transmitter
  \param[in]   num   Number of data items to send
  \return      \ref execution_status
*/
static int32_t SAI0_Send (const void *data, uint32_t num) {
#if ((RTE_SAI0) && (RTE_SAI0_DMA_TX == 1))
  edma_software_tcd_t  tcd;
  edma_transfer_size_t sz;
#endif

  if ((data == NULL) || (num == 0U)) {
    // Invalid parameters
    return ARM_DRIVER_ERROR_PARAMETER;
  }

  if ((sai->info->flags & SAI_FLAG_CONFIGURED) == 0U) {
    // SAI is not configured (mode not selected)
    return ARM_DRIVER_ERROR;
  }

  if (sai->info->status.tx_busy) {
    // Send is not completed yet
    return ARM_DRIVER_ERROR_BUSY;
  }

  // Set Send active flag
  sai->info->status.tx_busy = 1U;

  // Clear TX underflow flag
  sai->info->status.tx_underflow = 0U;

  // Save transmit buffer info
  sai->info->tx.buf = (uint8_t *)data;
  sai->info->tx.cnt = 0U;
  sai->info->tx.num = num;

  //  Enable TX channel
  SAI_HAL_TxSetDataChn (sai->base, 0);

  // DMA mode
#if ((RTE_SAI0) && (RTE_SAI0_DMA_TX == 1))
  if (sai->info->tx.data_bits > 16)  { sz = kEDMATransferSize_4Bytes; }
  else { 
    if (sai->info->tx.data_bits > 8) { sz = kEDMATransferSize_2Bytes; }
    else                             { sz = kEDMATransferSize_1Bytes; }
  }
  // Prepare config structure
  sai->tx_dma->cfg->srcTransferSize  = sz;
  sai->tx_dma->cfg->destTransferSize = sz;
  sai->tx_dma->cfg->destAddr         = SAI_HAL_TxGetFifoAddr (sai->base, 0);

  // If transmitter is in mute mode, send 0
  if (sai->info->tx.mute == 0U) {
    sai->tx_dma->cfg->srcAddr          = (uint32_t)data;
    sai->tx_dma->cfg->srcOffset        = 1;
  } else {
    sai->tx_dma->cfg->srcAddr          = (uint32_t)(&mute_data);
    sai->tx_dma->cfg->srcOffset        = 0;
  }
  sai->tx_dma->cfg->destOffset       = 0;
  sai->tx_dma->cfg->majorLoopCount   = num;

  EDMA_DRV_PrepareDescriptorTransfer (sai->tx_dma->ch_info, &tcd, sai->tx_dma->cfg, true, false);

  EDMA_HAL_STCDSetScatterGatherCmd   (&tcd, false);
  EDMA_HAL_STCDSetChannelMajorLink   (&tcd, sai->tx_dma->ch_info->channel, false);
  EDMA_HAL_STCDSetHalfCompleteIntCmd (&tcd, false);
  EDMA_HAL_STCDSetBandwidth          (&tcd, kEDMABandwidthStallNone);
  EDMA_HAL_STCDSetDisableDmaRequestAfterTCDDoneCmd (&tcd, true);

  EDMA_DRV_PushDescriptorToReg       (sai->tx_dma->ch_info, &tcd);
  EDMA_DRV_StartChannel              (sai->tx_dma->ch_info);

  // Enable SAI TX DMA request
  SAI_HAL_TxSetDmaCmd (sai->base, kSaiDmaReqFIFOWarning, true);
  SAI_HAL_TxSetDmaCmd (sai->base, kSaiDmaReqFIFORequest, true);

#else  // Interrupt mode
  // Enable SAI transmit interrupt;
  SAI_HAL_TxSetIntCmd (sai->base, kSaiIntrequestSyncError,   true);
  SAI_HAL_TxSetIntCmd (sai->base, kSaiIntrequestFIFORequest, true);
#endif

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t SAI0_Receive (void *data, uint32_t num)
  \brief       Start receiving data from SAI receiver.
  \param[out]  data  Pointer to buffer for data to receive from SAI receiver
  \param[in]   num   Number of data items to receive
  \return      \ref execution_status
*/
static int32_t SAI0_Receive (void *data, uint32_t num) {
#if ((RTE_SAI0) && (RTE_SAI0_DMA_RX == 1))
  edma_software_tcd_t  tcd;
  edma_transfer_size_t sz;
#endif

  if ((data == NULL) || (num == 0U)) {
    // Invalid parameters
    return ARM_DRIVER_ERROR_PARAMETER;
  }

  if ((sai->info->flags & SAI_FLAG_CONFIGURED) == 0U) {
    // SAI is not configured (mode not selected)
    return ARM_DRIVER_ERROR;
  }

  if (sai->info->status.rx_busy) {
    // Receive is not completed yet
    return ARM_DRIVER_ERROR_BUSY;
  }

  // Set Receive active flag
  sai->info->status.rx_busy = 1U;

  // Clear RX overflow flag
  sai->info->status.rx_overflow = 0U;

  // Save receive buffer info
  sai->info->rx.buf = (uint8_t *)data;
  sai->info->rx.cnt = 0U;
  sai->info->rx.num = num;

  // Enable RX channel
  SAI_HAL_RxSetDataChn (sai->base, 0);

  // DMA mode
#if ((RTE_SAI0) && (RTE_SAI0_DMA_RX == 1))
  if (sai->info->rx.data_bits > 16)  { sz = kEDMATransferSize_4Bytes; }
  else { 
    if (sai->info->rx.data_bits > 8) { sz = kEDMATransferSize_2Bytes; }
    else                             { sz = kEDMATransferSize_1Bytes; }
  }
  // Prepare config structure
  sai->rx_dma->cfg->srcTransferSize  = sz;
  sai->rx_dma->cfg->destTransferSize = sz;
  sai->rx_dma->cfg->destAddr         = (uint32_t)data;
  sai->rx_dma->cfg->destOffset       = 1;
  sai->rx_dma->cfg->srcAddr          = SAI_HAL_RxGetFifoAddr (sai->base, 0);
  sai->rx_dma->cfg->srcOffset        = 0;
  sai->rx_dma->cfg->majorLoopCount   = num;

  EDMA_DRV_PrepareDescriptorTransfer (sai->rx_dma->ch_info, &tcd, sai->rx_dma->cfg, true, false);

  EDMA_HAL_STCDSetScatterGatherCmd   (&tcd, false);
  EDMA_HAL_STCDSetChannelMajorLink   (&tcd, sai->rx_dma->ch_info->channel, false);
  EDMA_HAL_STCDSetHalfCompleteIntCmd (&tcd, false);
  EDMA_HAL_STCDSetBandwidth          (&tcd, kEDMABandwidthStallNone);
  EDMA_HAL_STCDSetDisableDmaRequestAfterTCDDoneCmd (&tcd, true);

  EDMA_DRV_PushDescriptorToReg       (sai->rx_dma->ch_info, &tcd);
  EDMA_DRV_StartChannel              (sai->rx_dma->ch_info);

  // Enable SAI RX DMA request
  SAI_HAL_RxSetDmaCmd (sai->base, kSaiDmaReqFIFORequest, true);

#else  // Interrupt mode
  // Enable SAI receive interrupt;
  SAI_HAL_RxSetIntCmd (sai->base, kSaiIntrequestSyncError,   true);
  SAI_HAL_RxSetIntCmd (sai->base, kSaiIntrequestFIFORequest, true);
#endif

  return ARM_DRIVER_OK;
}

/**
  \fn          uint32_t SAI0_GetTxCount (void)
  \brief       Get transmitted data count.
  \return      number of data items transmitted
*/
static uint32_t SAI0_GetTxCount (void) {
  return (sai->info->tx.cnt);
}

/**
  \fn          uint32_t SAI0_GetRxCount (void)
  \brief       Get received data count.
  \return      number of data items received
*/
static uint32_t SAI0_GetRxCount (void) {
  return (sai->info->rx.cnt);
}

/**
  \fn          int32_t SAI0_Control (uint32_t control, uint32_t arg1, uint32_t arg2)
  \brief       Control SAI Interface.
  \param[in]   control  Operation
  \param[in]   arg1     Argument 1 of operation (optional)
  \param[in]   arg2     Argument 2 of operation (optional)
  \return      common \ref execution_status and driver specific \ref sai_execution_status
*/
static int32_t SAI0_Control (uint32_t control, uint32_t arg1, uint32_t arg2) {
  uint32_t val, protocol, master, mclk, frame_len, frame_sync_width, frame_sync_pol;
  uint32_t bit_order, clock_pol, frame_sync_early, slot_count, slot_sz, slot_offset;
  I2S_STREAM_INFO *stream;

  if ((sai->info->flags & SAI_FLAG_POWERED) == 0U) {
    // SAI not powered
    return ARM_DRIVER_ERROR;
  }

  // Clear local variables
  val = protocol = master = mclk = frame_len = frame_sync_width = frame_sync_pol = 0U;
  bit_order = clock_pol = frame_sync_early = slot_count = slot_sz = slot_offset = 0U;

  switch (control & ARM_SAI_CONTROL_Msk) {
    case ARM_SAI_CONFIGURE_TX:
      stream = &sai->info->tx;
      break;

    case ARM_SAI_CONFIGURE_RX:
      stream = &sai->info->rx;
      break;

    case ARM_SAI_CONTROL_TX:
      if ((arg1 & 1U) == 0U) {
        // Disable transmitter
        SAI_HAL_TxDisable (sai->base);
      } else {
        // Enable transmitter
        SAI_HAL_TxEnable (sai->base);

        // Enable transmit FIFO error to detect underflow
        SAI_HAL_TxSetIntCmd (sai->base, kSaiIntrequestFIFOError, true);
      }
      // Mute
      sai->info->tx.mute = (((arg1 & 2U) != 0U) ? 1U : 0U);
      return ARM_DRIVER_OK;

    case ARM_SAI_CONTROL_RX:
      if ((arg1 & 1U) == 0U) {
        // Disable receiver
        SAI_HAL_RxDisable (sai->base);
      } else {
        // Enable receiver
        SAI_HAL_RxEnable (sai->base);

        // Enable transmit FIFO error to detect overflow
        SAI_HAL_RxSetIntCmd (sai->base, kSaiIntrequestFIFOError, true);
      }
      return ARM_DRIVER_OK;

    case ARM_SAI_MASK_SLOTS_TX:
      if (sai->info->tx.protocol != ARM_SAI_PROTOCOL_USER) { return ARM_DRIVER_ERROR; }
      I2S_WR_TMR(sai->base, arg1);
      return ARM_DRIVER_ERROR;

    case ARM_SAI_MASK_SLOTS_RX:
      if (sai->info->rx.protocol != ARM_SAI_PROTOCOL_USER) { return ARM_DRIVER_ERROR; }
      I2S_WR_RMR(sai->base, arg1);
      return ARM_DRIVER_ERROR;

    case ARM_SAI_ABORT_SEND:
      // Disable TX interrupt
      SAI_HAL_TxSetIntCmd (sai->base, kSaiIntrequestFIFORequest, false);

      // Disable DMA channel
#if ((RTE_SAI0) && (RTE_SAI0_DMA_TX == 1))
      SAI_HAL_TxSetDmaCmd (sai->base, kSaiDmaReqFIFORequest, false);

      // If DMA mode - disable DMA channel
      if (sai->info->status.tx_busy) {
        EDMA_DRV_StopChannel (sai->tx_dma->ch_info);
      }
#endif

      // Disable transmitter
      SAI_HAL_TxDisable (sai->base);

      // Reset TX FIFO
      SAI_HAL_TxSetReset (sai->base, kSaiResetTypeFIFO);

      // Reset counters
      sai->info->tx.cnt = 0U;
      sai->info->tx.num = 0U;

      // Clear busy flag
      sai->info->status.tx_busy = 0U;

      return ARM_DRIVER_OK;

    case ARM_SAI_ABORT_RECEIVE:
      // Disable RX interrupt
      SAI_HAL_RxSetIntCmd (sai->base, kSaiIntrequestFIFORequest, false);

      // Disable DMA channel;
#if ((RTE_SAI0) && (RTE_SAI0_DMA_RX == 1))
      SAI_HAL_RxSetDmaCmd (sai->base, kSaiDmaReqFIFORequest, false);

      // If DMA mode - disable DMA channel
      if (sai->info->status.rx_busy) {
        EDMA_DRV_StopChannel (sai->rx_dma->ch_info);
      }
#endif

      // Disable receiver
      SAI_HAL_RxDisable (sai->base);

      // Reset TX FIFO
      SAI_HAL_RxSetReset (sai->base, kSaiResetTypeFIFO);

      // Reset counters
      sai->info->tx.cnt = 0U;
      sai->info->tx.num = 0U;

      // Clear busy flag
      sai->info->status.tx_busy = 0U;
      return ARM_DRIVER_OK;

    default: return ARM_DRIVER_ERROR;
  }

  // Mode
  switch (control & ARM_SAI_MODE_Msk) {
    case ARM_SAI_MODE_MASTER:
      master = 1U;
      if ((control & ARM_SAI_CONTROL_Msk) == ARM_SAI_CONFIGURE_TX) {
        SAI_HAL_TxSetMasterSlave (sai->base, kSaiMaster);
      } else {
        SAI_HAL_RxSetMasterSlave (sai->base, kSaiMaster);
      }
      break;
    case ARM_SAI_MODE_SLAVE:
      if ((control & ARM_SAI_CONTROL_Msk) == ARM_SAI_CONFIGURE_TX) {
        SAI_HAL_TxSetMasterSlave (sai->base, kSaiSlave);
      } else {
        SAI_HAL_RxSetMasterSlave (sai->base, kSaiSlave);
      }
      break;
    default: return ARM_DRIVER_ERROR;
  }

  // Synchronization
  switch (control & ARM_SAI_SYNCHRONIZATION_Msk) {
    case ARM_SAI_ASYNCHRONOUS:
      if ((control & ARM_SAI_CONTROL_Msk) == ARM_SAI_CONFIGURE_TX) {
        SAI_HAL_TxSetSyncMode (sai->base, kSaiModeAsync);
      } else {
        SAI_HAL_RxSetSyncMode (sai->base, kSaiModeAsync);
      }
      break;
    case ARM_SAI_SYNCHRONOUS:
      if ((control & ARM_SAI_CONTROL_Msk) == ARM_SAI_CONFIGURE_TX) {
        // Transmitter is synchronous to receiver
        SAI_HAL_TxSetSyncMode (sai->base, kSaiModeSync);
      } else {
        // Receiver is synchronous to transmitter
        SAI_HAL_RxSetSyncMode (sai->base, kSaiModeSync);
      }
      break;
    default: return ARM_SAI_ERROR_SYNCHRONIZATION;
  }

  // Get Data size
  stream->data_bits = ((control & ARM_SAI_DATA_SIZE_Msk) >> ARM_SAI_DATA_SIZE_Pos) + 1;
  if (stream->data_bits < 8) {
    return ARM_SAI_ERROR_DATA_SIZE;
  }

  // Get Bit Order
  bit_order = control & ARM_SAI_BIT_ORDER_Msk;

  // Clock polarity
  clock_pol = control & ARM_SAI_CLOCK_POLARITY_Msk;

  // Get Frame length
  frame_len = ((arg1 & ARM_SAI_FRAME_LENGTH_Msk) >> ARM_SAI_FRAME_LENGTH_Pos) + 1U;
  // Check if default value
  if (frame_len == 1U) { frame_len = 0U; }

  // Get Slot size
  switch (arg1 & ARM_SAI_SLOT_SIZE_Msk) {
    case ARM_SAI_SLOT_SIZE_DEFAULT: slot_sz = stream->data_bits; break;
    case ARM_SAI_SLOT_SIZE_16:      slot_sz = 16;                break;
    case ARM_SAI_SLOT_SIZE_32:      slot_sz = 32;                break;
    default: return ARM_SAI_ERROR_SLOT_SIZE;
  }

  // Get Agument1 values
  frame_sync_width = ((arg1 & ARM_SAI_FRAME_SYNC_WIDTH_Msk) >> ARM_SAI_FRAME_SYNC_WIDTH_Pos) + 1;
  frame_sync_pol   =   arg1 & ARM_SAI_FRAME_SYNC_POLARITY_Msk;
  frame_sync_early = ((arg1 & ARM_SAI_FRAME_SYNC_EARLY) == 0) ? false : true;
  slot_count       = ((arg1 & ARM_SAI_SLOT_COUNT_Msk) >> ARM_SAI_SLOT_COUNT_Pos) + 1;
  slot_offset      =  (arg1 & ARM_SAI_SLOT_OFFSET_Msk) >> ARM_SAI_SLOT_OFFSET_Pos;


  // Protocol
  protocol = control & ARM_SAI_PROTOCOL_Msk;
  if ((control & ARM_SAI_CONTROL_Msk) == ARM_SAI_CONFIGURE_TX) {
    sai->info->tx.protocol = protocol;
  } else {
    sai->info->rx.protocol = protocol;
  }
  switch (protocol) {
    // User Protocol
    case ARM_SAI_PROTOCOL_USER:
      if (frame_len == 0U) { return ARM_SAI_ERROR_FRAME_LENGHT; }
      if (slot_offset > slot_sz) { return ARM_SAI_ERROR_SLOT_OFFESET; }
      if (bit_order == ARM_SAI_MSB_FIRST) {
        slot_offset = stream->data_bits + slot_offset;
      } else {
        slot_offset = (1U - slot_offset) & (0x1FU);
      }
      break;

    // I2S
    case ARM_SAI_PROTOCOL_I2S:
      if (frame_len == 0U) {
        frame_len = stream->data_bits * 2U;
      } else {
        if (frame_len != (stream->data_bits * 2U)) { return ARM_SAI_ERROR_FRAME_LENGHT; }
      }
      bit_order        = ARM_SAI_MSB_FIRST;
      frame_sync_width = frame_len / 2U;
      frame_sync_pol   = ARM_SAI_FRAME_SYNC_POLARITY_LOW;
      frame_sync_early = true;
      slot_count       = 2U;
      slot_sz          = stream->data_bits;
      slot_offset      = stream->data_bits;
      break;

    // MSB Justified
    case ARM_SAI_PROTOCOL_MSB_JUSTIFIED:
      if (frame_len == 0U) {
        frame_len = stream->data_bits * 2U;
      }
      switch (frame_len / 2) {
        case 8:  slot_sz = 8;  break;
        case 16: slot_sz = 16; break;
        case 32: slot_sz = 32; break;
        default: return ARM_SAI_ERROR_FRAME_LENGHT;
      }
      bit_order        = ARM_SAI_MSB_FIRST;
      frame_sync_width = frame_len / 2U;
      frame_sync_pol   = ARM_SAI_FRAME_SYNC_POLARITY_HIGH;
      frame_sync_early = false;
      slot_count       = 2U;
      slot_offset      = stream->data_bits;
      break;

    // LSB Justified
    case ARM_SAI_PROTOCOL_LSB_JUSTIFIED:
       if (frame_len == 0U) {
        frame_len = stream->data_bits * 2U;
      }
      switch (frame_len / 2) {
        case 8:  slot_sz = 8;  break;
        case 16: slot_sz = 16; break;
        case 32: slot_sz = 32; break;
        default: return ARM_SAI_ERROR_FRAME_LENGHT;
      }
      bit_order        = ARM_SAI_MSB_FIRST;
      frame_sync_width = frame_len / 2U;
      frame_sync_pol   = ARM_SAI_FRAME_SYNC_POLARITY_HIGH;
      frame_sync_early = false;
      slot_count       = 2U;
      slot_offset      = slot_sz;
      break;

    // PCM Short
    case ARM_SAI_PROTOCOL_PCM_SHORT:
       if (frame_len == 0U) {
        frame_len = stream->data_bits;
      }
      switch (frame_len) {
        case 8:  slot_sz = 8;  break;
        case 16: slot_sz = 16; break;
        case 32: slot_sz = 32; break;
        default: return ARM_SAI_ERROR_FRAME_LENGHT;
      }
      bit_order        = ARM_SAI_MSB_FIRST;
      frame_sync_width = 1U;
      frame_sync_pol   = ARM_SAI_FRAME_SYNC_POLARITY_HIGH;
      frame_sync_early = true;
      slot_count       = 1U;
      slot_offset      = stream->data_bits;
      break;

    // PCM Long
    case ARM_SAI_PROTOCOL_PCM_LONG:
      if (frame_len == 0U) {
        frame_len = stream->data_bits;
      }
      switch (frame_len) {
        case 8:  slot_sz = 8;  break;
        case 16: slot_sz = 16; break;
        case 32: slot_sz = 32; break;
        default: return ARM_SAI_ERROR_FRAME_LENGHT;
      }
      bit_order        = ARM_SAI_MSB_FIRST;
      if (stream->data_bits < 16U) { return ARM_SAI_ERROR_DATA_SIZE; }
      frame_sync_width = 13U;
      frame_sync_pol   = ARM_SAI_FRAME_SYNC_POLARITY_HIGH;
      frame_sync_early = false;
      slot_count       = 1U;
      slot_offset      = stream->data_bits;
      break;

    // AC97
    case ARM_SAI_PROTOCOL_AC97:
      if (frame_len == 0U) {
        frame_len = 256;
      } else {
        if (frame_len != 20U) {return ARM_SAI_ERROR_FRAME_LENGHT; }
      }
      bit_order        = ARM_SAI_MSB_FIRST;
      frame_sync_width = 16U;
      frame_sync_pol   = ARM_SAI_FRAME_SYNC_POLARITY_HIGH;
      frame_sync_early = true;
      slot_sz          = 20U;
      slot_count       = 1U;
      slot_offset      = stream->data_bits;

      // Generate Frame Sync
      I2S_BWR_TCR4_FSD(sai->base , 1);
      break;
    default: return ARM_SAI_ERROR_PROTOCOL;
  }

  // Get Bit Order
  switch (bit_order) {
    case ARM_SAI_MSB_FIRST:
      if ((control & ARM_SAI_CONTROL_Msk) == ARM_SAI_CONFIGURE_TX) {
        I2S_BWR_TCR4_MF(sai->base, 1U);
      } else {
        I2S_BWR_RCR4_MF(sai->base, 1U);
      }
      break;
    case ARM_SAI_LSB_FIRST:
      if ((control & ARM_SAI_CONTROL_Msk) == ARM_SAI_CONFIGURE_TX) {
        I2S_BWR_TCR4_MF(sai->base, 0U);
      } else {
        I2S_BWR_RCR4_MF(sai->base, 0U);
      }
      break;
    default: return ARM_SAI_ERROR_BIT_ORDER;
  }

  // Mono mode - not supported
  if (control & ARM_SAI_MONO_MODE) { return ARM_SAI_ERROR_MONO_MODE; }

  // Companding
  if ((control & ARM_SAI_COMPANDING_Msk) != ARM_SAI_COMPANDING_NONE)  { return ARM_SAI_ERROR_COMPANDING; }

  // Clock polarity
  switch (clock_pol) {
    case ARM_SAI_CLOCK_POLARITY_0:
      if ((control & ARM_SAI_CONTROL_Msk) == ARM_SAI_CONFIGURE_TX) {
        I2S_BWR_TCR2_BCP(sai->base, 1U);
      } else {
        I2S_BWR_RCR2_BCP(sai->base, 1U);
      }
      break;
    case ARM_SAI_CLOCK_POLARITY_1:
      if ((control & ARM_SAI_CONTROL_Msk) == ARM_SAI_CONFIGURE_TX) {
        I2S_BWR_TCR2_BCP(sai->base, 0U);
      } else {
        I2S_BWR_RCR2_BCP(sai->base, 0U);
      }
      break;
    default: return ARM_SAI_ERROR_BIT_ORDER;
  }

  // Master clock pin
  SAI_HAL_SetMclkSrc (sai->base, kSaiMclkSourceSysclk);
  switch (control & ARM_SAI_MCLK_PIN_Msk) {
    case ARM_SAI_MCLK_PIN_INACTIVE:
      SAI_HAL_SetMclkDividerCmd (sai->base, true);
      break;
    case ARM_SAI_MCLK_PIN_OUTPUT:
      SAI_HAL_SetMclkDividerCmd (sai->base, true);
      break;
    case ARM_SAI_MCLK_PIN_INPUT:
      SAI_HAL_SetMclkDividerCmd (sai->base, false);
      break;
    default: return ARM_SAI_ERROR_MCLK_PIN;
  }
  SAI_HAL_TxSetBclkSrc (sai->base, kSaiBclkSourceMclkDiv);
  SAI_HAL_RxSetBclkSrc (sai->base, kSaiBclkSourceMclkDiv);
  

  // Frame Sync Width
  // Sync width cannot be longer than data
  if ((control & ARM_SAI_CONTROL_Msk) == ARM_SAI_CONFIGURE_TX) {
    I2S_BWR_TCR4_SYWD (sai->base, frame_sync_width - 1);
  } else {
    I2S_BWR_RCR4_SYWD (sai->base, frame_sync_width - 1);
  }

  // Frame sync polarity
  switch (frame_sync_pol) {
    case ARM_SAI_FRAME_SYNC_POLARITY_HIGH:
      if ((control & ARM_SAI_CONTROL_Msk) == ARM_SAI_CONFIGURE_TX) {
        I2S_BWR_TCR4_FSP (sai->base, 0U);
      } else {
        I2S_BWR_RCR4_FSP (sai->base, 0U);
      }
      break;
    case ARM_SAI_FRAME_SYNC_POLARITY_LOW:
      if ((control & ARM_SAI_CONTROL_Msk) == ARM_SAI_CONFIGURE_TX) {
        I2S_BWR_TCR4_FSP (sai->base, 1U);
      } else {
        I2S_BWR_RCR4_FSP (sai->base, 1U);
      }
      break;
    default: return ARM_SAI_ERROR_FRAME_SYNC_POLARITY;
  }

  // Frame sync early
  if ((control & ARM_SAI_CONTROL_Msk) == ARM_SAI_CONFIGURE_TX) {
    I2S_BWR_TCR4_FSE (sai->base, frame_sync_early);
  } else {
    I2S_BWR_RCR4_FSE (sai->base, frame_sync_early);
  }

  // Slot Count
  if ((control & ARM_SAI_CONTROL_Msk) == ARM_SAI_CONFIGURE_TX) {
    I2S_BWR_TCR4_FRSZ (sai->base, slot_count - 1);
  } else {
    I2S_BWR_RCR4_FRSZ (sai->base, slot_count - 1);
  }

  // Slot size
  if ((control & ARM_SAI_CONTROL_Msk) == ARM_SAI_CONFIGURE_TX) {
    I2S_BWR_TCR5_WNW(sai->base, slot_sz - 1);
    if (protocol != ARM_SAI_PROTOCOL_AC97) {
      I2S_BWR_TCR5_W0W(sai->base, slot_sz - 1);
    } else {
      I2S_BWR_TCR5_W0W(sai->base, 16U - 1);
    }
  } else {
    I2S_BWR_RCR5_WNW(sai->base, slot_sz - 1);
    if (protocol != ARM_SAI_PROTOCOL_AC97) {
      I2S_BWR_RCR5_W0W(sai->base, slot_sz - 1);
    } else {
      I2S_BWR_RCR5_W0W(sai->base, 16U - 1);
    }
  }

  // Slot offset
  if ((control & ARM_SAI_CONTROL_Msk) == ARM_SAI_CONFIGURE_TX) {
    I2S_BWR_TCR5_FBT(sai->base, slot_offset - 1);
  } else {
    I2S_BWR_RCR5_FBT(sai->base, slot_offset - 1);
  }

  // Audio Frequency
  if (master == 1U) {
    // WS and SCK are generated only by master

    val  = ((arg2 & ARM_SAI_MCLK_PRESCALER_Msk) >> ARM_SAI_MCLK_PRESCALER_Pos) + 1U;
    mclk = (arg2 & ARM_SAI_AUDIO_FREQ_Msk) * val;
    val  = CLOCK_SYS_GetSaiFreq (I2S0_IDX, kClockSaiSrcPllClk);

    // Set master clock divider
    SAI_HAL_SetMclkDiv (sai->base, mclk, val);
  }

  // MCLK Prescaler
  val = (((arg2 & ARM_SAI_MCLK_PRESCALER_Msk) >> ARM_SAI_MCLK_PRESCALER_Pos) + 1U) / frame_len;
  if ((control & ARM_SAI_CONTROL_Msk) == ARM_SAI_CONFIGURE_TX) {
    SAI_HAL_TxSetBclkDiv (sai->base, val);
  } else {
    SAI_HAL_RxSetBclkDiv (sai->base, val);
  }

  sai->info->flags |= SAI_FLAG_CONFIGURED;

  return ARM_DRIVER_OK;
}

/**
  \fn          ARM_SAI_STATUS SAI0_GetStatus (void)
  \brief       Get SAI status.
  \return      SAI status \ref ARM_SAI_STATUS
*/
static ARM_SAI_STATUS SAI0_GetStatus (void) {
  ARM_SAI_STATUS status;

  status.frame_error  = sai->info->status.frame_error;
  status.rx_busy      = sai->info->status.rx_busy;
  status.rx_overflow  = sai->info->status.rx_overflow;
  status.tx_busy      = sai->info->status.tx_busy;
  status.tx_underflow = sai->info->status.tx_underflow;

  return status;
}

/**
  \fn          void I2S0_Tx_IRQHandler (void)
  \brief       SAI Interrupt handler.
*/
void I2S0_Tx_IRQHandler (void) {
  uint32_t tcsr, event;
  volatile uint32_t wp, rp;
  uint32_t data;

  event = 0U;
  tcsr = I2S_RD_TCSR(sai->base);

  // FIFO Error - Tx underrun
  if ((tcsr & I2S_TCSR_FEF_MASK) && (tcsr & I2S_TCSR_FEIE_MASK)) {
    // Disable FIFO error interrupt
    SAI_HAL_TxSetIntCmd (sai->base, kSaiIntrequestFIFOError, false);

    sai->info->status.tx_underflow = 1U;
    event |= ARM_SAI_EVENT_TX_UNDERFLOW;
  }

  // Sync error
  if ((tcsr & I2S_TCSR_SEF_MASK) && (tcsr & I2S_TCSR_SEIE_MASK)) {
    sai->info->status.frame_error = 1U;
    event |= ARM_SAI_EVENT_FRAME_ERROR;
  }

  // FIFO request
  if ((tcsr & I2S_TCSR_FRF_MASK) && (tcsr & I2S_TCSR_FRIE_MASK)) {
    if (sai->info->tx.cnt != sai->info->tx.num) {
      // Enable FIFO error interrupt
      SAI_HAL_TxSetIntCmd (sai->base, kSaiIntrequestFIFOError, true);
    }
    while (sai->info->tx.cnt != sai->info->tx.num) {
      wp = I2S_RD_TFR_WFP(sai->base, 0U);
      rp = I2S_RD_TFR_RFP(sai->base, 0U);

      // FIFO is full when read an write pointers are identical, except the MSB
      if (((wp & 0x7U) == ((rp & 0x7U))) && ((wp & 0x8U) != (rp & 0x8U))) { break; }

      if (sai->info->tx.mute) {
        data = mute_data;
      } else {
        data = *(sai->info->tx.buf++);
        if (sai->info->tx.data_bits > 8U) {
          data |= *(sai->info->tx.buf++) << 8;
          if (sai->info->tx.data_bits > 16U) {
            data |= *(sai->info->tx.buf++) << 16;
            data |= *(sai->info->tx.buf++) << 24;
          }
        }
      }
      // Write data to FIFO
      SAI_HAL_SendData (sai->base, 0U, data);
      sai->info->tx.cnt++;
    }

    // Enable FIFO error interrupt
    SAI_HAL_TxSetIntCmd (sai->base, kSaiIntrequestFIFOError, true);
    if (sai->info->tx.cnt == sai->info->tx.num) {
      // Disable FIFO request interrupt
      SAI_HAL_TxSetIntCmd (sai->base, kSaiIntrequestFIFORequest, false);

      // Clear busy flag
      sai->info->status.tx_busy = 0U;

      event |= ARM_SAI_EVENT_SEND_COMPLETE;
    }
  }

  // Signal Events
  if ((event != 0U) && (sai->info->cb_event != NULL)) {
    sai->info->cb_event (event);
  }
}

/**
  \fn          void I2S0_Rx_IRQHandler (void)
  \brief       SAI Interrupt handler.
*/
void I2S0_Rx_IRQHandler (void) {
  uint32_t rcsr, event, wp, rp;
  uint32_t data;

  event = 0U;
  rcsr  = I2S_RD_RCSR(sai->base);

  // FIFO Error - Rx overflow
  if ((rcsr & I2S_RCSR_FEF_MASK) && (rcsr & I2S_RCSR_FEIE_MASK)) {
    // Disable FIFO error interrupt
    SAI_HAL_RxSetIntCmd (sai->base, kSaiIntrequestFIFOError, false);
    sai->info->status.rx_overflow = 1U;
    event |= ARM_SAI_EVENT_RX_OVERFLOW;
  }

  // Sync error
  if ((rcsr & I2S_RCSR_SEF_MASK) && (rcsr & I2S_RCSR_SEIE_MASK)) {
    sai->info->status.frame_error = 1U;
    event |= ARM_SAI_EVENT_FRAME_ERROR;
  }

  // FIFO request
  if ((rcsr & I2S_RCSR_FRF_MASK) && (rcsr & I2S_RCSR_FRIE_MASK)) {
    if (sai->info->rx.cnt != sai->info->rx.num) {
      // Enable FIFO error interrupt
      SAI_HAL_RxSetIntCmd (sai->base, kSaiIntrequestFIFOError, true);
    }
    while (sai->info->rx.cnt != sai->info->rx.num) {
      wp = I2S_RD_RFR_WFP(sai->base, 0U);
      rp = I2S_RD_RFR_RFP(sai->base, 0U);
      // FIFO is empty when read and write pointers are identical
      if ((wp & 0x8U) == (rp & 0x8U)) { break; }

      data = SAI_HAL_ReceiveData (sai->base, 0);

      // Put data into buffer
      *(sai->info->rx.buf++) = (uint8_t)data;
      if (sai->info->rx.data_bits > 8U) {
        *(sai->info->rx.buf++) = (uint8_t)(data >> 8);
        if (sai->info->rx.data_bits > 16U) {
          *(sai->info->rx.buf++) = (uint8_t)(data >> 16);
          *(sai->info->rx.buf++) = (uint8_t)(data >> 24);
        }
      }
      sai->info->rx.cnt++;
    }
    if (sai->info->rx.cnt == sai->info->rx.num) {
      // Disable FIFO request interrupt
      SAI_HAL_RxSetIntCmd (sai->base, kSaiIntrequestFIFORequest, false);

      // Clear busy flag
      sai->info->status.rx_busy = 0U;

      event |= ARM_SAI_EVENT_RECEIVE_COMPLETE;
    }
  }

  // Signal Events
  if ((event != 0U) && (sai->info->cb_event != NULL)) {
    sai->info->cb_event (event);
  }
}

#if ((RTE_SAI0) && (RTE_SAI0_DMA_TX == 1))
static void SAI_TX_DMA_Callback (void *par, edma_chn_status_t status) {

  switch (status) {
    case kEDMAChnNormal:
      // Update count variable
      sai->info->tx.cnt         = sai->info->tx.num;
      // Clear busy flag
      sai->info->status.tx_busy = 0U;
      if (sai->info->cb_event) { sai->info->cb_event (ARM_SAI_EVENT_SEND_COMPLETE); }

      break;
    case kEDMAChnIdle:
      break;

    case kEDMAChnError:
    default:
      DMA_CERR = 1 << sai->tx_dma->ch_info->channel;
      break;
  }
}
#endif

#if ((RTE_SAI0) && (RTE_SAI0_DMA_RX == 1))
static void SAI_RX_DMA_Callback (void *par, edma_chn_status_t status) {

  switch (status) {
    case kEDMAChnNormal:
      // Update count variable
      sai->info->rx.cnt         = sai->info->rx.num;
      // Clear busy flag
      sai->info->status.rx_busy = 0U;
      if (sai->info->cb_event) { sai->info->cb_event (ARM_SAI_EVENT_RECEIVE_COMPLETE); }

      break;
    case kEDMAChnIdle:
      break;

    case kEDMAChnError:
    default:
      DMA_CERR = 1 << sai->rx_dma->ch_info->channel;
      break;
  }
}
#endif

// SAI0 Driver Control Block
ARM_DRIVER_SAI Driver_SAI0 = {
    SAI_GetVersion,
    SAI0_GetCapabilities,
    SAI0_Initialize,
    SAI0_Uninitialize,
    SAI0_PowerControl,
    SAI0_Send,
    SAI0_Receive,
    SAI0_GetTxCount,
    SAI0_GetRxCount,
    SAI0_Control,
    SAI0_GetStatus
};
