/* -----------------------------------------------------------------------------
 * Copyright (c) 2013-2015 ARM Ltd.
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from
 * the use of this software. Permission is granted to anyone to use this
 * software for any purpose, including commercial applications, and to alter
 * it and redistribute it freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software in
 *    a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 *
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 3. This notice may not be removed or altered from any source distribution.
 *
 *
 * $Date:        16. June 2015
 * $Revision:    V1.2
 *
 * Driver:       Driver_SPI0, Driver_SPI1, Driver_SPI2
 * Configured:   via RTE_Device.h configuration file
 * Project:      SPI Driver for Freescale MK64F
 * --------------------------------------------------------------------------
 * Use the following configuration settings in the middleware component
 * to connect to this driver.
 *
 *   Configuration Setting                 Value   SPI Interface
 *   ---------------------                 -----   -------------
 *   Connect to hardware via Driver_SPI# = 0       use SPI0
 *   Connect to hardware via Driver_SPI# = 1       use SPI1
 *   Connect to hardware via Driver_SPI# = 2       use SPI2
 * -------------------------------------------------------------------------- */

/* History:
 *  Version 1.2
 *    Updated initialization, uninitialization and power procedures
 *    Corrected status bit-field handling, to prevent race conditions.
 *  Version 1.1
 *    Updated for KSDK 1.2.0
 *  Version 1.0
 *    Initial release
 */

#include "SPI_MK64F.h"

#define ARM_SPI_DRV_VERSION ARM_DRIVER_VERSION_MAJOR_MINOR(1,2)

// Driver Version
static const ARM_DRIVER_VERSION DriverVersion = { ARM_SPI_API_VERSION, ARM_SPI_DRV_VERSION };

// Driver Capabilities
static const ARM_SPI_CAPABILITIES DriverCapabilities = {
  0,  /* Simplex Mode (Master and Slave) */
  0,  /* TI Synchronous Serial Interface */
  0,  /* Microwire Interface */
  0   /* Signal Mode Fault event: \ref ARM_SPI_EVENT_MODE_FAULT */
};


#if defined(RTE_SPI0) && (RTE_SPI0 != 0)
/* SPI0 Information (Run-Time) */
static SPI_CTRL SPI0_Ctrl;

/* SPI0 Resources */
static SPI_RESOURCES SPI0_Resources = {
  SPI0,
  &SPI0_Ctrl,
  SPI0_IRQn,
  kSimClockGateSpi0,
  (dspi_which_pcs_config_t)(1 << RTE_SPI0_PCS)
};
#endif /* RTE_SPI0 */


#if defined(RTE_SPI1) && (RTE_SPI1 != 0)
/* SPI1 Information (Run-Time) */
static SPI_CTRL SPI1_Ctrl;

/* SPI0 Resources */
static SPI_RESOURCES SPI1_Resources = {
  SPI1,
  &SPI1_Ctrl,
  SPI1_IRQn,
  kSimClockGateSpi1,
  (dspi_which_pcs_config_t)(1 << RTE_SPI1_PCS)
};
#endif /* RTE_SPI1 */


#if defined(RTE_SPI2) && (RTE_SPI2 != 0)
/* SPI2 Information (Run-Time) */
static SPI_CTRL SPI2_Ctrl;

/* SPI0 Resources */
static SPI_RESOURCES SPI2_Resources = {
  SPI2,
  &SPI2_Ctrl,
  SPI2_IRQn,
  kSimClockGateSpi2,
  (dspi_which_pcs_config_t)(1 << RTE_SPI2_PCS)
};
#endif /* RTE_SPI2 */


/* Interrupt handler prototype */
static void SPI_IRQHandler (const SPI_RESOURCES *spi);


/**
  \fn          ARM_DRIVER_VERSION SPIX_GetVersion (void)
  \brief       Get SPI driver version.
  \return      \ref ARM_DRV_VERSION
*/
static ARM_DRIVER_VERSION SPIX_GetVersion (void) {
  return DriverVersion;
}


/**
  \fn          ARM_SPI_CAPABILITIES SPI_GetCapabilities (void)
  \brief       Get driver capabilities.
  \return      \ref ARM_SPI_CAPABILITIES
*/
static ARM_SPI_CAPABILITIES SPIX_GetCapabilities (void) {
  return DriverCapabilities;
}


/**
  \fn          int32_t SPI_Initialize (ARM_SPI_SignalEvent_t cb_event, const SPI_RESOURCES *spi)
  \brief       Initialize SPI Interface.
  \param[in]   cb_event  Pointer to \ref ARM_SPI_SignalEvent
  \param[in]   spi       Pointer to SPI resources
  \return      \ref execution_status
*/
static int32_t SPI_Initialize (ARM_SPI_SignalEvent_t cb_event, const SPI_RESOURCES *spi) {

  if (spi->ctrl->flags & SPI_INIT)  { return ARM_DRIVER_OK; }

  /* Initialize SPI Run-Time Resources */
  spi->ctrl->cb_event          = cb_event;
  spi->ctrl->status.busy       = 0U;
  spi->ctrl->status.data_lost  = 0U;
  spi->ctrl->status.mode_fault = 0U;

  spi->ctrl->flags = 0U;

  /* Reset control structure */
  memset(&spi->ctrl->xfer, 0, sizeof(SPI_TRANSFER_INFO));

  spi->ctrl->flags |= SPI_INIT;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t SPI_Uninitialize (const SPI_RESOURCES *spi)
  \brief       De-initialize SPI Interface.
  \param[in]   spi  Pointer to SPI resources
  \return      \ref execution_status
*/
static int32_t SPI_Uninitialize (const SPI_RESOURCES *spi) {
  spi->ctrl->flags = 0U;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t SPI_PowerControl (ARM_POWER_STATE state, const SPI_RESOURCES *spi)
  \brief       Control SPI Interface Power.
  \param[in]   state  Power state
  \param[in]   spi    Pointer to SPI resources
  \return      \ref execution_status
*/
static int32_t SPI_PowerControl (ARM_POWER_STATE state, const SPI_RESOURCES *spi) {
  dspi_command_config_t cmd;

  switch (state) {
    case ARM_POWER_OFF:
      spi->ctrl->flags &= ~(SPI_POWER | SPI_SETUP);

      /* Disable SPI interrupt requests in NVIC */
      NVIC_DisableIRQ (spi->irq_num);

      /* Stop transfer */
      DSPI_HAL_StopTransfer (spi->base);

      /* Flush TX and RX FIFO */
      DSPI_HAL_SetFlushFifoCmd (spi->base, true, true);

      spi->ctrl->status.busy       = 0U;
      spi->ctrl->status.data_lost  = 0U;
      spi->ctrl->status.mode_fault = 0U;

      spi->ctrl->flags = SPI_INIT;

      /* Disable SPI peripheral */
      DSPI_HAL_Disable (spi->base);

      /* Disable SPI peripheral clock */
      SIM_HAL_DisableClock (SIM, spi->clock_gate);
      break;

    case ARM_POWER_FULL:
      if ((spi->ctrl->flags & SPI_POWER) == 0U) {

        spi->ctrl->status.busy       = 0U;
        spi->ctrl->status.data_lost  = 0U;
        spi->ctrl->status.mode_fault = 0U;

        /* Enable SPI peripheral clock */
        SIM_HAL_EnableClock (SIM, spi->clock_gate);

        /* Enable SPI interrupt requests in NVIC */
        NVIC_EnableIRQ (spi->irq_num);

        /* Initialize peripheral to known state */
        DSPI_HAL_Init (spi->base);

        /* Enable DSPI module */
        DSPI_HAL_Enable (spi->base);

        /* Flush TX and RX FIFO */
        DSPI_HAL_SetFlushFifoCmd (spi->base, true, true);

        /* Configure interrupt requests */
        DSPI_HAL_SetIntMode (spi->base, kDspiEndOfQueue,        true);
        DSPI_HAL_SetIntMode (spi->base, kDspiTxComplete,        false);
        DSPI_HAL_SetIntMode (spi->base, kDspiTxFifoFillRequest, false);
        DSPI_HAL_SetIntMode (spi->base, kDspiTxFifoUnderflow,   true);
        DSPI_HAL_SetIntMode (spi->base, kDspiRxFifoOverflow,     true);
        DSPI_HAL_SetIntMode (spi->base, kDspiRxFifoDrainRequest, true);

        DSPI_HAL_SetTxFifoFillDmaIntMode (spi->base, kDspiGenerateIntReq, false);
        DSPI_HAL_SetRxFifoDrainDmaIntMode(spi->base, kDspiGenerateIntReq, true);

        /* Chip select is active low */
        DSPI_HAL_SetPcsPolarityMode (spi->base, spi->pcs, kDspiPcs_ActiveLow);

        /* Prepare Tx command for master transfers */
        cmd.isChipSelectContinuous = false;
        cmd.whichCtar              = kDspiCtar0;
        cmd.whichPcs               = spi->pcs;
        cmd.isEndOfQueue           = false;
        cmd.clearTransferCount     = false;

        spi->ctrl->xfer.tx_cmd = DSPI_HAL_GetFormattedCommand (spi->base, &cmd);

        /* Ready for operation */
        spi->ctrl->flags |= SPI_POWER;
      }
      break;

    case ARM_POWER_LOW:
    default:
      return ARM_DRIVER_ERROR_UNSUPPORTED;
  }
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t SPI_Send (const void *data, uint32_t num, const SPI_RESOURCES *spi)
  \brief       Start sending data to SPI transmitter.
  \param[in]   data  Pointer to buffer with data to send to SPI transmitter
  \param[in]   num   Number of data items to send
  \param[in]   spi   Pointer to SPI resources
  \return      \ref execution_status
*/
static int32_t SPI_Send (const void *data, uint32_t num, const SPI_RESOURCES *spi) {

  if ((data == NULL) || (num == 0U))        { return ARM_DRIVER_ERROR_PARAMETER; }
  if ((spi->ctrl->flags & SPI_SETUP) == 0U) { return ARM_DRIVER_ERROR; }
  if ( spi->ctrl->status.busy)              { return ARM_DRIVER_ERROR_BUSY; }

  spi->ctrl->status.busy       = 1U;
  spi->ctrl->status.data_lost  = 0U;
  spi->ctrl->status.mode_fault = 0U;

  spi->ctrl->xfer.rx_buf = NULL;
  spi->ctrl->xfer.tx_buf = (uint8_t *)data;
  spi->ctrl->xfer.num    = num;
  spi->ctrl->xfer.rx_cnt = 0U;
  spi->ctrl->xfer.tx_cnt = 0U;

  /* Enable FIFO Fill request */
  DSPI_HAL_SetIntMode (spi->base, kDspiTxFifoFillRequest, true);

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t SPI_Receive (void *data, uint32_t num, const SPI_RESOURCES *spi)
  \brief       Start receiving data from SPI receiver.
  \param[out]  data  Pointer to buffer for data to receive from SPI receiver
  \param[in]   num   Number of data items to receive
  \param[in]   spi   Pointer to SPI resources
  \return      \ref execution_status
*/
static int32_t SPI_Receive (void *data, uint32_t num, const SPI_RESOURCES *spi) {

  if ((data == NULL) || (num == 0U))        { return ARM_DRIVER_ERROR_PARAMETER; }
  if ((spi->ctrl->flags & SPI_SETUP) == 0U) { return ARM_DRIVER_ERROR; }
  if ( spi->ctrl->status.busy)              { return ARM_DRIVER_ERROR_BUSY; }

  spi->ctrl->status.busy       = 1U;
  spi->ctrl->status.data_lost  = 0U;
  spi->ctrl->status.mode_fault = 0U;

  spi->ctrl->xfer.rx_buf = (uint8_t *)data;
  spi->ctrl->xfer.tx_buf = NULL;
  spi->ctrl->xfer.num    = num;
  spi->ctrl->xfer.rx_cnt = 0U;
  spi->ctrl->xfer.tx_cnt = 0U;

  /* Enable FIFO Fill request */
  DSPI_HAL_SetIntMode (spi->base, kDspiTxFifoFillRequest, true);

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t SPI_Transfer (const void *data_out, void *data_in, uint32_t num, const SPI_RESOURCES *spi)
  \brief       Start sending/receiving data to/from SPI transmitter/receiver.
  \param[in]   data_out  Pointer to buffer with data to send to SPI transmitter
  \param[out]  data_in   Pointer to buffer for data to receive from SPI receiver
  \param[in]   num       Number of data items to transfer
  \param[in]   spi       Pointer to SPI resources
  \return      \ref execution_status
*/
static int32_t SPI_Transfer (const void *data_out, void *data_in, uint32_t num, const SPI_RESOURCES *spi) {

  if ((data_out == NULL) || (data_in == NULL) || (num == 0U)) { return ARM_DRIVER_ERROR_PARAMETER; }
  if ((spi->ctrl->flags & SPI_SETUP) == 0U)                   { return ARM_DRIVER_ERROR; }
  if ( spi->ctrl->status.busy)                                { return ARM_DRIVER_ERROR_BUSY; }

  spi->ctrl->status.busy       = 1U;
  spi->ctrl->status.data_lost  = 0U;
  spi->ctrl->status.mode_fault = 0U;

  spi->ctrl->xfer.rx_buf = (uint8_t *)data_in;
  spi->ctrl->xfer.tx_buf = (uint8_t *)data_out;
  spi->ctrl->xfer.num    = num;
  spi->ctrl->xfer.rx_cnt = 0U;
  spi->ctrl->xfer.tx_cnt = 0U;

  /* Enable FIFO Fill request */
  DSPI_HAL_SetIntMode (spi->base, kDspiTxFifoFillRequest, true);

  return ARM_DRIVER_OK;
}

/**
  \fn          uint32_t SPI_GetDataCount (const SPI_RESOURCES *spi)
  \brief       Get transferred data count.
  \param[in]   spi  Pointer to SPI resources
  \return      number of data items transferred
*/
static uint32_t SPI_GetDataCount (const SPI_RESOURCES *spi) {
  return (spi->ctrl->xfer.rx_cnt);
}

/**
  \fn          int32_t SPI_Control (uint32_t control, uint32_t arg, const SPI_RESOURCES *spi)
  \brief       Control SPI Interface.
  \param[in]   control  operation
  \param[in]   arg      argument of operation (optional)
  \param[in]   spi      Pointer to SPI resources
  \return      \ref execution_status
*/
static int32_t SPI_Control (uint32_t control, uint32_t arg, const SPI_RESOURCES *spi) {
  dspi_data_format_config_t mode_cfg;
  dspi_master_slave_mode_t  mode_typ;
  bool master;
  uint32_t bps, idx;

  if ((spi->ctrl->flags & SPI_POWER) == 0U) { return ARM_DRIVER_ERROR; }

  if ((control & ARM_SPI_CONTROL_Msk) == ARM_SPI_ABORT_TRANSFER) {
    /* Abort data transfer */
    DSPI_HAL_StopTransfer (spi->base);

    spi->ctrl->xfer.rx_cnt = 0U;
    spi->ctrl->xfer.tx_cnt = 0U;
    spi->ctrl->status.busy = 0U;

    DSPI_HAL_StartTransfer (spi->base);
    return ARM_DRIVER_OK;
  }

  if (spi->ctrl->status.busy) { return ARM_DRIVER_ERROR_BUSY; }

  switch (control & ARM_SPI_CONTROL_Msk) {
    case ARM_SPI_MODE_INACTIVE:
      DSPI_HAL_StopTransfer (spi->base);
      return ARM_DRIVER_OK;

    case ARM_SPI_MODE_MASTER:           /* Master: Output on MOSI, Input on MISO */
      mode_typ = kDspiMaster;
      master   = true;
      break;

    case ARM_SPI_MODE_SLAVE:            /* Slave: Output on MISO, Input on MOSI */
      mode_typ = kDspiSlave;
      master   = false;
      break;

    case ARM_SPI_MODE_MASTER_SIMPLEX:   /* Master: Output and Input on MOSI */
    case ARM_SPI_MODE_SLAVE_SIMPLEX:    /* Slave:  Output and Input on MISO */
      return ARM_SPI_ERROR_MODE;

    case ARM_SPI_SET_BUS_SPEED:         /* Set the bus speed in bps */
      DSPI_HAL_StopTransfer  (spi->base);
      DSPI_HAL_SetBaudRate   (spi->base, kDspiCtar0, arg, CLOCK_SYS_GetSpiFreq(0));
      DSPI_HAL_StartTransfer (spi->base);
      return ARM_DRIVER_OK;

    case ARM_SPI_GET_BUS_SPEED:         /* Get the bus speed in bps */
      if      (spi->base == SPI0) { idx = SPI0_IDX; }
      else if (spi->base == SPI1) { idx = SPI1_IDX; }
      else if (spi->base == SPI2) { idx = SPI2_IDX; }
      else { return ARM_DRIVER_ERROR; }
      bps  = CLOCK_SYS_GetSpiFreq(0) * (1 + SPI_BWR_CTAR_DBR (spi->base, idx, kDspiCtar0));
      bps /= s_baudratePrescaler[SPI_BWR_CTAR_PBR (spi->base, idx, kDspiCtar0)];
      bps /= s_baudrateScaler[SPI_BWR_CTAR_BR (spi->base, idx, kDspiCtar0)];
      return (bps);

    case ARM_SPI_SET_DEFAULT_TX_VALUE:  /* Set the default transmission value */
      spi->ctrl->xfer.def_val = (uint16_t)(arg & 0xFFFFU);
      return ARM_DRIVER_OK;

    case ARM_SPI_CONTROL_SS:            /* Control the Slave Select signal */
    default:
      return ARM_DRIVER_ERROR_UNSUPPORTED;
  }

  /* Set the frame format */
  switch (control & ARM_SPI_FRAME_FORMAT_Msk) {
    case ARM_SPI_CPOL0_CPHA0:
      mode_cfg.clkPolarity = kDspiClockPolarity_ActiveHigh;
      mode_cfg.clkPhase    = kDspiClockPhase_FirstEdge;
      break;

    case ARM_SPI_CPOL0_CPHA1:
      mode_cfg.clkPolarity = kDspiClockPolarity_ActiveHigh;
      mode_cfg.clkPhase    = kDspiClockPhase_SecondEdge;
      break;

    case ARM_SPI_CPOL1_CPHA0:
      mode_cfg.clkPolarity = kDspiClockPolarity_ActiveLow;
      mode_cfg.clkPhase    = kDspiClockPhase_FirstEdge;
      break;

    case ARM_SPI_CPOL1_CPHA1:
      mode_cfg.clkPolarity = kDspiClockPolarity_ActiveLow;
      mode_cfg.clkPhase    = kDspiClockPhase_SecondEdge;
      break;

    case ARM_SPI_TI_SSI:
    case ARM_SPI_MICROWIRE:
    default:
      return ARM_SPI_ERROR_FRAME_FORMAT;
  }

  /* Set the number of bits per SPI frame */
  mode_cfg.bitsPerFrame = (control & ARM_SPI_DATA_BITS_Msk) >> ARM_SPI_DATA_BITS_Pos;

  if ((mode_cfg.bitsPerFrame < 4) || (mode_cfg.bitsPerFrame >16)) {
    return ARM_SPI_ERROR_DATA_BITS;
  }
  else if (mode_cfg.bitsPerFrame > 8) {
    /* Two bytes per frame */
    spi->ctrl->flags |= SPI_MBYTE;
  }
  else {
    /* One byte per frame */
    spi->ctrl->flags &= ~SPI_MBYTE;
  }

  /* Set the bit order */
  if ((control & ARM_SPI_BIT_ORDER_Msk) == ARM_SPI_LSB_MSB) {
    mode_cfg.direction = kDspiLsbFirst;
  }
  else {
    mode_cfg.direction = kDspiMsbFirst;
  }

  if (master) {
    switch (control & ARM_SPI_SS_MASTER_MODE_Msk) {
      case ARM_SPI_SS_MASTER_HW_OUTPUT:   /* Master SS is hardware controlled output */
        break;

      case ARM_SPI_SS_MASTER_UNUSED:      /* Master SS is not used */
      case ARM_SPI_SS_MASTER_SW:          /* Master SS is software controlled */
      case ARM_SPI_SS_MASTER_HW_INPUT:    /* Master SS is hardware monitored input */
      default:
        return ARM_SPI_ERROR_SS_MODE;
    }
  }
  else {
    switch (control & ARM_SPI_SS_SLAVE_MODE_Msk) {
      case ARM_SPI_SS_SLAVE_HW:           /* Hardware monitored SS for the slave */
        break;

      case ARM_SPI_SS_SLAVE_SW:           /* Software controlled SS for the slave */
      default:
        return ARM_SPI_ERROR_SS_MODE;
    }
  }

  /* Set configuration */
  DSPI_HAL_StopTransfer       (spi->base);
  DSPI_HAL_SetMasterSlaveMode (spi->base, mode_typ);
  DSPI_HAL_SetDataFormat      (spi->base, kDspiCtar0 , &mode_cfg);
  DSPI_HAL_SetBaudRate        (spi->base, kDspiCtar0, arg, CLOCK_SYS_GetSpiFreq(0));
  DSPI_HAL_StartTransfer      (spi->base);

  spi->ctrl->flags |= SPI_SETUP;

  return ARM_DRIVER_OK;
}

/**
  \fn          ARM_SPI_STATUS SPI_GetStatus (const SPI_RESOURCES *spi)
  \brief       Get SPI status.
  \param[in]   spi  Pointer to SPI resources
  \return      SPI status \ref ARM_SPI_STATUS
*/
static ARM_SPI_STATUS SPI_GetStatus (const SPI_RESOURCES *spi) {
  ARM_SPI_STATUS status;

  status.busy       = spi->ctrl->status.busy;
  status.data_lost  = spi->ctrl->status.data_lost;
  status.mode_fault = spi->ctrl->status.mode_fault;

  return (status);
}


/**
  \fn          void SPI_IRQHandler (SPI_RESOURCES *spi)
  \brief       SPI Interrupt handler.
  \param[in]   spi  Pointer to SPI resources
*/
void SPI_IRQHandler (const SPI_RESOURCES *spi) {
  SPI_TRANSFER_INFO *tr = &spi->ctrl->xfer;
  uint32_t event;
  uint32_t data;

  event = 0U;

  if (DSPI_HAL_GetStatusFlag (spi->base, kDspiRxFifoOverflow)) {
    DSPI_HAL_ClearStatusFlag (spi->base, kDspiRxFifoOverflow);
    spi->ctrl->status.data_lost = 1;
    event |= ARM_SPI_EVENT_DATA_LOST;
  }
  else if (DSPI_HAL_GetStatusFlag (spi->base, kDspiTxFifoUnderflow)) {
    DSPI_HAL_ClearStatusFlag (spi->base, kDspiTxFifoUnderflow);
    spi->ctrl->status.data_lost = 1;
    event |= ARM_SPI_EVENT_DATA_LOST;
  }
  else if (DSPI_HAL_GetStatusFlag (spi->base, kDspiEndOfQueue)) {
    DSPI_HAL_ClearStatusFlag (spi->base, kDspiEndOfQueue);
  }
  else {
    if (DSPI_HAL_GetStatusFlag(spi->base, kDspiTxFifoFillRequest)) {
      if (tr->tx_cnt < tr->num) {
        /* Fill up Tx FIFO */
        if (tr->tx_buf != NULL) {
          data = *tr->tx_buf++;

          if (spi->ctrl->flags & SPI_MBYTE) {
            data |= *tr->tx_buf++ << 8;
          }
        }
        else {
          data = tr->def_val;
        }
        tr->tx_cnt++;

        /* Write data to Tx FIFO */
        if (DSPI_HAL_IsMaster (spi->base)) {
          data |= tr->tx_cmd;

          if ((tr->num - tr->tx_cnt) == 0) {
            /* Last frame to send, set end of queue */
            data |= SPI_PUSHR_EOQ_MASK;
          }

          DSPI_HAL_WriteCmdDataMastermode(spi->base, data);
        }
        else {
          DSPI_HAL_WriteDataSlavemode(spi->base, data);
        }

        /* Clear request, it will be re-set in case if Tx FIFO not full */
        DSPI_HAL_ClearStatusFlag (spi->base, kDspiTxFifoFillRequest);

        if (tr->tx_cnt == tr->num) {
          /* Disable FIFO Fill interrupt */
          DSPI_HAL_SetIntMode (spi->base, kDspiTxFifoFillRequest, false);
        }
      }
    }

    if (DSPI_HAL_GetStatusFlag(spi->base, kDspiRxFifoDrainRequest)) {
      /* Fetch received data */
      data = DSPI_HAL_ReadData (spi->base);

      /* Clear request, it will be re-set in case of more data in FIFO */
      DSPI_HAL_ClearStatusFlag (spi->base, kDspiRxFifoDrainRequest);

      if (tr->rx_cnt < tr->num) {
        if (tr->rx_buf != NULL) {
          *(tr->rx_buf++) = (uint8_t)data;

          if (spi->ctrl->flags & SPI_MBYTE) {
            *(tr->rx_buf++) = (uint8_t)(data >> 8);
          }
        }
        tr->rx_cnt++;

        if (tr->rx_cnt == tr->num) {
          /* All frames transfered */
          tr->num = 0U;
          spi->ctrl->status.busy = 0U;

          event |= ARM_SPI_EVENT_TRANSFER_COMPLETE;
        }
      }
      else {
        spi->ctrl->status.data_lost = 1;
        event |= ARM_SPI_EVENT_DATA_LOST;
      }
    }
  }

  /* Send event */
  if ((event != 0U) && (spi->ctrl->cb_event != NULL)) {
    spi->ctrl->cb_event(event);
  }
}


#if defined(RTE_SPI0) && (RTE_SPI0 != 0)
/* SPI0 Wrapper Functions */
static int32_t        SPI0_Initialize          (ARM_SPI_SignalEvent_t pSignalEvent)                { return SPI_Initialize (pSignalEvent, &SPI0_Resources); }
static int32_t        SPI0_Uninitialize        (void)                                              { return SPI_Uninitialize (&SPI0_Resources); }
static int32_t        SPI0_PowerControl        (ARM_POWER_STATE state)                             { return SPI_PowerControl (state, &SPI0_Resources); }
static int32_t        SPI0_Send                (const void *data, uint32_t num)                    { return SPI_Send (data, num, &SPI0_Resources); }
static int32_t        SPI0_Receive             (void *data, uint32_t num)                          { return SPI_Receive (data, num, &SPI0_Resources); }
static int32_t        SPI0_Transfer            (const void *data_out, void *data_in, uint32_t num) { return SPI_Transfer (data_out, data_in, num, &SPI0_Resources); }
static uint32_t       SPI0_GetDataCount        (void)                                              { return SPI_GetDataCount (&SPI0_Resources); }
static int32_t        SPI0_Control             (uint32_t control, uint32_t arg)                    { return SPI_Control (control, arg, &SPI0_Resources); }
static ARM_SPI_STATUS SPI0_GetStatus           (void)                                              { return SPI_GetStatus (&SPI0_Resources); }
       void           SPI0_IRQHandler          (void)                                              {        SPI_IRQHandler (&SPI0_Resources); }

/* Driver_SPI0 */
ARM_DRIVER_SPI Driver_SPI0 = {
  SPIX_GetVersion,
  SPIX_GetCapabilities,
  SPI0_Initialize,
  SPI0_Uninitialize,
  SPI0_PowerControl,
  SPI0_Send,
  SPI0_Receive,
  SPI0_Transfer,
  SPI0_GetDataCount,
  SPI0_Control,
  SPI0_GetStatus
};
#endif

#if defined(RTE_SPI1) && (RTE_SPI1 != 0)
/* SPI1 Wrapper Functions */
static int32_t        SPI1_Initialize          (ARM_SPI_SignalEvent_t pSignalEvent)                { return SPI_Initialize (pSignalEvent, &SPI1_Resources); }
static int32_t        SPI1_Uninitialize        (void)                                              { return SPI_Uninitialize (&SPI1_Resources); }
static int32_t        SPI1_PowerControl        (ARM_POWER_STATE state)                             { return SPI_PowerControl (state, &SPI1_Resources); }
static int32_t        SPI1_Send                (const void *data, uint32_t num)                    { return SPI_Send (data, num, &SPI1_Resources); }
static int32_t        SPI1_Receive             (void *data, uint32_t num)                          { return SPI_Receive (data, num, &SPI1_Resources); }
static int32_t        SPI1_Transfer            (const void *data_out, void *data_in, uint32_t num) { return SPI_Transfer (data_out, data_in, num, &SPI1_Resources); }
static uint32_t       SPI1_GetDataCount        (void)                                              { return SPI_GetDataCount (&SPI1_Resources); }
static int32_t        SPI1_Control             (uint32_t control, uint32_t arg)                    { return SPI_Control (control, arg, &SPI1_Resources); }
static ARM_SPI_STATUS SPI1_GetStatus           (void)                                              { return SPI_GetStatus (&SPI1_Resources); }
       void           SPI1_IRQHandler          (void)                                              {        SPI_IRQHandler (&SPI1_Resources); }

/* Driver_SPI1 */
ARM_DRIVER_SPI Driver_SPI1 = {
  SPIX_GetVersion,
  SPIX_GetCapabilities,
  SPI1_Initialize,
  SPI1_Uninitialize,
  SPI1_PowerControl,
  SPI1_Send,
  SPI1_Receive,
  SPI1_Transfer,
  SPI1_GetDataCount,
  SPI1_Control,
  SPI1_GetStatus
};
#endif

#if defined(RTE_SPI2) && (RTE_SPI2 != 0)
/* SPI2 Wrapper Functions */
static int32_t        SPI2_Initialize          (ARM_SPI_SignalEvent_t pSignalEvent)                { return SPI_Initialize (pSignalEvent, &SPI2_Resources); }
static int32_t        SPI2_Uninitialize        (void)                                              { return SPI_Uninitialize (&SPI2_Resources); }
static int32_t        SPI2_PowerControl        (ARM_POWER_STATE state)                             { return SPI_PowerControl (state, &SPI2_Resources); }
static int32_t        SPI2_Send                (const void *data, uint32_t num)                    { return SPI_Send (data, num, &SPI2_Resources); }
static int32_t        SPI2_Receive             (void *data, uint32_t num)                          { return SPI_Receive (data, num, &SPI2_Resources); }
static int32_t        SPI2_Transfer            (const void *data_out, void *data_in, uint32_t num) { return SPI_Transfer (data_out, data_in, num, &SPI2_Resources); }
static uint32_t       SPI2_GetDataCount        (void)                                              { return SPI_GetDataCount (&SPI2_Resources); }
static int32_t        SPI2_Control             (uint32_t control, uint32_t arg)                    { return SPI_Control (control, arg, &SPI2_Resources); }
static ARM_SPI_STATUS SPI2_GetStatus           (void)                                              { return SPI_GetStatus (&SPI2_Resources); }
       void           SPI2_IRQHandler          (void)                                              {        SPI_IRQHandler (&SPI2_Resources); }

/* Driver_SPI2 */
ARM_DRIVER_SPI Driver_SPI2 = {
  SPIX_GetVersion,
  SPIX_GetCapabilities,
  SPI2_Initialize,
  SPI2_Uninitialize,
  SPI2_PowerControl,
  SPI2_Send,
  SPI2_Receive,
  SPI2_Transfer,
  SPI2_GetDataCount,
  SPI2_Control,
  SPI2_GetStatus
};
#endif
