/* -----------------------------------------------------------------------------
 * Copyright (c) 2013-2015 ARM Ltd.
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from
 * the use of this software. Permission is granted to anyone to use this
 * software for any purpose, including commercial applications, and to alter
 * it and redistribute it freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software in
 *    a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 *
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 3. This notice may not be removed or altered from any source distribution.
 *
 *
 * $Date:        15. June 2015
 * $Revision:    V1.2
 *
 * Driver:       Driver_USART0, Driver_USART1, Driver_USART2, Driver_USART3
 * Configured:   via RTE_Device.h configuration file
 * Project:      USART Driver for Freescale MK64F
 * --------------------------------------------------------------------------
 * Use the following configuration settings in the middleware component
 * to connect to this driver.
 *
 *   Configuration Setting                   Value   UART Interface
 *   ---------------------                   -----   --------------
 *   Connect to hardware via Driver_USART# = 0       use UART0
 *   Connect to hardware via Driver_USART# = 1       use UART1
 *   Connect to hardware via Driver_USART# = 2       use UART2
 *   Connect to hardware via Driver_USART# = 3       use UART3
 *   Connect to hardware via Driver_USART# = 4       use UART4
 *   Connect to hardware via Driver_USART# = 5       use UART5
 * -------------------------------------------------------------------------- */

/* History:
 *  Version 1.2
 *    - PowerControl for Power OFF and Uninitialize functions made unconditional.
 *    - Corrected status bit-field handling, to prevent race conditions.
 *  Version 1.1
 *    Updated for KSDK 1.2.0
 *  Version 1.0
 *    Initial release
 */
#include "UART_MK64F.h"

#define ARM_USART_DRV_VERSION ARM_DRIVER_VERSION_MAJOR_MINOR(1,2)

// Driver Version
static const ARM_DRIVER_VERSION usart_driver_version = { ARM_USART_API_VERSION, ARM_USART_DRV_VERSION };

#if (((RTE_UART0) && (RTE_UART0_DMA_TX == 1)) || \
     ((RTE_UART1) && (RTE_UART1_DMA_TX == 1)) || \
     ((RTE_UART2) && (RTE_UART2_DMA_TX == 1)) || \
     ((RTE_UART3) && (RTE_UART3_DMA_TX == 1)) || \
     ((RTE_UART4) && (RTE_UART4_DMA_TX == 1)) || \
     ((RTE_UART5) && (RTE_UART5_DMA_TX == 1)))
static void UARTx_TX_DMA_Callback (void *par, edma_chn_status_t ch);
#endif

#if (((RTE_UART0) && (RTE_UART0_DMA_RX == 1)) || \
     ((RTE_UART1) && (RTE_UART1_DMA_RX == 1)) || \
     ((RTE_UART2) && (RTE_UART2_DMA_RX == 1)) || \
     ((RTE_UART3) && (RTE_UART3_DMA_RX == 1)) || \
     ((RTE_UART4) && (RTE_UART4_DMA_RX == 1)) || \
     ((RTE_UART5) && (RTE_UART5_DMA_RX == 1)))
static void UARTx_RX_DMA_Callback (void *par, edma_chn_status_t ch);
#endif

#if (RTE_UART0)
// Capabilities
static const ARM_USART_CAPABILITIES UART1_Capabilities = {
  1,  // supports UART (Asynchronous) mode
  0,  // supports Synchronous Master mode
  0,  // supports Synchronous Slave mode
  1,  // supports UART Single-wire mode
  1,  // supports UART IrDA mode
  1,  // supports UART Smart Card mode
  0,  // Smart Card Clock generator
  1,  // RTS Flow Control available
  1,  // CTS Flow Control available
  1,  // Transmit completed event: \ref ARM_USART_EVENT_TX_COMPLETE
  1,  // Signal receive character timeout event: \ref ARM_USART_EVENT_RX_TIMEOUT
  1,  // RTS Line: 0=not available, 1=available
  1,  // CTS Line: 0=not available, 1=available
  0,  // DTR Line: 0=not available, 1=available
  0,  // DSR Line: 0=not available, 1=available
  0,  // DCD Line: 0=not available, 1=available
  0,  // RI Line: 0=not available, 1=available
  0,  // Signal CTS change event: \ref ARM_USART_EVENT_CTS
  0,  // Signal DSR change event: \ref ARM_USART_EVENT_DSR
  0,  // Signal DCD change event: \ref ARM_USART_EVENT_DCD
  0,  // Signal RI change event: \ref ARM_USART_EVENT_RI
};
#endif

#if (RTE_UART1 || RTE_UART2 || RTE_UART3 || RTE_UART4)
static const ARM_USART_CAPABILITIES UART_Capabilities = {
  1,  // supports UART (Asynchronous) mode
  0,  // supports Synchronous Master mode
  0,  // supports Synchronous Slave mode
  1,  // supports UART Single-wire mode
  1,  // supports UART IrDA mode
  0,  // supports UART Smart Card mode
  0,  // Smart Card Clock generator
  1,  // RTS Flow Control available
  1,  // CTS Flow Control available
  1,  // Transmit completed event: \ref ARM_USART_EVENT_TX_COMPLETE
  1,  // Signal receive character timeout event: \ref ARM_USART_EVENT_RX_TIMEOUT
  1,  // RTS Line: 0=not available, 1=available
  1,  // CTS Line: 0=not available, 1=available
  0,  // DTR Line: 0=not available, 1=available
  0,  // DSR Line: 0=not available, 1=available
  0,  // DCD Line: 0=not available, 1=available
  0,  // RI Line: 0=not available, 1=available
  0,  // Signal CTS change event: \ref ARM_USART_EVENT_CTS
  0,  // Signal DSR change event: \ref ARM_USART_EVENT_DSR
  0,  // Signal DCD change event: \ref ARM_USART_EVENT_DCD
  0,  // Signal RI change event: \ref ARM_USART_EVENT_RI
};
#endif

#if (RTE_UART0)

// FIFO watermark must be less than FIFO size
#ifndef UART0_TX_FIFO_WATERMARK
#define UART0_TX_FIFO_WATERMARK (1U)
#endif

// FIFO watermark must be less than FIFO size
#ifndef UART0_RX_FIFO_WATERMARK
#define UART0_RX_FIFO_WATERMARK (1U)
#endif

// UART0 DMA Configuration
#if (RTE_UART0_DMA_TX)
edma_chn_state_t       UART0_DMA_TxState;
edma_transfer_config_t UART0_DMA_TxCfg;

static UART_DMA UART0_DMA_Tx = {
  kDmaRequestMux0UART0Tx,
  &UART0_DMA_TxState,
  &UART0_DMA_TxCfg
};
#endif

#if (RTE_UART0_DMA_RX)
edma_chn_state_t       UART0_DMA_RxState;
edma_transfer_config_t UART0_DMA_RxCfg;

static UART_DMA UART0_DMA_Rx = {
  kDmaRequestMux0UART0Rx,
  &UART0_DMA_RxState,
  &UART0_DMA_RxCfg
};
#endif

// UART0 Run-time information
static UART_INFO UART0_info;

static const UART_RESOURCES UART0_Resources = {
 &UART1_Capabilities,
  UART0,
  kSimClockGateUart0,
  UART0_RX_TX_IRQn,
  UART0_ERR_IRQn,
#if (RTE_UART0_DMA_TX)
 &UART0_DMA_Tx,
#else
  NULL,
#endif
#if (RTE_UART0_DMA_RX)
 &UART0_DMA_Rx,
#else
  NULL,
#endif
  UART0_TX_FIFO_WATERMARK,
  UART0_RX_FIFO_WATERMARK,
 &UART0_info
};
#endif

#if (RTE_UART1)

// FIFO watermark must be less than FIFO size
#ifndef UART1_TX_FIFO_WATERMARK
#define UART1_TX_FIFO_WATERMARK (1U)
#endif

// FIFO watermark must be less than FIFO size
#ifndef UART1_RX_FIFO_WATERMARK
#define UART1_RX_FIFO_WATERMARK (1U)
#endif

// UART1 DMA Configuration
#if (RTE_UART1_DMA_TX)
edma_chn_state_t       UART1_DMA_TxState;
edma_transfer_config_t UART1_DMA_TxCfg;

static UART_DMA UART1_DMA_Tx = {
  kDmaRequestMux0UART1Tx,
  &UART1_DMA_TxState,
  &UART1_DMA_TxCfg
};
#endif

#if (RTE_UART1_DMA_RX)
edma_chn_state_t       UART1_DMA_RxState;
edma_transfer_config_t UART1_DMA_RxCfg;

static UART_DMA UART1_DMA_Rx = {
  kDmaRequestMux0UART1Rx,
  &UART1_DMA_RxState,
  &UART1_DMA_RxCfg
};
#endif

// UART1 Run-time information
static UART_INFO UART1_info;

static const UART_RESOURCES UART1_Resources = {
 &UART_Capabilities,
  UART1,
  kSimClockGateUart1,
  UART1_RX_TX_IRQn,
  UART1_ERR_IRQn,
#if (RTE_UART1_DMA_TX)
 &UART1_DMA_Tx,
#else
  NULL,
#endif
#if (RTE_UART1_DMA_RX)
 &UART1_DMA_Rx,
#else
  NULL,
#endif
  UART1_TX_FIFO_WATERMARK,
  UART1_RX_FIFO_WATERMARK,
 &UART1_info
};
#endif

#if (RTE_UART2)

// FIFO watermark must be less than FIFO size
#ifndef UART2_TX_FIFO_WATERMARK
#define UART2_TX_FIFO_WATERMARK (1U)
#endif

// FIFO watermark must be less than FIFO size
#ifndef UART2_RX_FIFO_WATERMARK
#define UART2_RX_FIFO_WATERMARK (1U)
#endif

// UART2 DMA Configuration
#if (RTE_UART2_DMA_TX)
edma_chn_state_t       UART2_DMA_TxState;
edma_transfer_config_t UART2_DMA_TxCfg;

static UART_DMA UART2_DMA_Tx = {
  kDmaRequestMux0UART2Tx,
  &UART2_DMA_TxState,
  &UART2_DMA_TxCfg
};
#endif

#if (RTE_UART2_DMA_RX)
edma_chn_state_t       UART2_DMA_RxState;
edma_transfer_config_t UART2_DMA_RxCfg;

static UART_DMA UART2_DMA_Rx = {
  kDmaRequestMux0UART2Rx,
  &UART2_DMA_RxState,
  &UART2_DMA_RxCfg
};
#endif

// UART2 Run-time information
static UART_INFO UART2_info;

static const UART_RESOURCES UART2_Resources = {
 &UART_Capabilities,
  UART2,
  kSimClockGateUart2,
  UART2_RX_TX_IRQn,
  UART2_ERR_IRQn,
#if (RTE_UART2_DMA_TX)
 &UART2_DMA_Tx,
#else
  NULL,
#endif
#if (RTE_UART2_DMA_RX)
 &UART2_DMA_Rx,
#else
  NULL,
#endif
  UART2_TX_FIFO_WATERMARK,
  UART2_RX_FIFO_WATERMARK,
 &UART2_info
};
#endif

#if (RTE_UART3)

// FIFO watermark must be less than FIFO size
#ifndef UART3_TX_FIFO_WATERMARK
#define UART3_TX_FIFO_WATERMARK (1U)
#endif

// FIFO watermark must be less than FIFO size
#ifndef UART3_RX_FIFO_WATERMARK
#define UART3_RX_FIFO_WATERMARK (1U)
#endif

// UART3 DMA Configuration
#if (RTE_UART3_DMA_TX)
edma_chn_state_t       UART3_DMA_TxState;
edma_transfer_config_t UART3_DMA_TxCfg;

static UART_DMA UART3_DMA_Tx = {
  kDmaRequestMux0UART3Tx,
  &UART3_DMA_TxState,
  &UART3_DMA_TxCfg
};
#endif

#if (RTE_UART3_DMA_RX)
edma_chn_state_t       UART3_DMA_RxState;
edma_transfer_config_t UART3_DMA_RxCfg;

static UART_DMA UART3_DMA_Rx = {
  kDmaRequestMux0UART3Rx,
  &UART3_DMA_RxState,
  &UART3_DMA_RxCfg
};
#endif

// UART3 Run-time information
static UART_INFO UART3_info;

static const UART_RESOURCES UART3_Resources = {
 &UART_Capabilities,
  UART3,
  kSimClockGateUart3,
  UART3_RX_TX_IRQn,
  UART3_ERR_IRQn,
#if (RTE_UART3_DMA_TX)
 &UART3_DMA_Tx,
#else
  NULL,
#endif
#if (RTE_UART3_DMA_RX)
 &UART3_DMA_Rx,
#else
  NULL,
#endif
  UART3_TX_FIFO_WATERMARK,
  UART3_RX_FIFO_WATERMARK,
 &UART3_info
};
#endif

#if (RTE_UART4)

// FIFO watermark must be less than FIFO size
#ifndef UART4_TX_FIFO_WATERMARK
#define UART4_TX_FIFO_WATERMARK (1U)
#endif

// FIFO watermark must be less than FIFO size
#ifndef UART4_RX_FIFO_WATERMARK
#define UART4_RX_FIFO_WATERMARK (1U)
#endif

// UART4 DMA Configuration
#if (RTE_UART4_DMA_TX)
edma_chn_state_t       UART4_DMA_TxState;
edma_transfer_config_t UART4_DMA_TxCfg;

static UART_DMA UART4_DMA_Tx = {
  kDmaRequestMux0UART4,
  &UART4_DMA_TxState,
  &UART4_DMA_TxCfg
};
#endif

#if (RTE_UART4_DMA_RX)
edma_chn_state_t       UART4_DMA_RxState;
edma_transfer_config_t UART4_DMA_RxCfg;

static UART_DMA UART4_DMA_Rx = {
  kDmaRequestMux0UART4,
  &UART4_DMA_RxState,
  &UART4_DMA_RxCfg
};
#endif

// UART4 Run-time information
static UART_INFO UART4_info;

static const UART_RESOURCES UART4_Resources = {
 &UART_Capabilities,
  UART4,
  kSimClockGateUart4,
  UART4_RX_TX_IRQn,
  UART4_ERR_IRQn,
#if (RTE_UART4_DMA_TX)
 &UART4_DMA_Tx,
#else
  NULL,
#endif
#if (RTE_UART4_DMA_RX)
 &UART4_DMA_Rx,
#else
  NULL,
#endif
  UART4_TX_FIFO_WATERMARK,
  UART4_RX_FIFO_WATERMARK,
 &UART4_info
};
#endif

#if (RTE_UART5)

// FIFO watermark must be less than FIFO size
#ifndef UART5_TX_FIFO_WATERMARK
#define UART5_TX_FIFO_WATERMARK (1U)
#endif

// FIFO watermark must be less than FIFO size
#ifndef UART5_RX_FIFO_WATERMARK
#define UART5_RX_FIFO_WATERMARK (1U)
#endif

// UART5 DMA Configuration
#if (RTE_UART5_DMA_TX)
edma_chn_state_t       UART5_DMA_TxState;
edma_transfer_config_t UART5_DMA_TxCfg;

static UART_DMA UART5_DMA_Tx = {
  kDmaRequestMux0UART5,
  &UART5_DMA_TxState,
  &UART5_DMA_TxCfg
};
#endif

#if (RTE_UART5_DMA_RX)
edma_chn_state_t       UART5_DMA_RxState;
edma_transfer_config_t UART5_DMA_RxCfg;

static UART_DMA UART5_DMA_Rx = {
  kDmaRequestMux0UART5,
  &UART5_DMA_RxState,
  &UART5_DMA_RxCfg
};
#endif

// UART5 Run-time information
static UART_INFO UART5_info;

static const UART_RESOURCES UART5_Resources = {
 &UART_Capabilities,
  UART5,
  kSimClockGateUart5,
  UART5_RX_TX_IRQn,
  UART5_ERR_IRQn,
#if (RTE_UART5_DMA_TX)
 &UART5_DMA_Tx,
#else
  NULL,
#endif
#if (RTE_UART5_DMA_RX)
 &UART5_DMA_Rx,
#else
  NULL,
#endif
  UART5_TX_FIFO_WATERMARK,
  UART5_RX_FIFO_WATERMARK,
 &UART5_info
};
#endif

// USART Driver functions

/**
  \fn          ARM_DRIVER_VERSION USARTx_GetVersion (void)
  \brief       Get driver version.
  \return      \ref ARM_DRIVER_VERSION
*/
static ARM_DRIVER_VERSION USARTx_GetVersion (void) {
  return usart_driver_version;
}

/**
  \fn          ARM_USART_CAPABILITIES USART_GetCapabilities (UART_RESOURCES *usart)
  \brief       Get driver capabilities
  \param[in]   usart     Pointer to USART resources
  \return      \ref ARM_USART_CAPABILITIES
*/
static ARM_USART_CAPABILITIES USART_GetCapabilities (UART_RESOURCES *usart) {
  return *(usart->capabilities);
}

/**
  \fn          int32_t USART_Initialize (ARM_UART_SignalEvent_t  cb_event
                                         USART_RESOURCES         *usart)
  \brief       Initialize USART Interface.
  \param[in]   cb_event  Pointer to \ref ARM_USART_SignalEvent
  \param[in]   usart     Pointer to USART resources
  \return      \ref execution_status
*/
static int32_t USART_Initialize (ARM_USART_SignalEvent_t  cb_event,
                                 UART_RESOURCES          *usart) {

  if (usart->info->flags & USART_FLAG_INITIALIZED) {
    // Driver is already initialized
    return ARM_DRIVER_OK;
  }

  // Initialize USART Run-time3 Resources
  usart->info->cb_event = cb_event;

  usart->info->rx_status.rx_busy          = 0U;
  usart->info->rx_status.rx_overflow      = 0U;
  usart->info->rx_status.rx_break         = 0U;
  usart->info->rx_status.rx_framing_error = 0U;
  usart->info->rx_status.rx_parity_error  = 0U;

  usart->info->mode        = 0U;
  usart->info->send_active = 0U;

#if (((RTE_UART0) && (RTE_UART0_DMA_TX == 1)) || \
     ((RTE_UART1) && (RTE_UART1_DMA_TX == 1)) || \
     ((RTE_UART2) && (RTE_UART2_DMA_TX == 1)) || \
     ((RTE_UART3) && (RTE_UART3_DMA_TX == 1)) || \
     ((RTE_UART4) && (RTE_UART4_DMA_TX == 1)) || \
     ((RTE_UART5) && (RTE_UART5_DMA_TX == 1)))
  // DMA Initialize
  if (usart->tx_dma != NULL) {
    if (kEDMAInvalidChannel != EDMA_DRV_RequestChannel (kEDMAAnyChannel,
                                                        usart->tx_dma->req_src,
                                                        usart->tx_dma->ch_info)) {

      EDMA_DRV_InstallCallback (usart->tx_dma->ch_info, UARTx_TX_DMA_Callback, (void *)usart);
    } else {
      return ARM_DRIVER_ERROR;
    }

    // Partly prepare DMA channel config structure
    usart->tx_dma->cfg->srcTransferSize    = kEDMATransferSize_1Bytes;
    usart->tx_dma->cfg->destTransferSize   = kEDMATransferSize_1Bytes;
    usart->tx_dma->cfg->srcLastAddrAdjust  = 0U;
    usart->tx_dma->cfg->destLastAddrAdjust = 0U;
    usart->tx_dma->cfg->srcModulo          = kEDMAModuloDisable;
    usart->tx_dma->cfg->destModulo         = kEDMAModuloDisable;
    usart->tx_dma->cfg->minorLoopCount     = 1;
  }
#endif

#if (((RTE_UART0) && (RTE_UART0_DMA_RX == 1)) || \
     ((RTE_UART1) && (RTE_UART1_DMA_RX == 1)) || \
     ((RTE_UART2) && (RTE_UART2_DMA_RX == 1)) || \
     ((RTE_UART3) && (RTE_UART3_DMA_RX == 1)) || \
     ((RTE_UART4) && (RTE_UART4_DMA_RX == 1)) || \
     ((RTE_UART5) && (RTE_UART5_DMA_RX == 1)))
  if (usart->rx_dma != NULL) {
    if (kEDMAInvalidChannel != EDMA_DRV_RequestChannel (kEDMAAnyChannel,
                                                        usart->rx_dma->req_src,
                                                        usart->rx_dma->ch_info)) {

      EDMA_DRV_InstallCallback (usart->rx_dma->ch_info, UARTx_RX_DMA_Callback, (void *)usart);
    } else {
      return ARM_DRIVER_ERROR;
    }

    // Partly prepare DMA channel config structure
    usart->rx_dma->cfg->srcTransferSize    = kEDMATransferSize_1Bytes;
    usart->rx_dma->cfg->destTransferSize   = kEDMATransferSize_1Bytes;
    usart->rx_dma->cfg->srcLastAddrAdjust  = 0U;
    usart->rx_dma->cfg->destLastAddrAdjust = 0U;
    usart->rx_dma->cfg->srcModulo          = kEDMAModuloDisable;
    usart->rx_dma->cfg->destModulo         = kEDMAModuloDisable;
    usart->rx_dma->cfg->minorLoopCount     = 1;
  }
#endif

  usart->info->flags = USART_FLAG_INITIALIZED;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USART_Uninitialize (UART_RESOURCES *usart)
  \brief       De-initialize USART Interface.
  \param[in]   usart     Pointer to USART resources
  \return      \ref execution_status
*/
static int32_t USART_Uninitialize (UART_RESOURCES *usart) {
  // Reset USART status flags
  usart->info->flags = 0U;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USART_PowerControl (ARM_POWER_STATE state, 
                                           UART_RESOURCES *usart)
  \brief       Control USART Interface Power.
  \param[in]   state  Power state
  \param[in]   usart  Pointer to USART resources
  \return      \ref execution_status
*/
static int32_t USART_PowerControl (ARM_POWER_STATE  state,
                                   UART_RESOURCES  *usart) {
  switch (state) {
    case ARM_POWER_OFF:
      // Disable USART IRQ
      NVIC_DisableIRQ(usart->irq_num);
      NVIC_DisableIRQ(usart->err_irq_num);

      // Disable USART register interface clock
      SIM_HAL_DisableClock (SIM, usart->clock_gate);

      usart->info->rx_status.rx_busy          = 0U;
      usart->info->rx_status.rx_overflow      = 0U;
      usart->info->rx_status.rx_break         = 0U;
      usart->info->rx_status.rx_framing_error = 0U;
      usart->info->rx_status.rx_parity_error  = 0U;

      usart->info->mode        = 0U;
      usart->info->send_active = 0U;

      usart->info->flags = USART_FLAG_INITIALIZED;
      break;

    case ARM_POWER_LOW:
      return ARM_DRIVER_ERROR_UNSUPPORTED;

    case ARM_POWER_FULL:
      if (usart->info->flags & USART_FLAG_POWERED) {
        return ARM_DRIVER_OK;
      }

      usart->info->rx_status.rx_busy          = 0U;
      usart->info->rx_status.rx_overflow      = 0U;
      usart->info->rx_status.rx_break         = 0U;
      usart->info->rx_status.rx_framing_error = 0U;
      usart->info->rx_status.rx_parity_error  = 0U;

      usart->info->mode        = 0U;
      usart->info->send_active = 0U;

      // Enable USART register interface clock
      SIM_HAL_EnableClock (SIM, usart->clock_gate);

      UART_HAL_Init (usart->base);

      // Disable transmitter
      UART_HAL_DisableTransmitter (usart->base);

      // Disable receiver
      UART_HAL_DisableReceiver (usart->base);

      // Disable interrupts 
      UART_HAL_SetIntMode (usart->base, kUartIntLinBreakDetect,  false);
      UART_HAL_SetIntMode (usart->base, kUartIntRxActiveEdge,    false);
      UART_HAL_SetIntMode (usart->base, kUartIntTxDataRegEmpty,  false);
      UART_HAL_SetIntMode (usart->base, kUartIntTxComplete,      false);
      UART_HAL_SetIntMode (usart->base, kUartIntRxDataRegFull,   false);
      UART_HAL_SetIntMode (usart->base, kUartIntIdleLine,        false);
      UART_HAL_SetIntMode (usart->base, kUartIntRxOverrun,       false);
      UART_HAL_SetIntMode (usart->base, kUartIntNoiseErrFlag,    false);
      UART_HAL_SetIntMode (usart->base, kUartIntFrameErrFlag,    false);
      UART_HAL_SetIntMode (usart->base, kUartIntParityErrFlag,   false);
      UART_HAL_SetIntMode (usart->base, kUartIntTxFifoOverflow,  false);
      UART_HAL_SetIntMode (usart->base, kUartIntRxFifoUnderflow, false);

      // Get FIFO sizes
      switch (UART_HAL_GetTxFifoSize(usart->base)) {
        case UART_FIFO_SZ_1_DATAWORD:   usart->info->tx_fifo_sz = 1U;   break;
        case UART_FIFO_SZ_4_DATAWORD:   usart->info->tx_fifo_sz = 4U;   break;
        case UART_FIFO_SZ_8_DATAWORD:   usart->info->tx_fifo_sz = 8U;   break;
        case UART_FIFO_SZ_16_DATAWORD:  usart->info->tx_fifo_sz = 16U;  break;
        case UART_FIFO_SZ_32_DATAWORD:  usart->info->tx_fifo_sz = 32U;  break;
        case UART_FIFO_SZ_64_DATAWORD:  usart->info->tx_fifo_sz = 64U;  break;
        case UART_FIFO_SZ_128_DATAWORD: usart->info->tx_fifo_sz = 128U; break;
        default: return ARM_DRIVER_ERROR;
      }
      switch (UART_HAL_GetRxFifoSize(usart->base)) {
        case UART_FIFO_SZ_1_DATAWORD:   usart->info->rx_fifo_sz = 1U;   break;
        case UART_FIFO_SZ_4_DATAWORD:   usart->info->rx_fifo_sz = 4U;   break;
        case UART_FIFO_SZ_8_DATAWORD:   usart->info->rx_fifo_sz = 8U;   break;
        case UART_FIFO_SZ_16_DATAWORD:  usart->info->rx_fifo_sz = 16U;  break;
        case UART_FIFO_SZ_32_DATAWORD:  usart->info->rx_fifo_sz = 32U;  break;
        case UART_FIFO_SZ_64_DATAWORD:  usart->info->rx_fifo_sz = 64U;  break;
        case UART_FIFO_SZ_128_DATAWORD: usart->info->rx_fifo_sz = 128U; break;
        default: return ARM_DRIVER_ERROR;
      }

      // Enable FIFOs and set FIFO watermark values
      if (usart->info->tx_fifo_sz != 1U) {
        // Check FIFO watermark value
        if (usart->tx_fifo_watermark >= usart->info->tx_fifo_sz) { return ARM_DRIVER_ERROR; }

        // Set FIFO watermark value
        UART_HAL_SetTxFifoWatermark (usart->base, usart->tx_fifo_watermark);

        // Enable and flush TX FIFO
        UART_HAL_SetTxFifoCmd (usart->base, true);
        UART_HAL_FlushTxFifo (usart->base);
      } else {
        // Disable TX FIFO
        UART_HAL_SetTxFifoCmd (usart->base, false);
      }

      if (usart->info->rx_fifo_sz != 1U) {
        // Check FIFO watermark value
        if (usart->rx_fifo_watermark >= usart->info->rx_fifo_sz) { return ARM_DRIVER_ERROR; }

        // Set FIFO watermark value
        UART_HAL_SetRxFifoWatermark (usart->base, usart->rx_fifo_watermark);

        // Enable and flush RX FIFO
        UART_HAL_SetRxFifoCmd (usart->base, true);
        UART_HAL_FlushRxFifo (usart->base);
      } else {
        // Disable RX FIFO
        UART_HAL_SetRxFifoCmd (usart->base, false);
      }

      usart->info->flags = USART_FLAG_POWERED | USART_FLAG_INITIALIZED;

      // Clear and Enable USART IRQ
      NVIC_ClearPendingIRQ(usart->irq_num);
      NVIC_ClearPendingIRQ(usart->err_irq_num);
      NVIC_EnableIRQ(usart->irq_num);
      NVIC_EnableIRQ(usart->err_irq_num);

      break;

    default: return ARM_DRIVER_ERROR_UNSUPPORTED;
  }
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USART_Send (const void            *data,
                                         uint32_t         num,
                                         UART_RESOURCES  *usart)
  \brief       Start sending data to USART transmitter.
  \param[in]   data  Pointer to buffer with data to send to USART transmitter
  \param[in]   num   Number of data items to send
  \param[in]   usart Pointer to USART resources
  \return      \ref execution_status
*/
static int32_t USART_Send (const void            *data,
                                 uint32_t         num,
                                 UART_RESOURCES  *usart) {
  edma_software_tcd_t tcd;

  if ((data == NULL) || (num == 0)) {
    // Invalid parameters
    return ARM_DRIVER_ERROR_PARAMETER;
  }

  if ((usart->info->flags & USART_FLAG_CONFIGURED) == 0U) {
    // USART is not configured (mode not selected)
    return ARM_DRIVER_ERROR;
  }

  if (usart->info->send_active != 0U) {
    // Send is not completed yet
    return ARM_DRIVER_ERROR_BUSY;
  }

  // Set Send active flag
  usart->info->send_active = 1U;

  // Save transmit buffer info
  usart->info->xfer.tx_buf = (uint8_t *)data;
  usart->info->xfer.tx_num = num;
  usart->info->xfer.tx_cnt = 0U;

  // DMA mode
  if (usart->tx_dma) {
    // Prepare config structure
    usart->tx_dma->cfg->srcAddr        = (uint32_t)data;
    usart->tx_dma->cfg->destAddr       = UART_HAL_GetDataRegAddr (usart->base);
    usart->tx_dma->cfg->srcOffset      = 1U;
    usart->tx_dma->cfg->destOffset     = 0U;
    usart->tx_dma->cfg->majorLoopCount = num;

    EDMA_DRV_PrepareDescriptorTransfer (usart->tx_dma->ch_info, &tcd, usart->tx_dma->cfg, true, false);

    EDMA_HAL_STCDSetScatterGatherCmd   (&tcd, false);
    EDMA_HAL_STCDSetChannelMajorLink   (&tcd, usart->tx_dma->ch_info->channel, false);
    EDMA_HAL_STCDSetHalfCompleteIntCmd (&tcd, false);
    EDMA_HAL_STCDSetBandwidth          (&tcd, kEDMABandwidthStallNone);
    EDMA_HAL_STCDSetDisableDmaRequestAfterTCDDoneCmd (&tcd, true);

    EDMA_DRV_PushDescriptorToReg       (usart->tx_dma->ch_info, &tcd);
    EDMA_DRV_StartChannel              (usart->tx_dma->ch_info);

    // Enable USART TX DMA request
    UART_HAL_SetTxDmaCmd (usart->base, true);

  // Interrupt mode
  } else {
    // Enable transmit data register empty interrupt
    UART_HAL_SetIntMode (usart->base, kUartIntTxDataRegEmpty, true);
  }

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USART_Receive (void            *data,
                                      uint32_t         num,
                                      UART_RESOURCES  *usart)
  \brief       Start receiving data from USART receiver.
  \param[out]  data  Pointer to buffer for data to receive from USART receiver
  \param[in]   num   Number of data items to receive
  \param[in]   usart Pointer to USART resources
  \return      \ref execution_status
*/
static int32_t USART_Receive (void            *data,
                              uint32_t         num,
                              UART_RESOURCES  *usart) {
  edma_software_tcd_t tcd;

  if ((data == NULL) || (num == 0U)) {
    // Invalid parameters
    return ARM_DRIVER_ERROR_PARAMETER;
  }

  if ((usart->info->flags & USART_FLAG_CONFIGURED) == 0U) {
    // USART is not configured (mode not selected)
    return ARM_DRIVER_ERROR;
  }

  // Check if receiver is busy
  if (usart->info->rx_status.rx_busy == 1) {
    return ARM_DRIVER_ERROR_BUSY;
  }

  // Set RX busy flag
  usart->info->rx_status.rx_busy = 1;

  // Save number of data to be received
  usart->info->xfer.rx_num = num;

  // Clear RX statuses
  usart->info->rx_status.rx_break          = 0U;
  usart->info->rx_status.rx_framing_error  = 0U;
  usart->info->rx_status.rx_overflow       = 0U;
  usart->info->rx_status.rx_parity_error   = 0U;

  // Save receive buffer info
  usart->info->xfer.rx_buf = (uint8_t *)data;
  usart->info->xfer.rx_cnt =            0U;

  // DMA mode
  if (usart->rx_dma) {

    // Prepare config structure
    usart->rx_dma->cfg->srcAddr        = UART_HAL_GetDataRegAddr (usart->base);
    usart->rx_dma->cfg->destAddr       = (uint32_t)data;
    usart->rx_dma->cfg->srcOffset      = 0U;
    usart->rx_dma->cfg->destOffset     = 1U;
    usart->rx_dma->cfg->majorLoopCount = num;

    EDMA_DRV_PrepareDescriptorTransfer (usart->rx_dma->ch_info, &tcd, usart->rx_dma->cfg, true, false);

    EDMA_HAL_STCDSetScatterGatherCmd   (&tcd, false);
    EDMA_HAL_STCDSetChannelMajorLink   (&tcd, usart->rx_dma->ch_info->channel, false);
    EDMA_HAL_STCDSetHalfCompleteIntCmd (&tcd, false);
    EDMA_HAL_STCDSetBandwidth          (&tcd, kEDMABandwidthStallNone);
    EDMA_HAL_STCDSetDisableDmaRequestAfterTCDDoneCmd (&tcd, true);

    EDMA_DRV_PushDescriptorToReg       (usart->rx_dma->ch_info, &tcd);
    EDMA_DRV_StartChannel              (usart->rx_dma->ch_info);

    // Enable USART RX DMA request
    UART_HAL_SetRxDmaCmd (usart->base, true);

  // Interrupt mode
  } else {
    // Enable receive data available interrupt
    // Enable transmit data register empty interrupt
    UART_HAL_SetIntMode (usart->base, kUartIntRxDataRegFull, true);
  }

  // Enable Idle line interrupt
  UART_HAL_SetIntMode (usart->base, kUartIntParityErrFlag, true);
  UART_HAL_SetIntMode (usart->base, kUartIntNoiseErrFlag,  true);
  UART_HAL_SetIntMode (usart->base, kUartIntFrameErrFlag,  true);
  UART_HAL_SetIntMode (usart->base, kUartIntIdleLine,      true);
  UART_HAL_SetIntMode (usart->base, kUartIntRxOverrun,     true);

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USART_Transfer (const void             *data_out,
                                             void             *data_in,
                                             uint32_t          num,
                                             UART_RESOURCES   *usart)
  \brief       Start sending/receiving data to/from USART transmitter/receiver.
  \param[in]   data_out  Pointer to buffer with data to send to USART transmitter
  \param[out]  data_in   Pointer to buffer for data to receive from USART receiver
  \param[in]   num       Number of data items to transfer
  \param[in]   usart     Pointer to USART resources
  \return      \ref execution_status
*/
static int32_t USART_Transfer (const void             *data_out,
                                     void             *data_in,
                                     uint32_t          num,
                                     UART_RESOURCES   *usart) {

  return ARM_DRIVER_ERROR;
}

/**
  \fn          uint32_t USART_GetTxCount (UART_RESOURCES *usart)
  \brief       Get transmitted data count.
  \param[in]   usart     Pointer to USART resources
  \return      number of data items transmitted
*/
static uint32_t USART_GetTxCount (UART_RESOURCES *usart) {
  return usart->info->xfer.tx_cnt;
}

/**
  \fn          uint32_t USART_GetRxCount (USART_RESOURCES *usart)
  \brief       Get received data count.
  \param[in]   usart     Pointer to USART resources
  \return      number of data items received
*/
static uint32_t USART_GetRxCount (UART_RESOURCES *usart) {
  return usart->info->xfer.rx_cnt;
}

/**
  \fn          int32_t USART_Control (uint32_t          control,
                                      uint32_t          arg,
                                      UART_RESOURCES   *usart)
  \brief       Control USART Interface.
  \param[in]   control  Operation
  \param[in]   arg      Argument of operation (optional)
  \param[in]   usart    Pointer to USART resources
  \return      common \ref execution_status and driver specific \ref usart_execution_status
*/
static int32_t USART_Control (uint32_t          control,
                              uint32_t          arg,
                              UART_RESOURCES   *usart) {
  uint32_t mode;
  uint32_t tx_en, rx_en, val, parity;

  if ((usart->info->flags & USART_FLAG_POWERED) == 0U) {
    // USART not powered
    return ARM_DRIVER_ERROR;
  }

  switch (control & ARM_USART_CONTROL_Msk) {
    // Control TX
    case ARM_USART_CONTROL_TX:
      if (arg) {
        // Enable transmitter
        UART_HAL_EnableTransmitter (usart->base);
      } else {
        // Disable transmitter
        UART_HAL_DisableTransmitter (usart->base);

        
      }
      return ARM_DRIVER_OK;

    // Control RX
    case ARM_USART_CONTROL_RX:
      if (arg) {

        // Enable receiver
        UART_HAL_EnableReceiver (usart->base);

        // Enable RX interrupts
        UART_HAL_SetIntMode (usart->base, kUartIntParityErrFlag, true);
        UART_HAL_SetIntMode (usart->base, kUartIntRxOverrun,     true);
        UART_HAL_SetIntMode (usart->base, kUartIntNoiseErrFlag,  true);
        UART_HAL_SetIntMode (usart->base, kUartIntFrameErrFlag,  true);
        UART_HAL_SetIntMode (usart->base, kUartIntRxDataRegFull, true);
        UART_HAL_SetIntMode (usart->base, kUartIntIdleLine,      true);
      } else {
        // Disable receiver
        UART_HAL_DisableReceiver (usart->base);

        // Disable RX Interrupts
        UART_HAL_SetIntMode (usart->base, kUartIntParityErrFlag, false);
        UART_HAL_SetIntMode (usart->base, kUartIntRxOverrun,     false);
        UART_HAL_SetIntMode (usart->base, kUartIntNoiseErrFlag,  false);
        UART_HAL_SetIntMode (usart->base, kUartIntFrameErrFlag,  false);
        UART_HAL_SetIntMode (usart->base, kUartIntRxDataRegFull, false);
        UART_HAL_SetIntMode (usart->base, kUartIntIdleLine,      false);
      }
      return ARM_DRIVER_OK;

    // Control break
    case ARM_USART_CONTROL_BREAK:
      if (arg) {
        if (usart->info->send_active != 0U) { return ARM_DRIVER_ERROR_BUSY; }
    
        // Set Send active flag
        usart->info->send_active = 1U;

        // Send break characters
        UART_HAL_SetBreakCharCmd (usart->base, true);
      }
      else {
        // Stop sending break characters
        UART_HAL_SetBreakCharCmd (usart->base, false);

        // Set Send active flag
        usart->info->send_active = 0U;
      }
      return ARM_DRIVER_OK;

    // Abort Send
    case ARM_USART_ABORT_SEND:
      // Disable transmit data register empty interrupt
      UART_HAL_SetIntMode (usart->base, kUartIntTxDataRegEmpty, false);

      // If DMA mode - disable DMA channel
      if ((usart->tx_dma) && (usart->info->send_active != 0U)) {
        EDMA_DRV_StopChannel (usart->tx_dma->ch_info);
        UART_HAL_SetTxDmaCmd (usart->base, false);
      }

      // Flush FIFO
      UART_HAL_FlushTxFifo (usart->base);

      // Clear Send active flag
      usart->info->send_active = 0U;
      return ARM_DRIVER_OK;

    // Abort receive
    case ARM_USART_ABORT_RECEIVE:
      // Disable transmit data register empty interrupt
      UART_HAL_SetIntMode (usart->base, kUartIntRxDataRegFull, false);

      // If DMA mode - disable DMA channel
      if ((usart->rx_dma) && (usart->info->rx_status.rx_busy)) {
        EDMA_DRV_StopChannel (usart->rx_dma->ch_info);
        UART_HAL_SetRxDmaCmd (usart->base, false);
      }

      // Flush FIFO
      UART_HAL_FlushRxFifo (usart->base);

      // Clear RX busy status
      usart->info->rx_status.rx_busy = 0U;
      return ARM_DRIVER_OK;

    // Abort transfer
    case ARM_USART_ABORT_TRANSFER:
      // Disable transmit data register empty interrupt
      UART_HAL_SetIntMode (usart->base, kUartIntTxDataRegEmpty, false);
      // Disable transmit data register empty interrupt
      UART_HAL_SetIntMode (usart->base, kUartIntRxDataRegFull, false);

      // If DMA mode - disable DMA channel
      if ((usart->tx_dma) && (usart->info->send_active != 0U)) {
        EDMA_DRV_StopChannel (usart->tx_dma->ch_info);
        UART_HAL_SetTxDmaCmd (usart->base, false);
      }

      // If DMA mode - disable DMA channel
      if ((usart->rx_dma) && (usart->info->rx_status.rx_busy)) {
        EDMA_DRV_StopChannel (usart->rx_dma->ch_info);
        UART_HAL_SetRxDmaCmd (usart->base, false);
      }

      // Flush FIFO
      UART_HAL_FlushTxFifo (usart->base);
      UART_HAL_FlushRxFifo (usart->base);

      // Clear busy statuses
      usart->info->rx_status.rx_busy = 0U;
      usart->info->send_active    = 0U;
      return ARM_DRIVER_OK;
  }

  // Check if Receiver/Transmitter is busy
  if ( usart->info->rx_status.rx_busy ||
      (usart->info->send_active != 0U)) {
    return ARM_DRIVER_ERROR_BUSY;
  }

  tx_en = UART_HAL_IsTransmitterEnabled (usart->base);
  rx_en = UART_HAL_IsReceiverEnabled (usart->base);

  // Disable receiver and transmitter
  if (tx_en == true) { UART_HAL_DisableTransmitter (usart->base); }
  if (rx_en == true) { UART_HAL_DisableReceiver    (usart->base); }

  // Stop single wire
  UART_HAL_SetLoopCmd (usart->base, false);

  // Disable IrDA
  UART_BWR_IR_IREN(usart->base, false);

  if (usart->capabilities->smart_card) {
    // Disable SmartCard
    UART_BWR_C7816_ISO_7816E(usart->base, false);
  }

  switch (control & ARM_USART_CONTROL_Msk) {
    case ARM_USART_MODE_ASYNCHRONOUS:
      mode = ARM_USART_MODE_ASYNCHRONOUS;
      break;
    case ARM_USART_MODE_SYNCHRONOUS_MASTER:
      return ARM_USART_ERROR_MODE;
    case ARM_USART_MODE_SYNCHRONOUS_SLAVE:
      return ARM_USART_ERROR_MODE;
    case ARM_USART_MODE_SINGLE_WIRE:
      // Set single wire
      UART_HAL_SetLoopCmd (usart->base, true);
      break;
    case ARM_USART_MODE_IRDA:
      if (usart->capabilities->irda) {
        // Enable IrDA mode
        UART_BWR_IR_IREN(usart->base, true);
      } else return ARM_USART_ERROR_MODE;
      mode = ARM_USART_MODE_IRDA;
      break;
    case ARM_USART_MODE_SMART_CARD:
      if (usart->capabilities->smart_card) {
        // Enable Smart card mode
        UART_BWR_C7816_ISO_7816E(usart->base, true);
      } else return ARM_USART_ERROR_MODE;
      mode = ARM_USART_MODE_SMART_CARD;
      break;

    // Default TX value
    case ARM_USART_SET_DEFAULT_TX_VALUE:
      return ARM_DRIVER_OK;

    // IrDA pulse
    case ARM_USART_SET_IRDA_PULSE:
      if (usart->capabilities->irda) {
        UART_BWR_IR_TNP(usart->base, 0U);
        if (arg != 0U) {
          val = 1000000000U / usart->info->baudrate;
          if      (((arg * 1U) / 32U) <=  val) { UART_BWR_IR_TNP(usart->base, 2U); }
          else if (((arg * 1U) / 16U) <=  val) { UART_BWR_IR_TNP(usart->base, 1U); }
          else if (((arg * 3U) / 16U) <=  val) { UART_BWR_IR_TNP(usart->base, 0U); }
          else if (((arg * 1U) /  4U) <=  val) { UART_BWR_IR_TNP(usart->base, 3U); }
          else return ARM_DRIVER_ERROR;
        }
      } else return ARM_DRIVER_ERROR;
      return ARM_DRIVER_OK;

    // SmartCard guard time
    case ARM_USART_SET_SMART_CARD_GUARD_TIME:
      if (usart->capabilities->smart_card) {
        if (arg >= (255U + 12U)) return ARM_DRIVER_ERROR;
        if (arg <    12U)        return ARM_DRIVER_ERROR;
        UART_HAL_SetGuardBandInteger(usart->base, arg - 12U);
      } else return ARM_DRIVER_ERROR;
      return ARM_DRIVER_OK;

    // SmartCard clock
    case ARM_USART_SET_SMART_CARD_CLOCK:
      if (usart->capabilities->smart_card == 0U)  { return ARM_DRIVER_ERROR; }
      if (arg == 0U)                              { return ARM_DRIVER_OK;    }
      return ARM_DRIVER_OK;

     // SmartCard NACK
    case ARM_USART_CONTROL_SMART_CARD_NACK:
      if (usart->capabilities->smart_card) {
        if (arg) {
          // Generate NACK on overflow and error
          UART_HAL_ConfigureNackOnOverflow (usart->base, kUartIso7816OnackEnable);
          UART_HAL_ConfigureNackOnError    (usart->base, kUartIso7816AnackEnable);
        } else {
          // Do not generate NACK on overflow and error
          UART_HAL_ConfigureNackOnOverflow (usart->base, kUartIso7816OnackDisable);
          UART_HAL_ConfigureNackOnError    (usart->base, kUartIso7816AnackDisable);
        }
      } else return ARM_DRIVER_ERROR;
      return ARM_DRIVER_OK;

    // Unsupported command
    default: return ARM_DRIVER_ERROR_UNSUPPORTED;
  }

  // USART Parity
  parity = control & ARM_USART_PARITY_Msk;
  switch (parity) {
    case ARM_USART_PARITY_NONE: UART_HAL_SetParityMode (usart->base, kUartParityDisabled); break;
    case ARM_USART_PARITY_EVEN: UART_HAL_SetParityMode (usart->base, kUartParityEven);     break;
    case ARM_USART_PARITY_ODD:  UART_HAL_SetParityMode (usart->base, kUartParityOdd);      break;
    default: return ARM_USART_ERROR_PARITY;
  }

  // USART Data bits
  if (parity) {
    switch (control & ARM_USART_DATA_BITS_Msk) {
      case ARM_USART_DATA_BITS_7:
        usart->info->data_bits = 7U;
        UART_HAL_SetBitCountPerChar (usart->base, kUart8BitsPerChar);
        break;
      case ARM_USART_DATA_BITS_8:
        usart->info->data_bits = 8U;
        UART_HAL_SetBitCountPerChar (usart->base, kUart9BitsPerChar);
        break;
      case ARM_USART_DATA_BITS_9:
        // Not available in DMA mode
        if (usart->tx_dma || usart->rx_dma)    { return ARM_USART_ERROR_DATA_BITS; }
        // Not available in SmartCard mode
        if (mode == ARM_USART_MODE_SMART_CARD) { return ARM_USART_ERROR_DATA_BITS; }

        usart->info->data_bits = 9U;
        UART_HAL_SetBitCountPerChar  (usart->base, kUart9BitsPerChar);
        UART_HAL_SetBit10AsParityBit (usart->base, true);
        break;
      default: return ARM_USART_ERROR_DATA_BITS;
    }
  } else {
    switch (control & ARM_USART_DATA_BITS_Msk) {
      case ARM_USART_DATA_BITS_8:
        usart->info->data_bits = 8U;
        UART_HAL_SetBitCountPerChar (usart->base, kUart8BitsPerChar);
        break;
      case ARM_USART_DATA_BITS_9: 
        // Not available in DMA mode
        if (usart->tx_dma || usart->rx_dma) { return ARM_USART_ERROR_DATA_BITS; }
        usart->info->data_bits = 9U;
        UART_HAL_SetBitCountPerChar (usart->base, kUart9BitsPerChar);
        break;
      default: return ARM_USART_ERROR_DATA_BITS;
    }
  }

  // USART Stop bits
  switch (control & ARM_USART_STOP_BITS_Msk) {
    case ARM_USART_STOP_BITS_1: UART_HAL_SetStopBitCount (usart->base, kUartOneStopBit); break;
    case ARM_USART_STOP_BITS_2: UART_HAL_SetStopBitCount (usart->base, kUartTwoStopBit); break;
    default: return ARM_USART_ERROR_STOP_BITS;
  }

  // USART Flow control

  // Disable RTS and CTS
  UART_HAL_SetReceiverRtsCmd    (usart->base, false);
  UART_HAL_SetTransmitterCtsCmd (usart->base, false);

  switch (control & ARM_USART_FLOW_CONTROL_Msk) {
    case ARM_USART_FLOW_CONTROL_NONE:
      break;
    case ARM_USART_FLOW_CONTROL_RTS:
      UART_HAL_SetTransmitterCtsCmd (usart->base, true);
      break;
    case ARM_USART_FLOW_CONTROL_CTS:
      UART_HAL_SetReceiverRtsCmd    (usart->base, true);
      break;
    case ARM_USART_FLOW_CONTROL_RTS_CTS:
      UART_HAL_SetTransmitterCtsCmd (usart->base, true);
      UART_HAL_SetReceiverRtsCmd    (usart->base, true);
      break;
    default:
      return ARM_USART_ERROR_FLOW_CONTROL;
  }

  // USART Baudrate
  switch ((uint32_t)usart->base) {
    case UART0_BASE: val = 0U; break;
    case UART1_BASE: val = 1U; break;
    case UART2_BASE: val = 2U; break;
    case UART3_BASE: val = 3U; break;
    case UART4_BASE: val = 4U; break;
    case UART5_BASE: val = 5U; break;
    default: return ARM_DRIVER_ERROR;
  }
  if (UART_HAL_SetBaudRate (usart->base, CLOCK_SYS_GetUartFreq(val), arg) != kStatus_UART_Success) {
    return ARM_USART_ERROR_BAUDRATE;
  } else {
    usart->info->baudrate = arg;
  }

  // Configuration is OK - Mode is valid
  usart->info->mode = mode;

  // Set configured flag
  usart->info->flags |= USART_FLAG_CONFIGURED;

  // Enable receiver and transmitter
  if (tx_en == true) { UART_HAL_EnableTransmitter (usart->base); }
  if (rx_en == true) { UART_HAL_EnableReceiver    (usart->base); }

  return ARM_DRIVER_OK;
}

/**
  \fn          ARM_USART_STATUS USART_GetStatus (UART_RESOURCES *usart)
  \brief       Get USART status.
  \param[in]   usart     Pointer to USART resources
  \return      USART status \ref ARM_USART_STATUS
*/
static ARM_USART_STATUS USART_GetStatus (UART_RESOURCES *usart) {
  ARM_USART_STATUS stat;

  stat.tx_busy          = (UART_BRD_S1_TC(usart->base) ? (0U) : (1U));
  stat.rx_busy          = usart->info->rx_status.rx_busy;
  stat.tx_underflow     = 0U;
  stat.rx_overflow      = usart->info->rx_status.rx_overflow;
  stat.rx_break         = usart->info->rx_status.rx_break;
  stat.rx_framing_error = usart->info->rx_status.rx_framing_error;
  stat.rx_parity_error  = usart->info->rx_status.rx_parity_error;
  return stat;
}

/**
  \fn          int32_t USART_SetModemControl (ARM_USART_MODEM_CONTROL  control,
                                              UART_RESOURCES          *usart)
  \brief       Set USART Modem Control line state.
  \param[in]   control   \ref ARM_USART_MODEM_CONTROL
  \param[in]   usart     Pointer to USART resources
  \return      \ref execution_status
*/
static int32_t USART_SetModemControl (ARM_USART_MODEM_CONTROL  control,
                                      UART_RESOURCES          *usart) {
  // Software modem line control is not supported
  return ARM_DRIVER_ERROR_UNSUPPORTED;
}

/**
  \fn          ARM_USART_MODEM_STATUS USART_GetModemStatus (UART_RESOURCES *usart)
  \brief       Get USART Modem Status lines state.
  \param[in]   usart     Pointer to USART resources
  \return      modem status \ref ARM_USART_MODEM_STATUS
*/
static ARM_USART_MODEM_STATUS USART_GetModemStatus (UART_RESOURCES  *usart) {
  ARM_USART_MODEM_STATUS modem_status = { 0 };

  return modem_status;
}

/**
  \fn          void USART_IRQHandler (UART_RESOURCES *usart)
  \brief       USART Interrupt handler.
  \param[in]   usart     Pointer to USART resources
*/
static void USART_IRQHandler (UART_RESOURCES  *usart) {
  uint32_t event, s1;
  uint8_t  val, data;

  event = 0U;
  s1    = UART_RD_S1(usart->base);

  // Receiver overrun flag
  if (s1 & UART_S1_OR_MASK) {
    usart->info->rx_status.rx_overflow = 1U;
    // Disable receiver overrun interrupt
    UART_HAL_SetIntMode (usart->base, kUartIntRxOverrun, false);
    event |= ARM_USART_EVENT_RX_OVERFLOW;
  }

  // Framing error flag flag
  if (s1 & UART_S1_FE_MASK) {
    usart->info->rx_status.rx_framing_error = 1U;
    // Disable framing error interrupt
    UART_HAL_SetIntMode (usart->base, kUartIntFrameErrFlag, false);
    event |= ARM_USART_EVENT_RX_FRAMING_ERROR;
  }

  // Parity error  flag
  if (s1 & UART_S1_PF_MASK) {
    usart->info->rx_status.rx_parity_error = 1U;
    // Disable parity error interrupt
    UART_HAL_SetIntMode (usart->base, kUartIntParityErrFlag, false);
    event |= ARM_USART_EVENT_RX_PARITY_ERROR;
  }

  // Idle line
  if (s1 & UART_S1_IDLE_MASK) {
    event |= ARM_USART_EVENT_RX_TIMEOUT;

    // Disable Idle line interrupt
    UART_HAL_SetIntMode (usart->base, kUartIntIdleLine, false);

    if (UART_HAL_GetIntMode(usart->base, kUartIntRxDataRegFull)) {
      val = UART_HAL_GetRxDatawordCountInFifo (usart->base);
      while ((val --) && (usart->info->xfer.rx_num != usart->info->xfer.rx_cnt)) {
        // Read data from RX FIFO into receive buffer
        data = UART_BRD_C3_R8(usart->base);
        *(usart->info->xfer.rx_buf++) = UART_RD_D(usart->base);
        if (usart->info->data_bits == 9U) {
          *(usart->info->xfer.rx_buf++) = data;
        }

        usart->info->xfer.rx_cnt++;

        // Check if requested amount of data is received
        if (usart->info->xfer.rx_cnt == usart->info->xfer.rx_num) {
          // Disable receive data register full flag
          UART_HAL_SetIntMode (usart->base, kUartIntRxDataRegFull, false);

          // Clear RX busy flag and set receive transfer complete event
          usart->info->rx_status.rx_busy = 0;
          event |= ARM_USART_EVENT_RECEIVE_COMPLETE;
        }
      }
    }
  }

  // Receive data register full flag
  if ((s1 & UART_S1_RDRF_MASK) && (UART_HAL_GetIntMode(usart->base, kUartIntRxDataRegFull))) {
    val = UART_HAL_GetRxDatawordCountInFifo (usart->base);

    while ((val --) && (usart->info->xfer.rx_num != usart->info->xfer.rx_cnt)) {
      // Read data from RX FIFO into receive buffer
      data = UART_BRD_C3_R8(usart->base);
      *(usart->info->xfer.rx_buf++) = UART_RD_D(usart->base);
      if (usart->info->data_bits == 9U) {
        *(usart->info->xfer.rx_buf++) = data;
      }

      usart->info->xfer.rx_cnt++;

      // Check if requested amount of data is received
      if (usart->info->xfer.rx_cnt == usart->info->xfer.rx_num) {
        // Disable receive data register full flag
        UART_HAL_SetIntMode (usart->base, kUartIntRxDataRegFull, false);

        // Clear RX busy flag and set receive transfer complete event
        usart->info->rx_status.rx_busy = 0U;
        event |= ARM_USART_EVENT_RECEIVE_COMPLETE;
      }
    }
  }

  // Transmit Data Register Empty Flag
  if (usart->tx_dma == NULL) {
    if ((s1 & UART_S1_TDRE_MASK) && UART_HAL_GetIntMode(usart->base, kUartIntTxDataRegEmpty)) {

      // Calculate available room in FIFO
      val = usart->info->tx_fifo_sz - UART_HAL_GetTxDatawordCountInFifo (usart->base);
      while ((val --) && (usart->info->xfer.tx_num != usart->info->xfer.tx_cnt)) {
        // Write data to Tx FIFO
        data = *(usart->info->xfer.tx_buf++);
        if (usart->info->data_bits == 9U) {
          UART_BWR_C3_T8(usart->base, *(usart->info->xfer.tx_buf++) & 1U);
        }
        UART_WR_D(usart->base, data);
        usart->info->xfer.tx_cnt++;
      }

      // Check if all data is transmitted
      if (usart->info->xfer.tx_num == usart->info->xfer.tx_cnt) {
        // Disable transmit Data Register Empty interrupt
        UART_HAL_SetIntMode (usart->base, kUartIntTxDataRegEmpty, false);

        // Clear TX busy flag
        usart->info->send_active = 0U;

        // Set send complete event
        event |= ARM_USART_EVENT_SEND_COMPLETE;

        // Enable Transfer complete interrupt
        UART_HAL_SetIntMode (usart->base, kUartIntTxComplete, true);
      }
    }
  }

  // Transmit complete flag
  if ((s1 & UART_S1_TC_MASK) && (UART_HAL_GetIntMode(usart->base, kUartIntTxComplete))) {
    if (usart->info->xfer.tx_num == usart->info->xfer.tx_cnt) {
      // Set send complete event
      event |= ARM_USART_EVENT_TX_COMPLETE;

      // Disable transmit complete interrupt
      UART_HAL_SetIntMode (usart->base, kUartIntTxComplete, false);
    }
  }

  if (usart->info->cb_event && event) {
    usart->info->cb_event (event);
  }
}


#if (((RTE_UART0) && (RTE_UART0_DMA_TX == 1)) || \
     ((RTE_UART1) && (RTE_UART1_DMA_TX == 1)) || \
     ((RTE_UART2) && (RTE_UART2_DMA_TX == 1)) || \
     ((RTE_UART3) && (RTE_UART3_DMA_TX == 1)) || \
     ((RTE_UART4) && (RTE_UART4_DMA_TX == 1)) || \
     ((RTE_UART5) && (RTE_UART5_DMA_TX == 1)))
/**
  \fn          void UARTx_TX_DMA_Callback (void *par, edma_chn_status_t status)
  \brief       I2C DMA callback handler.
  \param[in]   par (void)Pointer to I2C resources
*/
static void UARTx_TX_DMA_Callback (void *par, edma_chn_status_t status) {
  UART_RESOURCES *usart = (UART_RESOURCES *)par;

  switch (status) {
    case kEDMAChnNormal:
      usart->info->xfer.tx_cnt = usart->info->xfer.tx_num;

      // Clear TX busy flag
      usart->info->send_active = 0U;

      // Enable Transfer complete interrupt
      UART_HAL_SetIntMode (usart->base, kUartIntTxComplete, true);

      if (usart->info->cb_event) { usart->info->cb_event (ARM_USART_EVENT_SEND_COMPLETE); }

      break;
    case kEDMAChnIdle:
      break;

    case kEDMAChnError:
    default:
      DMA_CERR = 1 << usart->tx_dma->ch_info->channel;
      break;
  }
}
#endif

#if (((RTE_UART0) && (RTE_UART0_DMA_RX == 1)) || \
     ((RTE_UART1) && (RTE_UART1_DMA_RX == 1)) || \
     ((RTE_UART2) && (RTE_UART2_DMA_RX == 1)) || \
     ((RTE_UART3) && (RTE_UART3_DMA_RX == 1)) || \
     ((RTE_UART4) && (RTE_UART4_DMA_RX == 1)) || \
     ((RTE_UART5) && (RTE_UART5_DMA_RX == 1)))
/**
  \fn          void UARTx_RX_DMA_Callback (void *par, edma_chn_status_t status)
  \brief       I2C DMA callback handler.
  \param[in]   par (void)Pointer to I2C resources
*/
static void UARTx_RX_DMA_Callback (void *par, edma_chn_status_t status) {
  UART_RESOURCES *usart = (UART_RESOURCES *)par;

  switch (status) {
    case kEDMAChnNormal:
      usart->info->xfer.rx_cnt = usart->info->xfer.rx_num;

      // Clear RX busy flag and set receive transfer complete event
      usart->info->rx_status.rx_busy = 0;
      if (usart->info->cb_event) { usart->info->cb_event (ARM_USART_EVENT_RECEIVE_COMPLETE); }
      break;
    case kEDMAChnIdle:
      break;

    case kEDMAChnError:
    default:
      DMA_CERR = 1 << usart->rx_dma->ch_info->channel;
      break;
  }
}
#endif

#if (RTE_UART0)
// USART0 Driver Wrapper functions
static ARM_USART_CAPABILITIES USART0_GetCapabilities (void) {
  return USART_GetCapabilities (&UART0_Resources);
}
static int32_t USART0_Initialize (ARM_USART_SignalEvent_t cb_event) {
  return USART_Initialize (cb_event, &UART0_Resources);
}
static int32_t USART0_Uninitialize (void) {
  return USART_Uninitialize(&UART0_Resources);
}
static int32_t USART0_PowerControl (ARM_POWER_STATE state) {
  return USART_PowerControl (state, &UART0_Resources);
}
static int32_t USART0_Send (const void *data, uint32_t num) {
  return USART_Send (data, num, &UART0_Resources);
}
static int32_t USART0_Receive (void *data, uint32_t num) {
  return USART_Receive (data, num, &UART0_Resources);
}
static int32_t USART0_Transfer (const void      *data_out,
                                      void      *data_in,
                                      uint32_t   num) {
  return USART_Transfer (data_out, data_in, num, &UART0_Resources);
}
static uint32_t USART0_GetTxCount (void) {
  return USART_GetTxCount (&UART0_Resources);
}
static uint32_t USART0_GetRxCount (void) {
  return USART_GetRxCount (&UART0_Resources); 
}
static int32_t USART0_Control (uint32_t control, uint32_t arg) {
  return USART_Control (control, arg, &UART0_Resources);
}
static ARM_USART_STATUS USART0_GetStatus (void) {
  return USART_GetStatus (&UART0_Resources);
}
static int32_t USART0_SetModemControl (ARM_USART_MODEM_CONTROL control) {
  return USART_SetModemControl (control, &UART0_Resources);
}
static ARM_USART_MODEM_STATUS USART0_GetModemStatus (void) {
  return USART_GetModemStatus (&UART0_Resources);
}
void UART0_RX_TX_IRQHandler (void) {
  USART_IRQHandler (&UART0_Resources);
}
void UART0_ERR_IRQHandler (void) {
  USART_IRQHandler (&UART0_Resources);
}

// USART0 Driver Control Block
ARM_DRIVER_USART Driver_USART0 = {
    USARTx_GetVersion,
    USART0_GetCapabilities,
    USART0_Initialize,
    USART0_Uninitialize,
    USART0_PowerControl,
    USART0_Send, 
    USART0_Receive,
    USART0_Transfer,
    USART0_GetTxCount,
    USART0_GetRxCount,
    USART0_Control,
    USART0_GetStatus,
    USART0_SetModemControl,
    USART0_GetModemStatus
};
#endif

#if (RTE_UART1)
// USART1 Driver Wrapper functions
static ARM_USART_CAPABILITIES USART1_GetCapabilities (void) {
  return USART_GetCapabilities (&UART1_Resources);
}
static int32_t USART1_Initialize (ARM_USART_SignalEvent_t cb_event) {
  return USART_Initialize (cb_event, &UART1_Resources);
}
static int32_t USART1_Uninitialize (void) {
  return USART_Uninitialize(&UART1_Resources);
}
static int32_t USART1_PowerControl (ARM_POWER_STATE state) {
  return USART_PowerControl (state, &UART1_Resources);
}
static int32_t USART1_Send (const void *data, uint32_t num) {
  return USART_Send (data, num, &UART1_Resources);
}
static int32_t USART1_Receive (void *data, uint32_t num) {
  return USART_Receive (data, num, &UART1_Resources);
}
static int32_t USART1_Transfer (const void      *data_out,
                                      void      *data_in,
                                      uint32_t   num) {
  return USART_Transfer (data_out, data_in, num, &UART1_Resources);
}
static uint32_t USART1_GetTxCount (void) {
  return USART_GetTxCount (&UART1_Resources);
}
static uint32_t USART1_GetRxCount (void) {
  return USART_GetRxCount (&UART1_Resources); 
}
static int32_t USART1_Control (uint32_t control, uint32_t arg) {
  return USART_Control (control, arg, &UART1_Resources);
}
static ARM_USART_STATUS USART1_GetStatus (void) {
  return USART_GetStatus (&UART1_Resources);
}
static int32_t USART1_SetModemControl (ARM_USART_MODEM_CONTROL control) {
  return USART_SetModemControl (control, &UART1_Resources);
}
static ARM_USART_MODEM_STATUS USART1_GetModemStatus (void) {
  return USART_GetModemStatus (&UART1_Resources);
}
void UART1_RX_TX_IRQHandler (void) {
  USART_IRQHandler (&UART1_Resources);
}
void UART1_ERR_IRQHandler (void) {
  USART_IRQHandler (&UART1_Resources);
}

// USART1 Driver Control Block
ARM_DRIVER_USART Driver_USART1 = {
    USARTx_GetVersion,
    USART1_GetCapabilities,
    USART1_Initialize,
    USART1_Uninitialize,
    USART1_PowerControl,
    USART1_Send, 
    USART1_Receive,
    USART1_Transfer,
    USART1_GetTxCount,
    USART1_GetRxCount,
    USART1_Control,
    USART1_GetStatus,
    USART1_SetModemControl,
    USART1_GetModemStatus
};
#endif

#if (RTE_UART2)
// USART2 Driver Wrapper functions
static ARM_USART_CAPABILITIES USART2_GetCapabilities (void) {
  return USART_GetCapabilities (&UART2_Resources);
}
static int32_t USART2_Initialize (ARM_USART_SignalEvent_t cb_event) {
  return USART_Initialize (cb_event, &UART2_Resources);
}
static int32_t USART2_Uninitialize (void) {
  return USART_Uninitialize(&UART2_Resources);
}
static int32_t USART2_PowerControl (ARM_POWER_STATE state) {
  return USART_PowerControl (state, &UART2_Resources);
}
static int32_t USART2_Send (const void *data, uint32_t num) {
  return USART_Send (data, num, &UART2_Resources);
}
static int32_t USART2_Receive (void *data, uint32_t num) {
  return USART_Receive (data, num, &UART2_Resources);
}
static int32_t USART2_Transfer (const void      *data_out,
                                      void      *data_in,
                                      uint32_t   num) {
  return USART_Transfer (data_out, data_in, num, &UART2_Resources);
}
static uint32_t USART2_GetTxCount (void) {
  return USART_GetTxCount (&UART2_Resources);
}
static uint32_t USART2_GetRxCount (void) {
  return USART_GetRxCount (&UART2_Resources); 
}
static int32_t USART2_Control (uint32_t control, uint32_t arg) {
  return USART_Control (control, arg, &UART2_Resources);
}
static ARM_USART_STATUS USART2_GetStatus (void) {
  return USART_GetStatus (&UART2_Resources);
}
static int32_t USART2_SetModemControl (ARM_USART_MODEM_CONTROL control) {
  return USART_SetModemControl (control, &UART2_Resources);
}
static ARM_USART_MODEM_STATUS USART2_GetModemStatus (void) {
  return USART_GetModemStatus (&UART2_Resources);
}
void UART2_RX_TX_IRQHandler (void) {
  USART_IRQHandler (&UART2_Resources);
}
void UART2_ERR_IRQHandler (void) {
  USART_IRQHandler (&UART2_Resources);
}

// USART2 Driver Control Block
ARM_DRIVER_USART Driver_USART2 = {
    USARTx_GetVersion,
    USART2_GetCapabilities,
    USART2_Initialize,
    USART2_Uninitialize,
    USART2_PowerControl,
    USART2_Send, 
    USART2_Receive,
    USART2_Transfer,
    USART2_GetTxCount,
    USART2_GetRxCount,
    USART2_Control,
    USART2_GetStatus,
    USART2_SetModemControl,
    USART2_GetModemStatus
};
#endif

#if (RTE_UART3)
// USART3 Driver Wrapper functions
static ARM_USART_CAPABILITIES USART3_GetCapabilities (void) {
  return USART_GetCapabilities (&UART3_Resources);
}
static int32_t USART3_Initialize (ARM_USART_SignalEvent_t cb_event) {
  return USART_Initialize (cb_event, &UART3_Resources);
}
static int32_t USART3_Uninitialize (void) {
  return USART_Uninitialize(&UART3_Resources);
}
static int32_t USART3_PowerControl (ARM_POWER_STATE state) {
  return USART_PowerControl (state, &UART3_Resources);
}
static int32_t USART3_Send (const void *data, uint32_t num) {
  return USART_Send (data, num, &UART3_Resources);
}
static int32_t USART3_Receive (void *data, uint32_t num) {
  return USART_Receive (data, num, &UART3_Resources);
}
static int32_t USART3_Transfer (const void      *data_out,
                                      void      *data_in,
                                      uint32_t   num) {
  return USART_Transfer (data_out, data_in, num, &UART3_Resources);
}
static uint32_t USART3_GetTxCount (void) {
  return USART_GetTxCount (&UART3_Resources);
}
static uint32_t USART3_GetRxCount (void) {
  return USART_GetRxCount (&UART3_Resources); 
}
static int32_t USART3_Control (uint32_t control, uint32_t arg) {
  return USART_Control (control, arg, &UART3_Resources);
}
static ARM_USART_STATUS USART3_GetStatus (void) {
  return USART_GetStatus (&UART3_Resources);
}
static int32_t USART3_SetModemControl (ARM_USART_MODEM_CONTROL control) {
  return USART_SetModemControl (control, &UART3_Resources);
}
static ARM_USART_MODEM_STATUS USART3_GetModemStatus (void) {
  return USART_GetModemStatus (&UART3_Resources);
}
void UART3_RX_TX_IRQHandler (void) {
  USART_IRQHandler (&UART3_Resources);
}
void UART3_ERR_IRQHandler (void) {
  USART_IRQHandler (&UART3_Resources);
}

// USART3 Driver Control Block
ARM_DRIVER_USART Driver_USART3 = {
    USARTx_GetVersion,
    USART3_GetCapabilities,
    USART3_Initialize,
    USART3_Uninitialize,
    USART3_PowerControl,
    USART3_Send, 
    USART3_Receive,
    USART3_Transfer,
    USART3_GetTxCount,
    USART3_GetRxCount,
    USART3_Control,
    USART3_GetStatus,
    USART3_SetModemControl,
    USART3_GetModemStatus
};
#endif

#if (RTE_UART4)
// USART4 Driver Wrapper functions
static ARM_USART_CAPABILITIES USART4_GetCapabilities (void) {
  return USART_GetCapabilities (&UART4_Resources);
}
static int32_t USART4_Initialize (ARM_USART_SignalEvent_t cb_event) {
  return USART_Initialize (cb_event, &UART4_Resources);
}
static int32_t USART4_Uninitialize (void) {
  return USART_Uninitialize(&UART4_Resources);
}
static int32_t USART4_PowerControl (ARM_POWER_STATE state) {
  return USART_PowerControl (state, &UART4_Resources);
}
static int32_t USART4_Send (const void *data, uint32_t num) {
  return USART_Send (data, num, &UART4_Resources);
}
static int32_t USART4_Receive (void *data, uint32_t num) {
  return USART_Receive (data, num, &UART4_Resources);
}
static int32_t USART4_Transfer (const void      *data_out,
                                      void      *data_in,
                                      uint32_t   num) {
  return USART_Transfer (data_out, data_in, num, &UART4_Resources);
}
static uint32_t USART4_GetTxCount (void) {
  return USART_GetTxCount (&UART4_Resources);
}
static uint32_t USART4_GetRxCount (void) {
  return USART_GetRxCount (&UART4_Resources); 
}
static int32_t USART4_Control (uint32_t control, uint32_t arg) {
  return USART_Control (control, arg, &UART4_Resources);
}
static ARM_USART_STATUS USART4_GetStatus (void) {
  return USART_GetStatus (&UART4_Resources);
}
static int32_t USART4_SetModemControl (ARM_USART_MODEM_CONTROL control) {
  return USART_SetModemControl (control, &UART4_Resources);
}
static ARM_USART_MODEM_STATUS USART4_GetModemStatus (void) {
  return USART_GetModemStatus (&UART4_Resources);
}
void UART4_RX_TX_IRQHandler (void) {
  USART_IRQHandler (&UART4_Resources);
}
void UART4_ERR_IRQHandler (void) {
  USART_IRQHandler (&UART4_Resources);
}

// USART4 Driver Control Block
ARM_DRIVER_USART Driver_USART4 = {
    USARTx_GetVersion,
    USART4_GetCapabilities,
    USART4_Initialize,
    USART4_Uninitialize,
    USART4_PowerControl,
    USART4_Send, 
    USART4_Receive,
    USART4_Transfer,
    USART4_GetTxCount,
    USART4_GetRxCount,
    USART4_Control,
    USART4_GetStatus,
    USART4_SetModemControl,
    USART4_GetModemStatus
};
#endif

#if (RTE_UART5)
// USART5 Driver Wrapper functions
static ARM_USART_CAPABILITIES USART5_GetCapabilities (void) {
  return USART_GetCapabilities (&UART5_Resources);
}
static int32_t USART5_Initialize (ARM_USART_SignalEvent_t cb_event) {
  return USART_Initialize (cb_event, &UART5_Resources);
}
static int32_t USART5_Uninitialize (void) {
  return USART_Uninitialize(&UART5_Resources);
}
static int32_t USART5_PowerControl (ARM_POWER_STATE state) {
  return USART_PowerControl (state, &UART5_Resources);
}
static int32_t USART5_Send (const void *data, uint32_t num) {
  return USART_Send (data, num, &UART5_Resources);
}
static int32_t USART5_Receive (void *data, uint32_t num) {
  return USART_Receive (data, num, &UART5_Resources);
}
static int32_t USART5_Transfer (const void      *data_out,
                                      void      *data_in,
                                      uint32_t   num) {
  return USART_Transfer (data_out, data_in, num, &UART5_Resources);
}
static uint32_t USART5_GetTxCount (void) {
  return USART_GetTxCount (&UART5_Resources);
}
static uint32_t USART5_GetRxCount (void) {
  return USART_GetRxCount (&UART5_Resources); 
}
static int32_t USART5_Control (uint32_t control, uint32_t arg) {
  return USART_Control (control, arg, &UART5_Resources);
}
static ARM_USART_STATUS USART5_GetStatus (void) {
  return USART_GetStatus (&UART5_Resources);
}
static int32_t USART5_SetModemControl (ARM_USART_MODEM_CONTROL control) {
  return USART_SetModemControl (control, &UART5_Resources);
}
static ARM_USART_MODEM_STATUS USART5_GetModemStatus (void) {
  return USART_GetModemStatus (&UART5_Resources);
}
void UART5_RX_TX_IRQHandler (void) {
  USART_IRQHandler (&UART5_Resources);
}
void UART5_ERR_IRQHandler (void) {
  USART_IRQHandler (&UART5_Resources);
}

// USART5 Driver Control Block
ARM_DRIVER_USART Driver_USART5 = {
    USARTx_GetVersion,
    USART5_GetCapabilities,
    USART5_Initialize,
    USART5_Uninitialize,
    USART5_PowerControl,
    USART5_Send, 
    USART5_Receive,
    USART5_Transfer,
    USART5_GetTxCount,
    USART5_GetRxCount,
    USART5_Control,
    USART5_GetStatus,
    USART5_SetModemControl,
    USART5_GetModemStatus
};
#endif
