/* -----------------------------------------------------------------------------
 * Copyright (c) 2013-2015 ARM Ltd.
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from
 * the use of this software. Permission is granted to anyone to use this
 * software for any purpose, including commercial applications, and to alter
 * it and redistribute it freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software in
 *    a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 *
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 3. This notice may not be removed or altered from any source distribution.
 *
 *
 * $Date:        15. June 2015
 * $Revision:    V1.1
 *
 * Driver:       Driver_USBH0
 * Project:      USB Full/Low-Speed Host Driver for Freescale MK64F
 * --------------------------------------------------------------------------
 * Use the following configuration settings in the middleware component
 * to connect to this driver.
 *
 *   Configuration Setting                  Value
 *   ---------------------                  -----
 *   Connect to hardware via Driver_USBH# = 0
 *   USB Host controller interface        = Custom
 * --------------------------------------------------------------------------
 * Defines used for driver configuration (at compile time):
 *
 *   USBH_MAX_PIPE_NUM: defines maximum number of Pipes that driver will
 *                      support, this value impacts driver memory
 *                      requirements
 *     - default value: 16
 *     - maximum value: 32
 * -------------------------------------------------------------------------- */

/* History:
 *  Version 1.1
 *    Updated conditionality of functions
 *  Version 1.0
 *    Initial release for USB Host CMSIS Driver API v2.01
 */


#include <stdint.h>
#include <string.h>
#include "cmsis_os.h"

#include "Driver_USBH.h"

#include "fsl_device_registers.h"

#ifndef USBH_MAX_PIPE_NUM
#define USBH_MAX_PIPE_NUM               16U
#endif
#if    (USBH_MAX_PIPE_NUM > 32)
#error  Too many Pipes, maximum Pipes that this driver supports is 32 !!!
#endif

extern uint8_t otg_fs_role;
extern uint8_t otg_fs_state;

extern void OTG_FS_PinVbusOnOff    (bool state);
extern bool OTG_FS_PinGetOC        (void);


// USBH Driver *****************************************************************

#define ARM_USBH_DRV_VERSION ARM_DRIVER_VERSION_MAJOR_MINOR(1,1)

// Driver Version
static const ARM_DRIVER_VERSION usbh_driver_version = { ARM_USBH_API_VERSION, ARM_USBH_DRV_VERSION };

// Driver Capabilities
static const ARM_USBH_CAPABILITIES usbh_driver_capabilities = {
  0x0001U,      // Root HUB available Ports Mask
  0U,           // Automatic SPLIT packet handling
  1U,           // Signal Connect event
  1U,           // Signal Disconnect event
  0U            // Signal Overcurrent event
};

#define OTG_FS_USBH_DRIVER_INITIALIZED  (1U << 4)
#define OTG_FS_USBH_DRIVER_POWERED      (1U << 5)

#define BD_OWN_MASK                     (1U << 7)
#define BD_DATA01_MASK                  (1U << 6)
#define BD_KEEP_MASK                    (1U << 5)
#define BD_NINC_MASK                    (1U << 4)
#define BD_DTS_MASK                     (1U << 3)
#define BD_BDT_STALL_MASK               (1U << 2)
#define BD_TOK_PID_MASK                 (0x0FU << 2)
#define BD_TOK_PID(idx)                 ((BD[idx].stat & BD_TOK_PID_MASK) >> 2)
#define BD_IDX(tx, odd)                 (tx * 2U + odd)

#define TOK_PID_DATA0                   (0x03U)
#define TOK_PID_DATA1                   (0x0BU)
#define TOK_PID_ACK                     (0x02U)
#define TOK_PID_STALL                   (0x0EU)
#define TOK_PID_NAK                     (0x0AU)
#define TOK_PID_BUS_TIMEOUT             (0x00U)
#define TOK_PID_DATA_ERROR              (0x0FU)

#define TOKENPID_OUT                    (0x01U)
#define TOKENPID_IN                     (0x09U)
#define TOKENPID_SETUP                  (0x0DU)

typedef struct _BD_t {                                  // Buffer Descriptor table entry
  uint32_t  cmd_stat;
  uint32_t  buf_addr;
} BD_t;

typedef struct {                                        // Pipe structure definition
  uint32_t  packet;
  uint8_t  *data;
  uint32_t  num;
  uint32_t  num_transferred_total;
  uint16_t  num_transferring;
  uint16_t  ep_max_packet_size;
  uint16_t  interval_reload;
  uint16_t  interval;
  uint8_t   dev_addr;
  uint8_t   dev_speed;
  uint8_t   ep_addr;
  uint8_t   ep_type;
  uint8_t   active;
  uint8_t   next_data_toggle;
} PIPE_t;

static ARM_USBH_SignalPortEvent_t SignalPortEvent;
static ARM_USBH_SignalPipeEvent_t SignalPipeEvent;

static ARM_USBH_PORT_STATE usbh_port_state;             // Port state

static volatile uint8_t    hw_busy;                     // Hardware busy flag
static volatile uint8_t    hw_access;                   // Hardware access active flag
static volatile uint8_t    active_bd_index;             // Active Buffer Descriptor table entry
static volatile uint8_t    last_rx_bd_odd;              // Last RX Buffer Descriptor even or odd entry
static volatile uint8_t    last_tx_bd_odd;              // Last TX Buffer Descriptor even or odd entry
static volatile PIPE_t    *ptr_current_transfer_pipe;   // Pointer to currently transferring Pipe
static volatile uint32_t   configured_pipe;             // Bit mask of configured Pipes
static volatile uint32_t   int_pipe_pending;            // Bit mask of interrupt Pipe pending transfer
static volatile uint32_t   ctrl_pipe_pending;           // Bit mask of contrl Pipe pending transfer
static volatile uint32_t   bulk_pipe_pending;           // Bit mask of bulk Pipe pending transfer

static BD_t __align(512)   bd  [1U*2U*2U];              // Buffer Descriptor Table

static volatile PIPE_t     pipe[USBH_MAX_PIPE_NUM];     // Pipes runtime information


// Auxiliary functions

/**
  \fn          void USBH_Reset (void)
  \brief       Reset USB Controller and variables.
*/
static void USBH_Reset (void) {

  USB0->USBTRC0 |=  USB_USBTRC0_USBRESET_MASK;      // Reset OTG FS module
  __nop();__nop();__nop();__nop();__nop();__nop();  // Wait for reset to finish
  USB0->CTL     &= ~USB_CTL_USBENSOFEN_MASK;        // Disable USB
  USB0->CTL     |=  USB_CTL_USBENSOFEN_MASK;        // Enable USB
  USB0->USBCTRL  =  0U;                             // Reset USB CTRL register
                                                    // Reset variables
  memset((void *)&usbh_port_state, 0, sizeof(usbh_port_state));     // Clear Buffer Descriptor Table
  memset((void *)bd,   0, sizeof(bd));              // Clear Buffer Descriptor Table
  hw_busy             = 0U;                         // Hardware is not busy
  hw_access           = 0U;                         // Hardware access not active
  active_bd_index     = 0U;                         // Active Buffer Descriptor table entry is 0
  last_rx_bd_odd      = 1U;                         // Last RX buffer descriptor entry was odd (next will be even)
  last_tx_bd_odd      = 1U;                         // Last TX buffer descriptor entry was odd (next will be even)
  ptr_current_transfer_pipe = NULL;                 // No Pipe is transferring
  int_pipe_pending    = 0U;                         // No interrupt Pipe transfer is pending
  ctrl_pipe_pending   = 0U;                         // No control Pipe transfer is pending
  bulk_pipe_pending   = 0U;                         // No bulk Pipe transfer is pending

  USB0->CTL     |=  USB_CTL_HOSTMODEEN_MASK;        // Force Host mode
  USB0->CTL     |=  USB_CTL_ODDRST_MASK;            // Reset buffer descriptor table entry to EVEN
  USB0->CTL     &= ~USB_CTL_ODDRST_MASK;            // Clear buffer descriptor table entry reset
  USB0->USBCTRL |=  USB_USBCTRL_PDE_MASK;           // Enable weak pull-down and USB transceiver resume
  USB0->SOFTHLD  =  159U;                           // Maximum threshold

  USB0->BDTPAGE1 = ((uint32_t)(bd) >>  8) & 0xFEU;  // Set BDT address
  USB0->BDTPAGE2 = ((uint32_t)(bd) >> 16);
  USB0->BDTPAGE3 = ((uint32_t)(bd) >> 24);

  USB0->ENDPOINT[0].ENDPT = 0U;                     // Disable Endpoint 0 Control register

  USB0->ADDR     =  0U;                             // Clear ADDR (and LSEN bit)
  USB0->ERREN   |=  USB_ERREN_BTSERREN_MASK |       // Enable all error interrupts
                    USB_ERREN_DMAERREN_MASK |
                    USB_ERREN_BTOERREN_MASK |
                    USB_ERREN_DFN8EN_MASK   |
                    USB_ERREN_CRC16EN_MASK  |
                    USB_ERREN_CRC5EOFEN_MASK|
                    USB_ERREN_PIDERREN_MASK ;
  USB0->INTEN    =  USB_INTEN_ATTACHEN_MASK ;       // Enable ATTACH interrupt
}

/**
  \fn          uint32_t USBH_HW_GetPipeIndexFromAddress (PIPE_t *ptr_pipe)
  \brief       Get the Pipe Index from it's Address.
  \param[in]   ptr_pipe Pointer to the Pipe
  \return      Index of the Pipe
*/
__INLINE static uint32_t USBH_HW_GetPipeIndexFromAddress (PIPE_t *ptr_pipe) {
  return (ptr_pipe - pipe);
}

/**
  \fn          PIPE_t *USBH_HW_FindFreePipe (void)
  \brief       Find a free Pipe.
  \return      Pointer to the first free Pipe (NULL = no free Pipe is available)
*/
__INLINE static PIPE_t *USBH_HW_FindFreePipe (void) {
  PIPE_t   *ptr_pipe;
  uint32_t  i;

  ptr_pipe = (PIPE_t *)pipe;

  for (i = 0U; i < USBH_MAX_PIPE_NUM; i++) {
    if (((configured_pipe & (1U << i)) == 0U) && (ptr_pipe->active == 0U)) { return ptr_pipe; }
    ptr_pipe++;
  }

  return NULL;
}

/**
  \fn          bool USBH_HW_StartTransfer (PIPE_t *ptr_pipe)
  \brief       Start transfer on Pipe.
  \param[in]   ptr_pipe Pointer to Pipe
  \return      true = success, false = fail
*/
static bool USBH_HW_StartTransfer (PIPE_t *ptr_pipe) {
  volatile BD_t *ptr_bd;
  uint8_t        addr, endpt0, token;
  uint32_t       num_to_transfer, buf_addr, cmd, timeout_cnt;

  if (usbh_port_state.connected == 0U) { return false; }

  addr = ((ptr_pipe->dev_speed == ARM_USB_SPEED_LOW) << USB_ADDR_LSEN_SHIFT) | ptr_pipe->dev_addr;

  switch (ptr_pipe->ep_type) {
    case ARM_USB_ENDPOINT_CONTROL:
      endpt0 =                           USB_ENDPT_RETRYDIS_MASK | USB_ENDPT_EPRXEN_MASK | USB_ENDPT_EPTXEN_MASK | USB_ENDPT_EPHSHK_MASK;
      break;
    case ARM_USB_ENDPOINT_ISOCHRONOUS:                  // Not supported
//    endpt0 = USB_ENDPT_EPCTLDIS_MASK | USB_ENDPT_RETRYDIS_MASK | USB_ENDPT_EPRXEN_MASK | USB_ENDPT_EPTXEN_MASK | USB_ENDPT_EPHSHK_MASK;
      break;
    case ARM_USB_ENDPOINT_BULK:
      endpt0 = USB_ENDPT_EPCTLDIS_MASK |                           USB_ENDPT_EPRXEN_MASK | USB_ENDPT_EPTXEN_MASK | USB_ENDPT_EPHSHK_MASK;
      break;
    case ARM_USB_ENDPOINT_INTERRUPT:
      endpt0 = USB_ENDPT_EPCTLDIS_MASK | USB_ENDPT_RETRYDIS_MASK | USB_ENDPT_EPRXEN_MASK | USB_ENDPT_EPTXEN_MASK | USB_ENDPT_EPHSHK_MASK;
      break;
    default:
      endpt0 = 0U;
      break;
  }
  if (ptr_pipe->dev_speed == ARM_USB_SPEED_LOW) {       // If low-speed Pipe
    endpt0 |= USB_ENDPT_HOSTWOHUB_MASK;
  }

  cmd = BD_OWN_MASK;                                    // OWN bit to 1
  switch (ptr_pipe->packet & ARM_USBH_PACKET_TOKEN_Msk) {
    case ARM_USBH_PACKET_IN:
      active_bd_index = (last_rx_bd_odd ^ 1U);
      token   = USB_TOKEN_TOKENPID(TOKENPID_IN);
      break;
    case ARM_USBH_PACKET_OUT:
      switch (ptr_pipe->packet & ARM_USBH_PACKET_DATA_Msk) {
        case ARM_USBH_PACKET_DATA1:
          cmd |= BD_DATA01_MASK;                        // Next transfer forced DATA1
        case ARM_USBH_PACKET_DATA0:
                                                        // Next transfer forced DATA0
          break;
        default:
          if (ptr_pipe->next_data_toggle) {
            cmd |= BD_DATA01_MASK;
          }
          ptr_pipe->next_data_toggle ^= 1U;             // Toggle next DATA
          break;
      }
      active_bd_index = (last_tx_bd_odd ^ 1U) + 2U;
      token   = USB_TOKEN_TOKENPID(TOKENPID_OUT);
      break;
    case ARM_USBH_PACKET_SETUP:
      active_bd_index = (last_tx_bd_odd ^ 1U) + 2U;
      endpt0 |= USB_ENDPT_RETRYDIS_MASK;                // Disable NAK retries for setup packet
      token   = USB_TOKEN_TOKENPID(TOKENPID_SETUP);
      break;
  }
  token |= USB_TOKEN_TOKENENDPT(ptr_pipe->ep_addr);
  ptr_bd = &bd[active_bd_index];

  num_to_transfer = ptr_pipe->num-ptr_pipe->num_transferred_total;
  if (num_to_transfer > ptr_pipe->ep_max_packet_size) {
    num_to_transfer = ptr_pipe->ep_max_packet_size;     // Limit data in packet to maximum packet size
  }
  ptr_pipe->num_transferring   = num_to_transfer;

  buf_addr                     = (uint32_t)(ptr_pipe->data+ptr_pipe->num_transferred_total);
  cmd                         |= (num_to_transfer << 16);

  timeout_cnt                  = 1000000U;

  hw_busy                      = 1U;
  ptr_current_transfer_pipe    = ptr_pipe;
  USB0->ADDR                   = addr;
  USB0->ENDPOINT[0].ENDPT      = endpt0;
  ptr_bd->buf_addr             = buf_addr;
  ptr_bd->cmd_stat             = cmd;
  while (timeout_cnt != 0U) {
    if (!(USB0->CTL & USB_CTL_TXSUSPENDTOKENBUSY_MASK)) { break;        }
    if (!timeout_cnt)                                   { return false; }
    timeout_cnt--;
  }
  USB0->TOKEN                  = token;

  return true;
}


// USBH Driver functions

/**
  \fn          ARM_DRIVER_VERSION USBH_GetVersion (void)
  \brief       Get driver version.
  \return      \ref ARM_DRIVER_VERSION
*/
static ARM_DRIVER_VERSION USBH_GetVersion (void) { return usbh_driver_version; }

/**
  \fn          ARM_USBH_CAPABILITIES USBH_GetCapabilities (void)
  \brief       Get driver capabilities.
  \return      \ref ARM_USBH_CAPABILITIES
*/
static ARM_USBH_CAPABILITIES USBH_GetCapabilities (void) { return usbh_driver_capabilities; }

/**
  \fn          int32_t USBH_Initialize (ARM_USBH_SignalPortEvent_t cb_port_event,
                                        ARM_USBH_SignalPipeEvent_t cb_pipe_event)
  \brief       Initialize USB Host Interface.
  \param[in]   cb_port_event  Pointer to \ref ARM_USBH_SignalPortEvent
  \param[in]   cb_pipe_event  Pointer to \ref ARM_USBH_SignalPipeEvent
  \return      \ref execution_status
*/
static int32_t USBH_Initialize (ARM_USBH_SignalPortEvent_t cb_port_event,
                                ARM_USBH_SignalPipeEvent_t cb_pipe_event) {

  if ((otg_fs_state & OTG_FS_USBH_DRIVER_INITIALIZED) != 0U) { return ARM_DRIVER_OK; }

  SignalPortEvent = cb_port_event;
  SignalPipeEvent = cb_pipe_event;

  otg_fs_role  =  ARM_USB_ROLE_HOST;
  otg_fs_state =  OTG_FS_USBH_DRIVER_INITIALIZED;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBH_Uninitialize (void)
  \brief       De-initialize USB Host Interface.
  \return      \ref execution_status
*/
static int32_t USBH_Uninitialize (void) {

  otg_fs_role   =  ARM_USB_ROLE_NONE;
  otg_fs_state &= ~OTG_FS_USBH_DRIVER_INITIALIZED;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBH_PowerControl (ARM_POWER_STATE state)
  \brief       Control USB Host Interface Power.
  \param[in]   state  Power state
  \return      \ref execution_status
*/
static int32_t USBH_PowerControl (ARM_POWER_STATE state) {

  switch (state) {
    case ARM_POWER_OFF:
      NVIC_DisableIRQ      (USB0_IRQn);                 // Disable interrupt
      NVIC_ClearPendingIRQ (USB0_IRQn);                 // Clear pending interrupt
      otg_fs_state  &= ~OTG_FS_USBH_DRIVER_POWERED;     // Clear powered flag
      USB0->INTEN    =  0U;                             // Disable USB interrupts
      USB0->USBTRC0 |=  USB_USBTRC0_USBRESET_MASK;      // Reset OTG FS module
      __nop();__nop();__nop();__nop();__nop();__nop();  // Wait for reset to finish
                                                        // Reset variables
      memset((void *)&usbh_port_state, 0, sizeof(usbh_port_state));     // Clear Buffer Descriptor Table
      memset((void *)bd,   0, sizeof(bd));              // Clear Buffer Descriptor Table
      memset((void *)pipe, 0, sizeof(pipe));            // Clear all Pipes runtime information
      configured_pipe = 0U;                             // No Pipe is configured

      USB0->USBCTRL |=  USB_USBCTRL_SUSP_MASK;          // Suspend USB transceiver
      USB0->CTL     &= ~USB_CTL_USBENSOFEN_MASK;        // Disable USB
      SIM->SCGC4    &= ~SIM_SCGC4_USBOTG_MASK;          // Disable OTG FS clock
      break;

    case ARM_POWER_FULL:
      if ((otg_fs_state & OTG_FS_USBH_DRIVER_POWERED) != 0U) { return ARM_DRIVER_OK; }

      SIM->SOPT2    |=  SIM_SOPT2_USBSRC_MASK;          // MCGFLLCLK, MCGPLLCLK, or IRC48M used as src
      SIM->SCGC4    |=  SIM_SCGC4_USBOTG_MASK;          // Enable OTG FS Clock

      memset((void *)pipe, 0, sizeof(pipe));            // Clear all Pipes runtime information
      configured_pipe     = 0U;                         // No Pipe is configured
      USBH_Reset();                                     // Reset USB Controller and variables, and start USB communication

      otg_fs_state  |=  OTG_FS_USBH_DRIVER_POWERED;     // Set powered flag
      NVIC_SetPriority (USB0_IRQn, 0);                  // Set highest interrupt priority
      NVIC_EnableIRQ   (USB0_IRQn);                     // Enable interrupt
      break;

    default:
      return ARM_DRIVER_ERROR_UNSUPPORTED;
  }

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBH_PortVbusOnOff (uint8_t port, bool vbus)
  \brief       Root HUB Port VBUS on/off.
  \param[in]   port  Root HUB Port Number
  \param[in]   vbus
                - \b false VBUS off
                - \b true  VBUS on
  \return      \ref execution_status
*/
static int32_t USBH_PortVbusOnOff (uint8_t port, bool vbus) {

  if ((otg_fs_state & OTG_FS_USBH_DRIVER_POWERED) == 0U) { return ARM_DRIVER_ERROR;           }
  if (port != 0U)                                        { return ARM_DRIVER_ERROR_PARAMETER; }

  OTG_FS_PinVbusOnOff (vbus);

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBH_PortReset (uint8_t port)
  \brief       Do Root HUB Port Reset.
  \param[in]   port  Root HUB Port Number
  \return      \ref execution_status
*/
static int32_t USBH_PortReset (uint8_t port) {

  if ((otg_fs_state & OTG_FS_USBH_DRIVER_POWERED) == 0U) { return ARM_DRIVER_ERROR;           }
  if (port != 0U)                                        { return ARM_DRIVER_ERROR_PARAMETER; }

  USB0->INTEN  =  0U;                           // Disable all interrupts
  USB0->CTL   |=  USB_CTL_RESET_MASK;           // Generate USB reset
  osDelay (15U);                                // for ~15 ms
  USB0->CTL   &= ~USB_CTL_RESET_MASK;           // Stop USB reset signalling
  USB0->CTL   |=  USB_CTL_USBENSOFEN_MASK;      // Enable SOF generation

  USB0->ERRSTAT=  0xFFU;                        // Clear all error interrupts
  USB0->ISTAT  =  0xFFU;                        // Clear interrupt flags
  USB0->INTEN  =  USB_INTEN_TOKDNEEN_MASK |     // Enable TOKDNE interrupt
                  USB_INTEN_SOFTOKEN_MASK |     // Enable SOFTOKNE interrupt
                  USB_INTEN_ERROREN_MASK  |     // Enable ERROR interrupt
                  USB_INTEN_USBRSTEN_MASK ;     // Enable USBRST interrupt

  SignalPortEvent(port, ARM_USBH_EVENT_RESET);
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBH_PortSuspend (uint8_t port)
  \brief       Suspend Root HUB Port (stop generating SOFs).
  \param[in]   port  Root HUB Port Number
  \return      \ref execution_status
*/
static int32_t USBH_PortSuspend (uint8_t port) {

  if ((otg_fs_state & OTG_FS_USBH_DRIVER_POWERED) == 0U) { return ARM_DRIVER_ERROR;           }
  if (port != 0U)                                        { return ARM_DRIVER_ERROR_PARAMETER; }

  USB0->CTL &= ~USB_CTL_USBENSOFEN_MASK;// Stop generating SOFs

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBH_PortResume (uint8_t port)
  \brief       Resume Root HUB Port (start generating SOFs).
  \param[in]   port  Root HUB Port Number
  \return      \ref execution_status
*/
static int32_t USBH_PortResume (uint8_t port) {

  if ((otg_fs_state & OTG_FS_USBH_DRIVER_POWERED) == 0U) { return ARM_DRIVER_ERROR;           }
  if (port != 0U)                                        { return ARM_DRIVER_ERROR_PARAMETER; }

  USBH_Reset();                         // Reset USB Controller and variables, and start USB communication

  return ARM_DRIVER_OK;
}

/**
  \fn          ARM_USBH_PORT_STATE USBH_PortGetState (uint8_t port)
  \brief       Get current Root HUB Port State.
  \param[in]   port  Root HUB Port Number
  \return      Port State \ref ARM_USBH_PORT_STATE
*/
static ARM_USBH_PORT_STATE USBH_PortGetState (uint8_t port) {
  ARM_USBH_PORT_STATE port_state = { 0U, 0U, 0U };

  if ((otg_fs_state & OTG_FS_USBH_DRIVER_POWERED) == 0U) { return port_state; }
  if (port != 0U)                                        { return port_state; }

  port_state = usbh_port_state;
  port_state.overcurrent = OTG_FS_PinGetOC ();

  return port_state;
}

/**
  \fn          ARM_USBH_PIPE_HANDLE USBH_PipeCreate (uint8_t  dev_addr,
                                                     uint8_t  dev_speed,
                                                     uint8_t  hub_addr,
                                                     uint8_t  hub_port,
                                                     uint8_t  ep_addr,
                                                     uint8_t  ep_type,
                                                     uint16_t ep_max_packet_size,
                                                     uint8_t  ep_interval)
  \brief       Create Pipe in System.
  \param[in]   dev_addr   Device Address
  \param[in]   dev_speed  Device Speed
  \param[in]   hub_addr   Hub Address
  \param[in]   hub_port   Hub Port
  \param[in]   ep_addr    Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \param[in]   ep_type    Endpoint Type (ARM_USB_ENDPOINT_xxx)
  \param[in]   ep_max_packet_size Endpoint Maximum Packet Size
  \param[in]   ep_interval        Endpoint Polling Interval
  \return      Pipe Handle \ref ARM_USBH_PIPE_HANDLE
*/
static ARM_USBH_PIPE_HANDLE USBH_PipeCreate (uint8_t dev_addr, uint8_t dev_speed, uint8_t hub_addr, uint8_t hub_port, uint8_t ep_addr, uint8_t ep_type, uint16_t ep_max_packet_size, uint8_t  ep_interval) {
  PIPE_t *ptr_pipe;

  if ((otg_fs_state & OTG_FS_USBH_DRIVER_POWERED) == 0U) { return NULL; }

  ptr_pipe = USBH_HW_FindFreePipe ();           // Find free Pipe
  if (ptr_pipe == 0U) { return NULL; }          // If no free Pipe

  memset((void *)ptr_pipe, 0, sizeof(PIPE_t));  // Clear Pipes runtime information

  // Store Pipe settings
  ptr_pipe->dev_addr              = dev_addr;
  ptr_pipe->dev_speed             = dev_speed;
  ptr_pipe->ep_addr               = ep_addr;
  ptr_pipe->ep_type               = ep_type;
  ptr_pipe->ep_max_packet_size    = ep_max_packet_size;
  if (ep_type == ARM_USB_ENDPOINT_INTERRUPT) {
    ptr_pipe->interval_reload     = ep_interval;
    ptr_pipe->interval            = ep_interval;
  } else {
    ptr_pipe->interval            = 0U;
    ptr_pipe->interval_reload     = 0U;
  }

  // Set flag in configured_pipe for Pipe we have configured
  configured_pipe |=  (1U << USBH_HW_GetPipeIndexFromAddress(ptr_pipe));

  return ((ARM_USBH_EP_HANDLE)ptr_pipe);
}

/**
  \fn          int32_t USBH_PipeModify (ARM_USBH_PIPE_HANDLE pipe_hndl,
                                        uint8_t              dev_addr,
                                        uint8_t              dev_speed,
                                        uint8_t              hub_addr,
                                        uint8_t              hub_port,
                                        uint16_t             ep_max_packet_size)
  \brief       Modify Pipe in System.
  \param[in]   pipe_hndl  Pipe Handle
  \param[in]   dev_addr   Device Address
  \param[in]   dev_speed  Device Speed
  \param[in]   hub_addr   Hub Address
  \param[in]   hub_port   Hub Port
  \param[in]   ep_max_packet_size Endpoint Maximum Packet Size
  \return      \ref execution_status
*/
static int32_t USBH_PipeModify (ARM_USBH_PIPE_HANDLE pipe_hndl, uint8_t dev_addr, uint8_t dev_speed, uint8_t hub_addr, uint8_t hub_port, uint16_t ep_max_packet_size) {
  PIPE_t   *ptr_pipe;

  if ((otg_fs_state & OTG_FS_USBH_DRIVER_POWERED) == 0U) { return ARM_DRIVER_ERROR;           }
  if (pipe_hndl == 0U)                                   { return ARM_DRIVER_ERROR_PARAMETER; }

  ptr_pipe = (PIPE_t *)(pipe_hndl);
  if (ptr_pipe->active != 0U)                            { return ARM_DRIVER_ERROR_BUSY;      }

  // Change modified Pipe settings
  ptr_pipe->dev_addr           = dev_addr;
  ptr_pipe->dev_speed          = dev_speed;
  ptr_pipe->ep_max_packet_size = ep_max_packet_size;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBH_PipeDelete (ARM_USBH_PIPE_HANDLE pipe_hndl)
  \brief       Delete Pipe from System.
  \param[in]   pipe_hndl  Pipe Handle
  \return      \ref execution_status
*/
static int32_t USBH_PipeDelete (ARM_USBH_PIPE_HANDLE pipe_hndl) {
  PIPE_t *ptr_pipe;

  if ((otg_fs_state & OTG_FS_USBH_DRIVER_POWERED) == 0U) { return ARM_DRIVER_ERROR;           }
  if (pipe_hndl == 0U)                                   { return ARM_DRIVER_ERROR_PARAMETER; }

  ptr_pipe = (PIPE_t *)(pipe_hndl);
  if (ptr_pipe->active != 0U)                            { return ARM_DRIVER_ERROR_BUSY;      }

  // Clear all fields of Pipe structure
  memset((void *)ptr_pipe, 0, sizeof(PIPE_t));

  // Clear flag in configured_pipe for Pipe we have deleted
  configured_pipe &= ~(1U << USBH_HW_GetPipeIndexFromAddress(ptr_pipe));

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBH_PipeReset (ARM_USBH_PIPE_HANDLE pipe_hndl)
  \brief       Reset Pipe.
  \param[in]   pipe_hndl  Pipe Handle
  \return      \ref execution_status
*/
static int32_t USBH_PipeReset (ARM_USBH_PIPE_HANDLE pipe_hndl) {
  PIPE_t *ptr_pipe;

  if ((otg_fs_state & OTG_FS_USBH_DRIVER_POWERED) == 0U) { return ARM_DRIVER_ERROR;           }
  if (pipe_hndl == 0U)                                   { return ARM_DRIVER_ERROR_PARAMETER; }

  ptr_pipe = (PIPE_t *)(pipe_hndl);
  if (ptr_pipe->active != 0U)                            { return ARM_DRIVER_ERROR_BUSY;      }

  ptr_pipe->next_data_toggle = 0U;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBH_PipeTransfer (ARM_USBH_PIPE_HANDLE pipe_hndl,
                                          uint32_t             packet,
                                          uint8_t             *data,
                                          uint32_t             num)
  \brief       Transfer packets through USB Pipe.
  \param[in]   pipe_hndl  Pipe Handle
  \param[in]   packet     Packet information
  \param[in]   data       Pointer to buffer with data to send or for data to receive
  \param[in]   num        Number of data bytes to transfer
  \return      \ref execution_status
*/
static int32_t USBH_PipeTransfer (ARM_USBH_PIPE_HANDLE pipe_hndl, uint32_t packet, uint8_t *data, uint32_t num) {
  PIPE_t   *ptr_pipe;

  if ((otg_fs_state & OTG_FS_USBH_DRIVER_POWERED) == 0U) { return ARM_DRIVER_ERROR;           }
  if (pipe_hndl == 0U)                                   { return ARM_DRIVER_ERROR_PARAMETER; }
  if (usbh_port_state.connected == 0U)                   { return ARM_DRIVER_ERROR;           }

  ptr_pipe = (PIPE_t *)(pipe_hndl);
  if (ptr_pipe->active != 0U)                            { return ARM_DRIVER_ERROR_BUSY;      }

  // Update current transfer information
  ptr_pipe->packet                = packet;
  ptr_pipe->data                  = data;
  ptr_pipe->num                   = num;
  ptr_pipe->num_transferred_total = 0U;
  ptr_pipe->num_transferring      = 0U;

  hw_access  = 1U;
  while (hw_busy != 0U);

  ptr_pipe->active = 1U;
  if (USBH_HW_StartTransfer (ptr_pipe) == false) {
    ptr_pipe->active = 0U;
    return ARM_DRIVER_ERROR;
  }
  hw_access  = 0U;

  return ARM_DRIVER_OK;
}

/**
  \fn          uint32_t USBH_PipeTransferGetResult (ARM_USBH_PIPE_HANDLE pipe_hndl)
  \brief       Get result of USB Pipe transfer.
  \param[in]   pipe_hndl  Pipe Handle
  \return      number of successfully transferred data bytes
*/
static uint32_t USBH_PipeTransferGetResult (ARM_USBH_PIPE_HANDLE pipe_hndl) {

  if (pipe_hndl == 0U) { return 0U; }

  return (((PIPE_t *)(pipe_hndl))->num_transferred_total);
}

/**
  \fn          int32_t USBH_PipeTransferAbort (ARM_USBH_PIPE_HANDLE pipe_hndl)
  \brief       Abort current USB Pipe transfer.
  \param[in]   pipe_hndl  Pipe Handle
  \return      \ref execution_status
*/
static int32_t USBH_PipeTransferAbort (ARM_USBH_PIPE_HANDLE pipe_hndl) {
  PIPE_t *ptr_pipe;

  if ((otg_fs_state & OTG_FS_USBH_DRIVER_POWERED) == 0U) { return ARM_DRIVER_ERROR;           }
  if (pipe_hndl == 0U)                                   { return ARM_DRIVER_ERROR_PARAMETER; }

  ptr_pipe = (PIPE_t *)(pipe_hndl);

  if (ptr_pipe->active != 0U) {
    ptr_pipe->active = 0U;
    if (ptr_current_transfer_pipe == (PIPE_t *)ptr_pipe) {
      // Wait for active transfer to finish as there is no way to cancel already
      // active transfer on hardware
      hw_access = 1U;
      while (hw_busy != 0U) {
        if (usbh_port_state.connected == 0U) {
          hw_busy = 0U;
        }
      }
      hw_access = 0U;
    }
    ptr_pipe->num_transferring = 0U;    // Clear abort affected transfer information
  }

  return ARM_DRIVER_OK;
}

/**
  \fn          uint16_t USBH_GetFrameNumber (void)
  \brief       Get current USB Frame Number.
  \return      Frame Number
*/
static uint16_t USBH_GetFrameNumber (void) {

  if ((otg_fs_state & OTG_FS_USBH_DRIVER_POWERED) == 0U) { return 0U; }

  return ((((uint32_t)(USB0->FRMNUML)) | (((uint32_t)(USB0->FRMNUMH)) << 8)) & 0x07FFU);
}

/**
  \fn          void USBH_FS_IRQ (void)
  \brief       USB Host Interrupt Routine (IRQ).
*/
void USBH_FS_IRQ (void) {
  volatile PIPE_t *ptr_pipe;
  volatile BD_t   *ptr_bd;
  uint32_t         istat, stat, errstat, pipe_msk;
  uint8_t          tx, odd, event, pipe_num;

  istat   = USB0->ISTAT & USB0->INTEN;                  // Read active interrupts
  stat    = USB0->STAT;
  errstat = USB0->ERRSTAT;
  USB0->ISTAT = istat;                                  // Clear interrupts

  if ((istat & USB_ISTAT_ATTACH_MASK) != 0U) {          // Attach interrupt
    if ((USB0->CTL & USB_CTL_JSTATE_MASK) != 0U) {      // If full-speed
      usbh_port_state.speed = ARM_USB_SPEED_FULL;
    } else {                                            // If low-speed
      USB0->ADDR = 0x80U;
      usbh_port_state.speed = ARM_USB_SPEED_LOW;
    }
    usbh_port_state.connected = 1U;
    SignalPortEvent(0, ARM_USBH_EVENT_CONNECT);
    USB0->INTEN  &= ~USB_INTEN_ATTACHEN_MASK;           // Disable attach interrupt
  }

  if ((istat & USB_ISTAT_USBRST_MASK) != 0U) {          // Reset interrupt
    if ((usbh_port_state.connected) != 0U) {
      usbh_port_state.connected = 0U;
      ptr_pipe = pipe;                                  // Deactivate all active Pipes
      for (pipe_num = 0; pipe_num < USBH_MAX_PIPE_NUM; pipe_num++) {
        if ((configured_pipe >> pipe_num) == 0U) { break; }
        if (ptr_pipe->active              != 0U) { 
          ptr_pipe->active = 0U;
          SignalPipeEvent((ARM_USBH_PIPE_HANDLE)ptr_pipe, ARM_USBH_EVENT_BUS_ERROR);
        }
        ptr_pipe++;
      }
      SignalPortEvent(0, ARM_USBH_EVENT_DISCONNECT);
    }
    USB0->ADDR    =  0U;                                // Clear LSEN bit
    USB0->INTEN  |=  USB_INTEN_ATTACHEN_MASK;           // Enable attach interrupt
  }

  if ((istat & USB_ISTAT_ERROR_MASK) != 0U) {           // Error interrupt
    if (ptr_current_transfer_pipe != 0U) {
      SignalPipeEvent((ARM_USBH_EP_HANDLE)ptr_current_transfer_pipe, ARM_USBH_EVENT_BUS_ERROR);
    }
    USB0->ERRSTAT = errstat;                            // Clear all errors
  }

  if ((istat & USB_ISTAT_TOKDNE_MASK) != 0U) {          // Token interrupt
    tx       = (stat >> 3) & 0x01U;
    odd      = (stat >> 2) & 0x01U;
    ptr_bd   = &bd[BD_IDX(tx,odd)];
    pipe_msk = (1U << USBH_HW_GetPipeIndexFromAddress((PIPE_t *)ptr_current_transfer_pipe));
    event    =  0U;
    if (tx) { last_tx_bd_odd = odd; }
    else    { last_rx_bd_odd = odd; }
    switch ((ptr_bd->cmd_stat & BD_TOK_PID_MASK) >> 2U) {
      case TOK_PID_DATA0:                               // If DATA0 packet received
      case TOK_PID_DATA1:                               // If DATA1 packet received
        ptr_current_transfer_pipe->num_transferred_total += (ptr_bd->cmd_stat >> 16) & 0x3FFU;
        ptr_current_transfer_pipe->num_transferring       =  0U;
        if ((ptr_current_transfer_pipe->num == ptr_current_transfer_pipe->num_transferred_total)     || // If all data was transferred or
            (((ptr_bd->cmd_stat >> 16) & 0x3FF) == 0U)                                               || // if zero-length packet was received or
            (((ptr_bd->cmd_stat >> 16) & 0x3FF)%ptr_current_transfer_pipe->ep_max_packet_size) != 0U) { // if short packet was received
          ptr_current_transfer_pipe->active = 0U;
          event = ARM_USBH_EVENT_TRANSFER_COMPLETE;
        } else {                                                                                        // else more data to transfer
          switch (ptr_current_transfer_pipe->ep_type) {
            case ARM_USB_ENDPOINT_CONTROL:
              ctrl_pipe_pending |= pipe_msk;
              break;
            case ARM_USB_ENDPOINT_BULK:
              bulk_pipe_pending |= pipe_msk;
              break;
            default:
              break;
          }
        }
        break;
      case TOK_PID_ACK:                                 // If packet sent
        ptr_current_transfer_pipe->num_transferred_total += ptr_current_transfer_pipe->num_transferring;
        ptr_current_transfer_pipe->num_transferring       = 0U;
        if (ptr_current_transfer_pipe->num == ptr_current_transfer_pipe->num_transferred_total) {       // If all data was transferred
          ptr_current_transfer_pipe->active = 0U;
          event = ARM_USBH_EVENT_TRANSFER_COMPLETE;
        } else {                                                                                        // else more data to transfer
          ptr_current_transfer_pipe->packet &= ~ARM_USBH_PACKET_DATA_Msk;       // Clear forced DATA toggle
          switch (ptr_current_transfer_pipe->ep_type) {
            case ARM_USB_ENDPOINT_CONTROL:
              ctrl_pipe_pending |= pipe_msk;
              break;
            case ARM_USB_ENDPOINT_BULK:
              bulk_pipe_pending |= pipe_msk;
              break;
            default:
              break;
          }
        }
        break;
      case TOK_PID_STALL:
        ptr_current_transfer_pipe->active = 0U;
        event = ARM_USBH_EVENT_HANDSHAKE_STALL;
        break;
      case TOK_PID_NAK:
        if (ptr_current_transfer_pipe->active != 0U) {
          // If current pipe is still active restart the transmission on NAK
          switch (ptr_current_transfer_pipe->ep_type) {
            case ARM_USB_ENDPOINT_CONTROL:
              ctrl_pipe_pending |= pipe_msk;
              break;
            case ARM_USB_ENDPOINT_BULK:
              bulk_pipe_pending |= pipe_msk;
              break;
            default:
              break;
          }
        }
        break;
      case TOK_PID_BUS_TIMEOUT:
        ptr_current_transfer_pipe->active = 0U;
        event = ARM_USBH_EVENT_BUS_ERROR;
        break;
      case TOK_PID_DATA_ERROR:
        ptr_current_transfer_pipe->active = 0U;
        event = ARM_USBH_EVENT_BUS_ERROR;
        break;
      default:
        break;
    }
    if (event != 0U) { SignalPipeEvent((ARM_USBH_EP_HANDLE)ptr_current_transfer_pipe, event); }
  }

  if ((istat & USB_ISTAT_SOFTOK_MASK) != 0U) {  // If SOFTOK interrupt occurred
    ptr_pipe = pipe;
    for (pipe_num = 0U; pipe_num < USBH_MAX_PIPE_NUM; pipe_num++) {
      if ((configured_pipe >> pipe_num) == 0U){ // If all configured endpoints were handled
        break;
      }
      // If interrupt transfer is active handle period (interval)
      if ((ptr_pipe->ep_type == ARM_USB_ENDPOINT_INTERRUPT) && (ptr_pipe->active != 0U) && (ptr_pipe->interval != 0U)) { 
        ptr_pipe->interval--;
        if (ptr_pipe->interval == 0U) {
          ptr_pipe->interval = ptr_pipe->interval_reload;
          int_pipe_pending  |= (1U << pipe_num);
        }
      }
      ptr_pipe++;
    }
  }

  if ((int_pipe_pending == 0U) && (ctrl_pipe_pending == 0U) && (bulk_pipe_pending == 0U)) {     // If no transfer is pending
    hw_busy = 0U;
  } else if (hw_access == 0U){                                          // If HW is accessed from application, skip activating new transfer
    if (int_pipe_pending != 0U) {                                       // If there is interrupt transfer pending
      ptr_pipe = pipe;
      for (pipe_num = 0U; pipe_num < USBH_MAX_PIPE_NUM; pipe_num++) {
        if ((configured_pipe >> pipe_num) == 0U) { break; }             // If all configured Pipes were handled
        if ((int_pipe_pending & (1U << pipe_num)) != 0U) {
          if (ptr_pipe->active != 0U) {
            if (USBH_HW_StartTransfer ((PIPE_t *)ptr_pipe) == 0U) { ptr_pipe->active = 0U; }
          }
          int_pipe_pending &= ~(1U << pipe_num);
        }
        ptr_pipe++;
      }
    } else if (ctrl_pipe_pending != 0U) {                               // If there is control transfer pending
      ptr_pipe = pipe;
      for (pipe_num = 0U; pipe_num < USBH_MAX_PIPE_NUM; pipe_num++) {
        if ((configured_pipe >> pipe_num) == 0U) { break; }             // If all configured Pipes were handled
        if ((ctrl_pipe_pending & (1U << pipe_num)) != 0U) {
          if (ptr_pipe->active != 0U) {
            if (USBH_HW_StartTransfer ((PIPE_t *)ptr_pipe) == 0U) { ptr_pipe->active = 0U; }
          }
          ctrl_pipe_pending &= ~(1U << pipe_num);
        }
        ptr_pipe++;
      }
    } else if (bulk_pipe_pending != 0U) {                               // If there is bulk transfer pending
      ptr_pipe = pipe;
      for (pipe_num = 0U; pipe_num < USBH_MAX_PIPE_NUM; pipe_num++) {
        if ((configured_pipe >> pipe_num) == 0U) { break; }              // If all configured Pipes were handled
        if ((bulk_pipe_pending & (1U << pipe_num)) != 0U) {
          if (ptr_pipe->active != 0U) {
            if (USBH_HW_StartTransfer ((PIPE_t *)ptr_pipe) == 0U) { ptr_pipe->active = 0U; }
          }
          bulk_pipe_pending &= ~(1U << pipe_num);
        }
        ptr_pipe++;
      }
    }
  }
}

ARM_DRIVER_USBH Driver_USBH0 = {
  USBH_GetVersion,
  USBH_GetCapabilities,
  USBH_Initialize,
  USBH_Uninitialize,
  USBH_PowerControl,
  USBH_PortVbusOnOff,
  USBH_PortReset,
  USBH_PortSuspend,
  USBH_PortResume,
  USBH_PortGetState,
  USBH_PipeCreate,
  USBH_PipeModify,
  USBH_PipeDelete,
  USBH_PipeReset,
  USBH_PipeTransfer,
  USBH_PipeTransferGetResult,
  USBH_PipeTransferAbort,
  USBH_GetFrameNumber
};
