/*-----------------------------------------------------------------------------
 * Name:    FXOS8700.h
 * Purpose: FXOS8700CQ 6-Axis Accelerometer and Magnetometer interface
 * Rev.:    1.00
 *-----------------------------------------------------------------------------*/

/* Copyright (c) 2013 - 2014 ARM LIMITED

   All rights reserved.
   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:
   - Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   - Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   - Neither the name of ARM nor the names of its contributors may be used
     to endorse or promote products derived from this software without
     specific prior written permission.
   *
   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
   ARE DISCLAIMED. IN NO EVENT SHALL COPYRIGHT HOLDERS AND CONTRIBUTORS BE
   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
   POSSIBILITY OF SUCH DAMAGE.
   ---------------------------------------------------------------------------*/

#include "Driver_I2C.h"
#include "Board_Accelerometer.h"
#include "Board_Magnetometer.h"
#include "RTE_Components.h"

#ifdef RTE_CMSIS_RTOS
#include "cmsis_os.h"
#endif

#ifndef FXOS8700_I2C_PORT
#define FXOS8700_I2C_PORT       0
#endif

#ifndef FXOS8700_I2C_ADDR
#define FXOS8700_I2C_ADDR       0x1D
#endif

#define FXOS8700_ID             0xC7

/* I2C Driver */
#define _I2C_Driver_(n)  Driver_I2C##n
#define  I2C_Driver_(n) _I2C_Driver_(n)
extern ARM_DRIVER_I2C    I2C_Driver_(FXOS8700_I2C_PORT);
#define ptrI2C         (&I2C_Driver_(FXOS8700_I2C_PORT))

/* Register addresses */
#define OUT_X_MSB               0x01
#define OUT_X_LSB               0x02
#define OUT_Y_MSB               0x03
#define OUT_Y_LSB               0x04
#define OUT_Z_MSB               0x05
#define OUT_Z_LSB               0x06

#define WHO_AM_I                0x0D
#define XYZ_DATA_CFG            0x0E

#define CTRL_REG1               0x2A
#define CTRL_REG2               0x2B

#define M_OUT_X_MSB             0x33
#define M_OUT_X_LSB             0x34
#define M_OUT_Y_MSB             0x35
#define M_OUT_Y_LSB             0x36
#define M_OUT_Z_MSB             0x37
#define M_OUT_Z_LSB             0x38

#define M_CTRL_REG1             0x5B
#define M_CTRL_REG2             0x5C
#define M_CTRL_REG3             0x5D

/* Driver status */
#define STATUS_INIT             0x01
#define STATUS_INIT_ACCEL       0x02
#define STATUS_INIT_MAGNETO     0x04
static int8_t Status;

#ifdef RTE_CMSIS_RTOS
static osSemaphoreDef(Semaphore);
static osSemaphoreId  Semaphore_Id;
#endif


/**
  \fn          int32_t Register_Read (uint8_t reg, uint8_t *val)
  \brief       Read value from device register
  \param[in]   reg    Register address
  \param[out]  val    Pointer where data will be read from register
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
static int32_t Register_Read (uint8_t reg, uint8_t *val) {
  uint8_t data[1];

  data[0] = reg;
  ptrI2C->MasterTransmit(FXOS8700_I2C_ADDR, data, 1, true);
  while (ptrI2C->GetStatus().busy);
  if (ptrI2C->GetDataCount() != 1) { return -1; }
  ptrI2C->MasterReceive (FXOS8700_I2C_ADDR, val, 1, false);
  while (ptrI2C->GetStatus().busy);
  if (ptrI2C->GetDataCount() != 1) { return -1; }

  return 0;
}

/**
  \fn          int32_t Register_WriteData (uint8_t reg, const uint8_t *val)
  \brief       Write value to device register
  \param[in]   reg    Register address
  \param[in]   val    Pointer with data to write to register
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
static int32_t Register_Write (uint8_t reg, uint8_t val) {
  uint8_t data[2];

  data[0] = reg;
  data[1] = val;
  ptrI2C->MasterTransmit(FXOS8700_I2C_ADDR, data, 2, false);
  while (ptrI2C->GetStatus().busy);
  if (ptrI2C->GetDataCount() != 2) { return -1; }

  return 0;
}

/**
  \fn          int32_t Initialize (void)
  \brief       Initialize magnetometer and accelerometer device
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t Initialize (void) {
  uint8_t who;
  int32_t res;

  if ((Status & STATUS_INIT) != 0) {
    return 0;
  }

  #ifdef RTE_CMSIS_RTOS
  Semaphore_Id = osSemaphoreCreate (osSemaphore(Semaphore), 1);
  if (Semaphore_Id == NULL) {
    return -1;
  }
  #endif

  ptrI2C->Initialize  (NULL);
  ptrI2C->PowerControl(ARM_POWER_FULL);
  ptrI2C->Control     (ARM_I2C_BUS_SPEED, ARM_I2C_BUS_SPEED_FAST);
  ptrI2C->Control     (ARM_I2C_BUS_CLEAR, 0);

  /* Identify FXOS8700 device */
  who = 0;
  res = Register_Read(WHO_AM_I, &who);
  if (res == 0) {
    if (who != FXOS8700_ID) {
      return -1;                        /* Wrong ID */
    }
  }

  Register_Write(CTRL_REG1,    0x00);   /* Put device into standby */
  Register_Write(CTRL_REG2,    0x00);
  Register_Write(XYZ_DATA_CFG, 0x00);
  Register_Write(CTRL_REG1,    0x0D);   /* Enable active mode */
  Register_Write(M_CTRL_REG2,  0x00);
  Register_Write(M_CTRL_REG1,  0x03);

  Status = STATUS_INIT;

  return 0;
}


/**
  \fn          int32_t Accelerometer_Initialize (void)
  \brief       Initialize accelerometer
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t Accelerometer_Initialize (void) {

  if (Initialize() == 0) {
    Status |= STATUS_INIT_ACCEL;
    return 0;
  }
  return -1;
}

/**
  \fn          int32_t Accelerometer_Uninitialize (void)
  \brief       De-initialize accelerometer
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t Accelerometer_Uninitialize (void) {
  Status &= ~STATUS_INIT_ACCEL;
  return 0;
}

/**
  \fn          int32_t Accelerometer_GetState (ACCELEROMETER_STATE *pState)
  \brief       Get accelerometer state
  \param[out]  pState  pointer to ACCELEROMETER_STATE structure
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t Accelerometer_GetState (ACCELEROMETER_STATE *pState) {
  uint8_t val[6];
  uint8_t reg;
  int16_t mg;

  if ((Status & STATUS_INIT_ACCEL) == 0) {
    return -1;
  }

#ifdef RTE_CMSIS_RTOS
  if (osSemaphoreWait(Semaphore_Id, 500) == 0) {
    /* No semaphore available */
    return -1;
  }
#endif
  reg = OUT_X_MSB;

  ptrI2C->MasterTransmit(FXOS8700_I2C_ADDR, &reg, 1, true);
  while (ptrI2C->GetStatus().busy);
  if (ptrI2C->GetDataCount() != 1) { return -1; }
  ptrI2C->MasterReceive (FXOS8700_I2C_ADDR, &val[0], 6, false);
  while (ptrI2C->GetStatus().busy);
  if (ptrI2C->GetDataCount() != 6) { return -1; }

  mg = (val[0] << 8) | (val[1] >> 2);
  pState->x = (mg * 244) / 4096;
  mg = (val[2] << 8) | (val[3] >> 2);
  pState->y = (mg * 244) / 4096;
  mg = (val[4] << 8) | (val[5] >> 2);
  pState->z = (mg * 244) / 4096;

#ifdef RTE_CMSIS_RTOS
  osSemaphoreRelease (Semaphore_Id);
#endif
  return 0;
}


/**
  \fn          int32_t Magnetometer_Initialize (void)
  \brief       Initialize magnetometer
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t Magnetometer_Initialize (void) {

  if (Initialize() == 0) {
    Status |= STATUS_INIT_MAGNETO;
    return 0;
  }
  return -1;
}

/**
  \fn          int32_t Magnetometer_Uninitialize (void)
  \brief       De-initialize magnetometer
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t Magnetometer_Uninitialize (void) {
  Status &= ~STATUS_INIT_MAGNETO;
  return (0);
}

/**
  \fn          int32_t Magnetometer_GetState (MAGNETOMETER_STATE *pState)
  \brief       Get magnetometer state
  \param[out]  pState  pointer to MAGNETOMETER_STATE structure
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t Magnetometer_GetState (MAGNETOMETER_STATE *pState) {
  uint8_t val[6];
  uint8_t reg;
  int16_t uT;

  if ((Status & STATUS_INIT_MAGNETO) == 0) {
    return -1;
  }

#ifdef RTE_CMSIS_RTOS
  if (osSemaphoreWait(Semaphore_Id, 500) == 0) {
    /* No semaphore available */
    return -1;
  }
#endif

  reg = M_OUT_X_MSB;

  ptrI2C->MasterTransmit(FXOS8700_I2C_ADDR, &reg, 1, true);
  while (ptrI2C->GetStatus().busy);
  if (ptrI2C->GetDataCount() != 1) { return -1; }
  ptrI2C->MasterReceive (FXOS8700_I2C_ADDR, &val[0], 6, false);
  while (ptrI2C->GetStatus().busy);
  if (ptrI2C->GetDataCount() != 6) { return -1; }
  
  uT = (val[0] << 8) | val[1];
  pState->x = uT;
  uT = (val[2] << 8) | val[3];
  pState->y = uT;
  uT = (val[4] << 8) | val[5];
  pState->z = uT;

#ifdef RTE_CMSIS_RTOS
  osSemaphoreRelease (Semaphore_Id);
#endif
  return 0;
}
